/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl;

import java.util.Calendar;
import java.util.GregorianCalendar;

/**
 * Calendar utilities and constants.
 *
 * @author johan.walles@gmail.com
 */
public class CalendarUtil {
    /**
     * Month names in Swedish.
     */
    public final static String[] MONTHS = new String[] {
        "januari", "februari", "mars",
        "april", "maj", "juni",
        "juli", "augusti", "september",
        "oktober", "november", "december"
    };

    /**
     * Names of weekdays in Swedish, starting on Sunday.
     */
    public final static String[] DAYS = new String[] {
        "söndag", "måndag", "tisdag", "onsdag", "torsdag", "fredag", "lördag"
    };

    /**
     * Are these two dates the same day?
     *
     * @param a A date to compare
     * @param b A date to compare
     * @return True if the dates are the same day.  False otherwise.
     */
    public static boolean sameDay(Calendar a, Calendar b) {
        if (a.get(Calendar.DAY_OF_YEAR) != b.get(Calendar.DAY_OF_YEAR)) {
            return false;
        }
        if (a.get(Calendar.YEAR) != b.get(Calendar.YEAR)) {
            return false;
        }
        return true;
    }

    /**
     * Format a date like "Idag måndag 12 januari".
     *
     * @param when The date to render into a string.
     *
     * @return A human readable (for humans who read Swedish) string describing
     * the date.
     */
    public static CharSequence formatDate(Calendar when) {
        StringBuilder result = new StringBuilder(40);
        Calendar reference = Calendar.getInstance();

        boolean sameYear = true;
        if (reference.get(Calendar.YEAR) != when.get(Calendar.YEAR)) {
            sameYear = false;
        }

        boolean hasPrefix = false;
        if (sameDay(when, reference)) {
            hasPrefix = true;
            result.append("i dag ");
        }
        reference.add(Calendar.DAY_OF_MONTH, 1);
        if (sameDay(when, reference)) {
            hasPrefix = true;
            result.append("i morgon ");
        }
        reference.add(Calendar.DAY_OF_MONTH, 1);
        if (sameDay(when, reference)) {
            hasPrefix = true;
            result.append("i övermorgon ");
        }

        result.append(DAYS[when.get(Calendar.DAY_OF_WEEK) - 1]);
        result.append(" ");
        result.append(when.get(Calendar.DAY_OF_MONTH));
        result.append(" ");
        result.append(MONTHS[when.get(Calendar.MONTH)]);

        if ((!sameYear) && (!hasPrefix)) {
            result.append(" ");
            result.append(when.get(Calendar.YEAR));
        }

        result.setCharAt(0, Character.toUpperCase(result.charAt(0)));

        return result;
    }

    /**
     * Describe a point in time for telling the user the age of an old location
     * fix.
     *
     * @param when Time stamp of the location fix.
     *
     * @return A string in Swedish describing the time, like "2 timmar gammal",
     * or "från 9 december 1973".
     */
    public static CharSequence describeTime(long when) {
        long minutesAgo = (System.currentTimeMillis() - when) / 60000L;
        if (minutesAgo < 1) {
            return "mindre än en minut gammal";
        }

        if (minutesAgo < 2) {
            return "en minut gammal";
        }

        if (minutesAgo < 60) {
            return minutesAgo + " minuter gammal";
        }

        long hoursAgo = minutesAgo / 60;
        if (hoursAgo < 2) {
            return "en timme gammal";
        }
        if (hoursAgo < 24) {
            return hoursAgo + " timmar gammal";
        }

        // Just put the date in there
        Calendar date = new GregorianCalendar();
        int currentYear = date.get(Calendar.YEAR);

        date.setTimeInMillis(when);
        int year = date.get(Calendar.YEAR);
        int month = date.get(Calendar.MONTH);
        int day = date.get(Calendar.DAY_OF_MONTH);

        StringBuilder builder = new StringBuilder(30);
        builder.append("från ");
        builder.append(day);
        builder.append(" ");
        builder.append(CalendarUtil.MONTHS[month]);
        if (year != currentYear) {
            builder.append(" ");
            builder.append(year);
        }
        return builder;
    }
}
