/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl;

import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;

import net.launchpad.sl.LocationSelectionAdapter.CompletionsExistListener;
import net.launchpad.sl.location.LocationUtil;
import net.launchpad.sl.location.LocationsDbAdapter;
import net.launchpad.sl.location.NamedLocation;
import net.launchpad.sl.location.ScoredLocation;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.ContentUris;
import android.content.DialogInterface;
import android.content.Intent;
import android.database.Cursor;
import android.graphics.drawable.Drawable;
import android.location.Address;
import android.location.Geocoder;
import android.net.Uri;
import android.os.Bundle;
import android.provider.Contacts;
import android.provider.Contacts.ContactMethodsColumns;
import android.provider.Contacts.People;
import android.provider.Contacts.PeopleColumns;
import android.text.Editable;
import android.text.Spannable;
import android.text.SpannableStringBuilder;
import android.text.TextWatcher;
import android.text.method.LinkMovementMethod;
import android.text.style.URLSpan;
import android.util.Log;
import android.view.KeyEvent;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.inputmethod.EditorInfo;
import android.widget.AdapterView;
import android.widget.AutoCompleteTextView;
import android.widget.Button;
import android.widget.TextView;
import android.widget.TextView.BufferType;
import android.widget.TextView.OnEditorActionListener;

/**
 * This activity creates and returns new {@link NamedLocation}s.
 *
 * @author johan.walles@gmail.com
 */
public class CreateLocationActivity extends Activity {
    /**
     * Lower left latitude of SL's coverage.
     */
    public static final double SL_LOWER_LEFT_LAT = 58.87;

    /**
     * Lower left longitude of SL's coverage.
     */
    public static final double SL_LOWER_LEFT_LON = 17.1754;

    /**
     * Upper right latitude of SL's coverage.
     */
    public static final double SL_UPPER_RIGHT_LAT = 59.8307;

    /**
     * Upper right longitude of SL's coverage.
     */
    public static final double SL_UPPER_RIGHT_LON = 19.1907;

    /**
     * A text field where the user can enter a new address.
     */
    private AutoCompleteTextView addressEditor;

    /**
     * Hints to the user can be displayed here.
     */
    private TextView helpTextView;

    /**
     * Used for converting between addresses and named locations.
     */
    private Geocoder geocoder;

    /**
     * Naming suggestion for the selected location.  Can be null.
     */
    private String name;

    /**
     * This text will be shown as a balloon tip to the user.
     */
    private String balloonText;

    /**
     * Are any completions showing?  In that case the {@link #balloonText}
     * shouldn't be displayed.
     */
    private boolean completionsExist = false;

    /**
     * An info icon for displaying with the {@link #balloonText}.
     */
    private Drawable infoIcon;

    /**
     * Disambiguate between several hits for one address.
     *
     * @param addresses A list of addresses.  May be empty, or null.
     *
     * @return One address from the list, or null if no address was selected.
     */
    private Address selectLocation(final List<Address> addresses)
    {
        if (addresses == null) {
            // No addresses, no selection possible
            return null;
        }
        if (addresses.size() == 0) {
            // Empty list of addresses, no selection possible
            return null;
        }
        if (addresses.size() == 1) {
            // One address, only one selection possible
            return addresses.get(0);
        }

        List<String> descriptionsList = new ArrayList<String>();
        for (Address address : addresses) {
            descriptionsList.add(address.getAddressLine(1));
        }
        String descriptionsArray[] =
            descriptionsList.toArray(new String[addresses.size()]);

        DialogInterface.OnClickListener selectorListener =
            new DialogInterface.OnClickListener()
        {
            public void onClick(DialogInterface dialog, int which) {
                dialog.dismiss();
                if (which == DialogInterface.BUTTON_NEGATIVE) {
                    // User cancel, do nothing
                    return;
                }
                returnLocation(addresses.get(which));
            }
        };
        AlertDialog.Builder builder = new AlertDialog.Builder(this);
        builder.setItems(descriptionsArray, selectorListener);
        builder.setNegativeButton("Ingen av ovanstående", selectorListener);
        builder.setTitle("Välj plats");
        builder.show();

        // We don't know yet
        return null;
    }

    /**
     * Show / don't show the balloon help depending on whether we have any text
     * to show and whether there are any completions that we should stay out of
     * the way from.
     */
    private void updateBalloon() {
        if (completionsExist || balloonText == null) {
            addressEditor.setError(null);
        } else {
            addressEditor.setError(balloonText, infoIcon);
        }
    }

    /**
     * Clean out out-of-bounds results caused by
     * <a href="http://code.google.com/p/android/issues/detail?id=4189">Android
     * bug 4189</a>.
     *
     * @param dirty A possibly dirty list of addresses
     *
     * @return A clean list of addresses
     */
    private List<Address> cleanLocations(Iterable<Address> dirty) {
        List<Address> clean = new LinkedList<Address>();

        for (Address prospect : dirty) {
            if (prospect.getLatitude() < SL_LOWER_LEFT_LAT) {
                continue;
            }
            if (prospect.getLatitude() > SL_UPPER_RIGHT_LAT) {
                continue;
            }
            if (prospect.getLongitude() < SL_LOWER_LEFT_LON) {
                continue;
            }
            if (prospect.getLongitude() > SL_UPPER_RIGHT_LON) {
                continue;
            }

            clean.add(prospect);
        }

        return clean;
    }

    /**
     * Generate a help text when no results have been found for a location
     * description.
     *
     * @param description The location description to generate the help text
     * for.
     *
     * @param isBug4189 True if this is because of
     * "http://code.google.com/p/android/issues/detail?id=4189".  False
     * otherwise.
     *
     * @return A help text saying that no locations were found for the given
     * location description.
     */
    private Spannable getHelpFor(String description, boolean isBug4189) {
        SpannableStringBuilder builder = new SpannableStringBuilder();

        builder.append("Hittade inte <");
        builder.append(description);
        builder.append(">. Testa att lägga till t.ex. \"Stockholm\" efter för att förbättra sökresultatet, eller kontrollera din stavning.");

        if (!isBug4189) {
            return builder;
        }

        final String BUG_URL =
            "http://code.google.com/p/android/issues/detail?id=4189";

        builder.append("\n\n");
        builder.append("Det här beror på ");
        builder.append("ett fel i Androids kartfunktioner");
        builder.append(". ");

        int linkStart = builder.length();
        builder.append("Stjärnmarkera gärna felet");
        int linkEnd = builder.length();

        builder.append(" ifall du vill få det fixat så du får bättre sökresultat i framtiden!");

        // builder.setSpan(new URLSpan(BUG_URL), link1start, link1end, 0);
        builder.setSpan(new URLSpan(BUG_URL), linkStart, linkEnd, 0);

        return builder;
    }

    /**
     * Convert an address string to a location.
     *
     * @param locationDescription The address to convert into a location.
     *
     * @return A location for the given address, or null if the conversion
     * fails.
     */
    private Address toLocation(String locationDescription) {
        // Clear the help text before attempting to search.
        helpTextView.setText("");

        List<Address> locations;
        try {
            // FIXME: On both 1.5 and 1.6 this gets us tons of location outside
            // of the box, but no locations inside the box for "Storgatan 5".
            //
            // Reported here:
            // http://groups.google.com/group/android-developers/browse_thread/thread/2a0f59849e7b3a21
            // http://code.google.com/p/android/issues/detail?id=4189
            // http://groups.google.com/group/android-developers/browse_thread/thread/7db1a9697f311ff
            locations =
                geocoder.getFromLocationName(locationDescription, 30,
                                             SL_LOWER_LEFT_LAT,
                                             SL_LOWER_LEFT_LON,
                                             SL_UPPER_RIGHT_LAT,
                                             SL_UPPER_RIGHT_LON);
        } catch (IOException e) {
            Log.e(SL.TAG, "Address decoding failed for: " + locationDescription, e);
            return null;
        }

        List<Address> cleanLocations = cleanLocations(locations);

        if (cleanLocations.size() == 1) {
            // Return the one location found inside the SL area
            return cleanLocations.get(0);
        }

        if (cleanLocations.size() == 0) {
            Log.w(SL.TAG, "No location found for: " + locationDescription);
            boolean isBug4189 = locations.size() > 0;

            helpTextView.setText(getHelpFor(locationDescription, isBug4189),
                                 BufferType.SPANNABLE);

            return null;
        }

        return selectLocation(cleanLocations);
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        // Set up our Geocoder
        geocoder = LocationUtil.createGeocoder(this);

        // UI code
        setContentView(R.layout.create_location);
        final Button searchAddressButton =
            (Button)findViewById(R.id.searchAddressButton);
        searchAddressButton.setEnabled(false);

        infoIcon =
            getResources().getDrawable(android.R.drawable.ic_dialog_info);
        infoIcon.setBounds(0, 0,
                           infoIcon.getIntrinsicWidth(),
                           infoIcon.getIntrinsicHeight());

        addressEditor = (AutoCompleteTextView)findViewById(R.id.addressEntry);
        LocationSelectionAdapter adapter = new LocationSelectionAdapter(this);
        adapter.setCompletionsExistListener(new CompletionsExistListener() {
            public void completionsExist(boolean completions) {
                CreateLocationActivity.this.completionsExist = completions;
                updateBalloon();
            }
        });
        addressEditor.setAdapter(adapter);
        addressEditor.setThreshold(3);
        addressEditor.setOnEditorActionListener(new OnEditorActionListener() {
            public boolean onEditorAction(TextView v, int actionId, KeyEvent event)
            {
                final int SEARCH = EditorInfo.IME_ACTION_SEARCH;
                if ((actionId & SEARCH) == SEARCH) {
                    if (!searchAddressButton.isEnabled()) {
                        return true;
                    }

                    Address address =
                        toLocation(addressEditor.getText().toString());
                    if (address == null) {
                        // We didn't get any location, do nothing
                        return false;
                    }

                    returnLocation(address);
                    return true;
                }

                return false;
            }
        });
        addressEditor.setOnItemClickListener(new AdapterView.OnItemClickListener() {
            /**
             * The user has chosen one of the suggested auto-completions.
             */
            public void onItemClick(AdapterView<?> parent, View view,
                int position, long id)
            {
                if (id == 0) {
                    // This is just a street name, not a full entry match.  Just
                    // update the text field.
                    LocationSelectionAdapter.Entry entry =
                        (LocationSelectionAdapter.Entry)parent.getItemAtPosition(position);
                    String address = entry.toString();

                    // Strip any trailing ellipsis from the street name
                    if (address.endsWith("...")) {
                        address = address.substring(0, address.length() - 3);
                    }

                    // Add a space character at the end so that the user can
                    // start typing the street number at once.
                    address += " ";

                    // FIXME: Change the virtual keyboard type to numeric

                    addressEditor.setText(address);
                    addressEditor.setSelection(address.length());
                    return;
                }

                // Get a name for this auto-completion
                name = getPersonName(id);

                // Get the address of this auto-completion
                String addressString = getPersonAddress(id);

                // If necessary, disambiguate between several possible locations
                Address address = toLocation(addressString);
                if (address != null) {
                    NamedLocation location = new NamedLocation(
                        LocationUtil.getSingleLineDescription(address),
                        address.getLatitude(),
                        address.getLongitude(),
                        name);
                    location = readOrCreate(location);
                    returnLocation(location);
                }
            }
        });

        // Hint user to include a street number in the address, otherwise the
        // travel suggestions will potentially be crap.
        final Pattern containsDigits = Pattern.compile("^.*[0-9].*$");
        addressEditor.addTextChangedListener(new TextWatcher() {
            public void onTextChanged(CharSequence s, int start, int before, int count)
            {
                // This block intentionally left blank
            }

            public void beforeTextChanged(CharSequence s, int start, int count,
                    int after)
            {
                // This block intentionally left blank
            }

            public void afterTextChanged(Editable s) {
                // Enable search button if we have text
                searchAddressButton.setEnabled(s.length() > 0);

                if (s.length() == 0) {
                    // No input, it should be obvious to the user that she has
                    // to start typing
                    balloonText = null;
                } else if (containsDigits.matcher(s).matches()) {
                    // We have digits, this is probably an address
                    balloonText = null;
                } else {
                    balloonText =
                        "Glöm inte gatunummer, annars kommer du få mycket sämre reseförslag!";
                }
                updateBalloon();
            }
        });

        searchAddressButton.setOnClickListener(new OnClickListener() {
            public void onClick(View v) {
                Address address =
                    toLocation(addressEditor.getText().toString());
                if (address == null) {
                    // We didn't get any location, do nothing
                    return;
                }

                returnLocation(address);
            }
        });

        helpTextView = (TextView)findViewById(R.id.helpTextView);
        helpTextView.setMovementMethod(new LinkMovementMethod());
    }

    /**
     * Fetch a contact's name from the contacts database.
     *
     * @param id The database ID of the contact we're looking for.
     *
     * @return The name of the contact, or the name of the contact's
     * organization.
     */
    private String getPersonName(long id) {
        // Look up the contact's own name
        Uri uri = ContentUris.withAppendedId(People.CONTENT_URI, id);
        Cursor cursor =
            getContentResolver().query(uri,
                new String[] { PeopleColumns.NAME },
                null, null, null);
        try {
            for (cursor.moveToFirst();
                cursor.getPosition() < cursor.getCount();
                cursor.moveToNext())
            {
                String returnMe = cursor.getString(0);
                if (returnMe != null && returnMe.length() > 0) {
                    return returnMe;
                }
            }
        } finally {
            cursor.close();
        }

        // Look up the contact's organization
        cursor = getContentResolver().query(Contacts.Organizations.CONTENT_URI,
            new String[] { Contacts.OrganizationColumns.COMPANY },
            Contacts.OrganizationColumns.PERSON_ID + "=" + id,
            null,
            null);
        try {
            for (cursor.moveToFirst();
                cursor.getPosition() < cursor.getCount();
                cursor.moveToNext())
            {
                String returnMe = cursor.getString(0);
                if (returnMe != null && returnMe.length() > 0) {
                    return returnMe;
                }
            }
        } finally {
            cursor.close();
        }

        return null;
    }

    /**
     * Fetch a contact's address from the contacts database.
     *
     * @param id The database ID of the contact we're looking for.
     *
     * @return The contact's address.
     */
    private String getPersonAddress(long id) {
        Uri uri = Contacts.ContactMethods.CONTENT_URI;

        // Ordering the results by type as below makes us use the home address
        // primarily and the work address secondarily.
        Cursor cursor =
            getContentResolver().query(uri,
                new String[] { ContactMethodsColumns.DATA },
                String.format("%s=%d and %s<>%d and %s=%d",
                    Contacts.People.Extensions.PERSON_ID, id,
                    Contacts.ContactMethodsColumns.TYPE,
                    Contacts.ContactMethodsColumns.TYPE_CUSTOM,
                    Contacts.ContactMethodsColumns.KIND, Contacts.KIND_POSTAL),
                null,
                Contacts.ContactMethodsColumns.TYPE + " asc");

        try {
            if (cursor.getCount() >= 0) {
                cursor.moveToFirst();
                return cursor.getString(0);
            }
        } finally {
            cursor.close();
        }

        // No address found
        return null;
    }

    /**
     * Return a named location to the caller of this {@link Activity}.
     *
     * @param address The location to return.
     */
    private void returnLocation(Address address) {
        if (name == null) {
            // Name this location properly before returning it.  The actual
            // returning is done in onActivityResult().
            NameLocationActivity.launch(this, address);
            return;
        }

        NamedLocation location = new NamedLocation(
            LocationUtil.getSingleLineDescription(address),
            address.getLatitude(),
            address.getLongitude(),
            name);
        location = readOrCreate(location);
        returnLocation(location);
    }

    /**
     * Launch a location creation activity.  Implement
     * {@link Activity#onActivityResult(int, int, android.content.Intent)} to
     * get the reply.
     *
     * @param launcher The activity we're being launched from.
     *
     * @param spinnerId For which spinner should the new location be created?
     */
    public static void launch(Activity launcher, int spinnerId) {
        Intent intent =
            new Intent(launcher, CreateLocationActivity.class);
        intent.putExtra(RequestKeys.LOCATION_SELECTOR_ID, spinnerId);
        launcher.startActivityForResult(intent, RequestCodes.CREATE_LOCATION);
    }

    @Override
    protected void onActivityResult(int requestCode, int resultCode, Intent data)
    {
        if (requestCode != RequestCodes.NAME_LOCATION) {
            Log.w(SL.TAG, "CreateLocationActivity got unknown result request code "
                + requestCode);
            return;
        }

        if (resultCode != RESULT_OK) {
            Log.w(SL.TAG, "CreateLocationActivity ignoring not-OK result code "
                + resultCode);
            return;
        }

        NamedLocation namedLocation =
            NameLocationActivity.getNamedLocation(data);

        namedLocation = readOrCreate(namedLocation);
        returnLocation(namedLocation);
    }

    /**
     * Finish this activity and return the new location to our caller.
     *
     * @param namedLocation The result to return.
     */
    private void returnLocation(NamedLocation namedLocation) {
        setResult(RESULT_OK,
            SL.createLocationResult(namedLocation,
                getIntent().getIntExtra(RequestKeys.LOCATION_SELECTOR_ID, -1)));
        finish();
    }

    /**
     * Store a location in the database if it doesn't already exist.  If it
     * already exists, read the existing location.
     *
     * @param namedLocation The location to store.
     *
     * @return A named location that exists in the database.
     */
    private NamedLocation readOrCreate(NamedLocation namedLocation) {
        // Look for an existing location with the same name
        LocationsDbAdapter db = new LocationsDbAdapter(this, SL.DB_NAME);
        ScoredLocation existing = db.fetchByName(namedLocation.getAlias());
        if (existing != null) {
            return existing.getLocation();
        }

        // Update the database with the new location
        Set<ScoredLocation> locations = db.fetchLocations();
        locations.add(new ScoredLocation(namedLocation, 0, 0));
        db.storeLocations(locations);
        return namedLocation;
    }
}
