/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.launchpad.sl;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import android.content.Context;
import android.database.Cursor;
import android.provider.BaseColumns;
import android.provider.Contacts;
import android.provider.Contacts.People;
import android.provider.Contacts.PeopleColumns;
import android.util.Log;
import android.widget.ArrayAdapter;
import android.widget.AutoCompleteTextView;
import android.widget.Filter;

/**
 * Provides auto completion when adding new locations.
 *
 * @author johan.walles@gmail.com
 */
public class LocationSelectionAdapter
extends ArrayAdapter<LocationSelectionAdapter.Entry>
{
    /**
     * This listener should be notified about whether any completions exist.
     */
    private CompletionsExistListener completionsExistListener;

    /**
     * An auto completion entry.
     */
    protected static class Entry implements Comparable<Entry> {
        /**
         * The ID of this person in the contacts database.
         */
        private long id;

        /**
         * A description of this entry in the contacts database.
         */
        private String description;

        /**
         * Constructor.
         *
         * @param id The ID of this person in the contacts database.
         *
         * @param description A description of this contacts database entry.
         */
        public Entry(long id, String description) {
            this.id = id;
            this.description = description;
        }

        /**
         * Get the ID of this contacts database entry.
         *
         * @return The ID of this contacts database entry.
         */
        public long getId() {
            return id;
        }

        @Override
        public String toString() {
            return description;
        }

        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result
                + ((description == null) ? 0 : description.hashCode());
            result = prime * result + (int) (id ^ (id >>> 32));
            return result;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj) {
                return true;
            }
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            Entry other = (Entry) obj;
            if (description == null) {
                if (other.description != null) {
                    return false;
                }
            } else if (!description.equals(other.description)) {
                return false;
            }
            if (id != other.id) {
                return false;
            }
            return true;
        }

        public int compareTo(Entry another) {
            // Put pure addresses before people
            if (id == 0 && another.id != 0) {
                return -1;
            }
            if (another.id == 0 && id != 0) {
                return 1;
            }

            return description.compareTo(another.description);
        }
    }

    /**
     * Information collected from the contacts database.
     */
    protected static class Contact {
        /**
         * This contact's name.
         */
        public String name;

        /**
         * This contact's organization.
         */
        public String organization;

        /**
         * This contact's postal address.
         */
        public String postalAddress;

        @Override
        public String toString() {
            if (name != null && name.length() > 0) {
                return name;
            }

            if (organization != null && organization.length() > 0) {
                return organization;
            }

            if (postalAddress != null && postalAddress.length() > 0) {
                return postalAddress;
            }

            return null;
        }
    }

    /**
     * Receiver of {@link LocationSelectionAdapter} notifications about whether
     * or not any completions exist for the current text box contents.
     */
    public static interface CompletionsExistListener {
        /**
         * Called with notification about whether or not any completions exist
         * for the current text box contents.
         *
         * @param completionsExist True if there are completions.  False
         * otherwise.
         */
        public void completionsExist(boolean completionsExist);
    }

    /**
     * Construct a new location selection adapter.
     *
     * @param context The context in which our {@link AutoCompleteTextView} is
     * running.
     */
    public LocationSelectionAdapter(Context context) {
        super(context, android.R.layout.simple_dropdown_item_1line);
    }

    /**
     * Extracts a street name from the beginning of a string.
     */
    private final static Pattern street =
        Pattern.compile("([^\n0-9]+) +[0-9]+");

    /**
     * Extract the street name from an address.
     * <p>
     * Has default protection for testing purposes.
     *
     * @param address The address string.
     *
     * @return The street name, or null if no street name could be extracted
     * from the address.
     */
    static String streetify(String address) {
        Matcher matcher =  street.matcher(address);
        if (matcher.find()) {
            String streetName = matcher.group(1).trim();
            if (streetName.length() > 0) {
                return streetName;
            }
        }

        return null;
    }

    @Override
    public Filter getFilter() {
        return new Filter() {
            @SuppressWarnings("unchecked")
            @Override
            protected void publishResults(CharSequence constraint,
                FilterResults results)
            {
                clear();

                if (results != null && results.values != null) {
                    for (Entry value : (TreeSet<Entry>)results.values) {
                        add(value);
                    }
                }
                notifyDataSetChanged();

                if (completionsExistListener != null) {
                    completionsExistListener.completionsExist(!isEmpty());
                }
            }

            @Override
            protected FilterResults performFiltering(CharSequence constraint) {
                FilterResults result = null;
                try {
                    result = filteringHelper(constraint);
                } catch (Exception e) {
                    Log.e(SL.TAG,
                        "Completion filtering failed for constraint: <"
                        + constraint
                        + ">",
                        e);
                }
                return result;
            }

            private FilterResults filteringHelper(CharSequence constraint) {
                if (constraint == null) {
                    return null;
                }

                Map<Long, Contact> contacts = new HashMap<Long, Contact>();

                // Collect addresses
                Cursor cursor = getContext().getContentResolver().query(
                    Contacts.ContactMethods.CONTENT_URI,
                    new String[] { Contacts.ContactMethodsColumns.DATA,
                        Contacts.People.Extensions.PERSON_ID },
                    String.format("%s<>%d and %s=%d",
                        Contacts.ContactMethodsColumns.TYPE,
                        Contacts.ContactMethodsColumns.TYPE_CUSTOM,
                        Contacts.ContactMethodsColumns.KIND,
                        Contacts.KIND_POSTAL), null, null);
                try {
                    for (cursor.moveToFirst(); cursor.getPosition() < cursor
                        .getCount(); cursor.moveToNext())
                    {
                        String address = cursor.getString(0);
                        long id = cursor.getLong(1);
                        Contact contact = contacts.get(id);
                        if (contact == null) {
                            contact = new Contact();
                        }
                        contact.postalAddress = address;
                        contacts.put(id, contact);
                    }
                } finally {
                    cursor.close();
                }

                // Read all names
                cursor = getContext().getContentResolver().query(
                    People.CONTENT_URI,
                    new String[] { PeopleColumns.NAME, BaseColumns._ID, },
                    null, null, null);
                try {
                    for (cursor.moveToFirst(); cursor.getPosition() < cursor
                        .getCount(); cursor.moveToNext())
                    {
                        long id = cursor.getLong(1);
                        Contact contact = contacts.get(id);
                        if (contact == null) {
                            // This contact has no address, so we can't use it
                            continue;
                        }
                        String name = cursor.getString(0);
                        contact.name = name;
                        contacts.put(id, contact);
                    }
                } finally {
                    cursor.close();
                }

                // Read all organizations
                cursor = getContext().getContentResolver().query(
                    Contacts.Organizations.CONTENT_URI,
                    new String[] { Contacts.OrganizationColumns.COMPANY,
                        Contacts.OrganizationColumns.PERSON_ID },
                    null, null, null);
                try {
                    for (cursor.moveToFirst(); cursor.getPosition() < cursor
                        .getCount(); cursor.moveToNext())
                    {
                        long id = cursor.getLong(1);
                        Contact contact = contacts.get(id);
                        if (contact == null) {
                            // This contact has no address, so we can't use it
                            continue;
                        }
                        String organization = cursor.getString(0);
                        contact.organization = organization;
                        contacts.put(id, contact);
                    }
                } finally {
                    cursor.close();
                }

                // Collect matches
                SortedSet<Entry> entries = new TreeSet<Entry>();
                Pattern pattern = Pattern.compile(constraint.toString(),
                    Pattern.LITERAL | Pattern.CASE_INSENSITIVE);
                Set<String> matchingAddresses = new HashSet<String>();
                for (Map.Entry<Long, Contact> contact : contacts.entrySet()) {
                    long id = contact.getKey();
                    boolean matches = false;

                    // Does the name match?
                    String name = contact.getValue().name;
                    if (name != null && name.length() > 0) {
                        if (pattern.matcher(name).find()) {
                            matches = true;
                        }
                    } else {
                        // Name not available, does the organization match?
                        String organization = contact.getValue().organization;
                        if (organization != null && organization.length() > 0) {
                            if (pattern.matcher(organization).find()) {
                                matches = true;
                            }
                        }
                    }

                    // Does the address match?
                    String address = contact.getValue().postalAddress;
                    if (address != null && address.length() > 0) {
                        if (pattern.matcher(address).find()) {
                            String streetName = streetify(address);
                            if (streetName != null) {
                                // The address field contains a proper address
                                matches = true;
                            }

                            if (streetName != null
                                && constraint.length() > streetName.length())
                            {
                                // The user has typed more than this street
                                // name, possibly a number, so she isn't helped
                                // by having this completion available.
                                streetName = null;
                            }
                            if (streetName != null) {
                                matchingAddresses.add(streetName + "...");
                            }
                        }
                    }

                    if (matches) {
                        String description = contact.getValue().toString();
                        if (description == null) {
                            // Nondescript entry, displaying it won't help
                            continue;
                        }

                        Entry entry = new Entry(id, description);
                        entries.add(entry);
                    }
                }
                for (String address : matchingAddresses) {
                    entries.add(new Entry(0, address));
                }

                FilterResults results = new FilterResults();
                results.values = entries;
                results.count = entries.size();
                return results;
            }
        };
    }

    @Override
    public long getItemId(int position) {
        return getItem(position).getId();
    }

    /**
     * Register a listener for whether or not we have any completions.
     *
     * @param listener The listener that will be notified when completion
     * existence changes.
     */
    public void setCompletionsExistListener(CompletionsExistListener listener) {
        this.completionsExistListener = listener;
    }
}
