/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl;

import net.launchpad.sl.location.LocationNameWatcher;
import net.launchpad.sl.location.LocationUtil;
import net.launchpad.sl.location.LocationsDbAdapter;
import net.launchpad.sl.location.NamedLocation;
import android.app.Activity;
import android.content.Intent;
import android.location.Address;
import android.os.Bundle;
import android.text.TextWatcher;
import android.view.KeyEvent;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.inputmethod.EditorInfo;
import android.widget.Button;
import android.widget.EditText;
import android.widget.TextView;
import android.widget.TextView.OnEditorActionListener;

/**
 * Name a location.
 *
 * @author johan.walles@gmail.com
 */
public class NameLocationActivity extends Activity {
    /**
     * Show an UI for naming a given location.
     *
     * @param launcher The activity we're being launched from.
     *
     * @param location The location to name.
     */
    public static void launch(Activity launcher, Address location) {
        Intent intent = new Intent(launcher, NameLocationActivity.class);
        intent.putExtra(RequestKeys.LOCATION, location);
        launcher.startActivityForResult(intent, RequestCodes.NAME_LOCATION);
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        // UI code
        setContentView(R.layout.name_location);

        // Describe the location
        final Address location =
            getIntent().getParcelableExtra(RequestKeys.LOCATION);
        TextView locationDescription =
            (TextView)findViewById(R.id.LocationDescriptionText);
        locationDescription.setText(
            LocationUtil.getMultiLineDescription(location));

        // Set up the edit box
        final EditText nameEditor = (EditText)findViewById(R.id.NameEditor);
        nameEditor.setText(LocationUtil.getSingleLineDescription(location));
        final Button nameOkButton = (Button)findViewById(R.id.NameOkButton);
        nameEditor.setOnEditorActionListener(new OnEditorActionListener() {
            public boolean onEditorAction(TextView v, int actionId, KeyEvent event)
            {
                if (!nameOkButton.isEnabled()) {
                    // Current name is not valid, ignore the DONE button
                    return false;
                }

                final int DONE = EditorInfo.IME_ACTION_DONE;
                if ((actionId & DONE) == DONE) {
                    returnLocation(nameEditor.getText().toString(), location);
                    return true;
                }

                return false;
            }
        });

        // Set up the OK button
        nameOkButton.setOnClickListener(new OnClickListener() {
            public void onClick(View v) {
                returnLocation(nameEditor.getText().toString(), location);
            }
        });

        TextWatcher textWatcher = new LocationNameWatcher(
                new LocationsDbAdapter(this, SL.DB_NAME),
                nameEditor,
                nameOkButton);
        nameEditor.addTextChangedListener(textWatcher);
    }

    /**
     * Return a reply to the activity that called
     * {@link #launch(Activity, Address)} on us.
     *
     * @param location A bicycle.
     * @param name The location name.
     */
    private void returnLocation(String name, Address location)
    {
        NamedLocation namedLocation =
            new NamedLocation(
                LocationUtil.getSingleLineDescription(location).toString(),
                location.getLatitude(),
                location.getLongitude(),
                name);

        Intent result = new Intent();
        result.putExtra(RequestKeys.LOCATION, namedLocation);
        setResult(RESULT_OK, result);
        finish();
    }

    /**
     * Extract a named location from the response to a
     * {@link #launch(Activity, Address)}.
     *
     * @param response The response to parse.
     *
     * @return The named location in the response.
     */
    public static NamedLocation getNamedLocation(Intent response) {
        NamedLocation location =
            response.getParcelableExtra(RequestKeys.LOCATION);
        if (location == null) {
            throw new IllegalArgumentException("Key missing in response: "
                                               + RequestKeys.LOCATION);
        }
        return location;
    }
}
