/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl;

import java.util.Calendar;
import java.util.GregorianCalendar;

import net.launchpad.sl.location.NamedLocation;

import android.location.Location;
import android.net.Uri;
import android.os.Parcel;
import android.os.Parcelable;

/**
 * This class represents a query that can be sent to
 * http://reseplanerare.sl.se/bin/query.exe.
 *
 * @author johan.walles@gmail.com
 */
public class Query implements Parcelable {
    /**
     * We can't parcel null times, so we'll use this as a marker for "now".
     */
    private final static long PARCEL_TIME_NOW = -1L;

    /**
     * Where do we start our journey?
     */
    private NamedLocation from;

    /**
     * Where are we going?
     */
    private NamedLocation to;

    /**
     * When are we going?
     *
     * @see #whenIsDeparture
     */
    private Calendar when;

    /**
     * True if {@link #when} is the departure time.  False if it is the
     * arrival time.
     */
    private boolean whenIsDeparture = true;

    /**
     * Create a new query.
     *
     * @param from Where does the journey start?
     *
     * @param to Where are we going?
     */
    public Query(NamedLocation from, NamedLocation to)
    {
        this.from = from;
        this.to = to;
    }

    /**
     * Convert this query into a sl.se URL.
     *
     * @param currentLocation Used if either to or from represents the current
     * location.
     *
     * @param tableDisplay True to display the search result in a table.  False
     * to display it as text.
     *
     * @return An URL that should answer this query.
     */
    public String toUriString(Location currentLocation, boolean tableDisplay) {
        Uri.Builder uriBuilder = new Uri.Builder();
        uriBuilder.scheme("http");
        uriBuilder.authority("reseplanerare.sl.se");
        uriBuilder.appendEncodedPath("bin/query.exe/sn");
        uriBuilder.appendQueryParameter("queryPageDisplayed", "yes");
        uriBuilder.appendQueryParameter("ignoreTypeCheck", "yes");

        uriBuilder.appendQueryParameter("REQ0JourneyStopsS0A", "255");
        String description = from.getAddressOrStop();
        if (description == null) {
            description = from.getAlias();
        }
        uriBuilder.appendQueryParameter("REQ0JourneyStopsS0G", description);
        uriBuilder.appendQueryParameter("REQ0JourneyStopsSID",
                                        getLocationString(from, currentLocation));

        uriBuilder.appendQueryParameter("REQ0JourneyStopsZ0A", "255");
        description = to.getAddressOrStop();
        if (description == null) {
            description = to.getAlias();
        }
        uriBuilder.appendQueryParameter("REQ0JourneyStopsZ0G", description);
        uriBuilder.appendQueryParameter("REQ0JourneyStopsZID",
                                        getLocationString(to, currentLocation));

        String departureOrArrival;
        if (whenIsDeparture) {
            // Timestamp below is departure time
            departureOrArrival = "1";
        } else {
            // Timestamp below is arrival time
            departureOrArrival = "0";
        }
        uriBuilder.appendQueryParameter("REQ0HafasSearchForw", departureOrArrival);

        uriBuilder.appendQueryParameter("REQ0JourneyDate", getDateString());
        uriBuilder.appendQueryParameter("REQ0JourneyTime", getTimeString());

        // Set REQ0HafasUnsharpSearch to 1 to accept walking a bit at either
        // end of your trip if it would make your trip faster.
        // builder.appendQueryParameter("REQ0HafasUnsharpSearch", "1");

        uriBuilder.appendQueryParameter("existUnsharpSearch", "yes");
        uriBuilder.appendQueryParameter("SetGlobalOptionGO_displayDataInTable",
            tableDisplay ? "yes" : "no");

        uriBuilder.appendQueryParameter("start", "Sök resa");
        uriBuilder.fragment("FormMainHeader");

        String uriString = uriBuilder.build().toString();

        // Replace all system provided (UTF8) encodings with what SL wants
        // (Unicode).  Do this for ÅÄÖåäöé.
        // Reference: http://www.jeppesn.dk/utf-8.html
        uriString = uriString.replace("%C3%85", "%C5"); // Å
        uriString = uriString.replace("%C3%84", "%C4"); // Ä
        uriString = uriString.replace("%C3%96", "%D6"); // Ö

        uriString = uriString.replace("%C3%A5", "%E5"); // å
        uriString = uriString.replace("%C3%A4", "%E4"); // ä
        uriString = uriString.replace("%C3%B6", "%F6"); // ö
        uriString = uriString.replace("%C3%A9", "%E9"); // é

        return uriString;
    }

    /**
     * Get a query string describing a location in a form that can be parsed by
     * sl.se.
     * <p>
     * This method has default protection for testing purposes.
     *
     * @param namedLocation The location to describe.
     *
     * @param currentLocation This location is used if either to or from
     * represents the phone's current location.
     *
     * @return A query string describing a location.
     */
    String getLocationString(NamedLocation namedLocation,
        Location currentLocation)
    {
        StringBuilder builder = new StringBuilder(99);

        builder.append("A=16@O=");
        String description = namedLocation.getAddressOrStop();
        if (description == null) {
            description = namedLocation.getAlias();
        }
        builder.append(description);

        Location location;
        if (namedLocation.isHere()) {
            location = currentLocation;
        } else {
            location = namedLocation.getCoordinate();
        }
        builder.append("@X=").append((long)(location.getLongitude() * 1000000));
        builder.append("@Y=").append((long)(location.getLatitude() * 1000000));
        builder.append("@");

        return builder.toString();
    }

    /**
     * Convert {@link #when} into a time string for sl.se, like "17:50".
     * <p>
     * This method has default protection for testing purposes.
     *
     * @return A string representation of when, containing hour and minute.
     */
    String getTimeString() {
        Calendar time = when;
        if (time == null) {
            time = new GregorianCalendar();
        }

        int hour = time.get(Calendar.HOUR_OF_DAY);
        int minute = time.get(Calendar.MINUTE);

        return String.format("%02d:%02d", hour, minute);
    }

    /**
     * Convert {@link #when} into a date string for sl.se, like "20.09.09".
     * <p>
     * This method has default protection for testing purposes.
     *
     * @return A string representation of when, containing year, month and day.
     */
    String getDateString() {
        Calendar time = when;
        if (time == null) {
            time = new GregorianCalendar();
        }

        int year = time.get(Calendar.YEAR);
        if (year > 2000) {
            // sl.se wants a two digit year
            year -= 2000;
        }

        // API months are zero based, sl.se months are one based
        int month = time.get(Calendar.MONTH) + 1;

        int day = time.get(Calendar.DAY_OF_MONTH);

        return String.format("%02d.%02d.%02d", day, month, year);
    }

    /**
     * Set when we want to go.  Null means "go now".
     * @param when When do we want to go?  null means now.
     * @param whenIsDeparture True if when is the departure time.  False if when
     * is the arrival time.
     */
    public void setWhen(Calendar when, boolean whenIsDeparture) {
        this.when = when;
        this.whenIsDeparture = whenIsDeparture;

        if (when == null) {
            // We can never arrive now, so if when is now it means when is the
            // departure time.
            this.whenIsDeparture = true;
        }
    }

    @Override
    public String toString() {
        String whenString = "go now";
        if (when != null) {
            if (whenIsDeparture) {
                whenString = "go at " + when.toString();
            } else {
                whenString = "be there at " + when.toString();
            }
        }

        return String.format("From %s to %s, %s",
                             from.getAlias(), to.getAlias(), whenString);
    }

    public int describeContents() {
        return 0;
    }

    public void writeToParcel(Parcel dest, int flags) {
        dest.writeParcelable(from, 0);
        dest.writeParcelable(to, 0);
        dest.writeLong(when == null ? PARCEL_TIME_NOW : when.getTimeInMillis());
        dest.writeBooleanArray(new boolean[] { whenIsDeparture});
    }

    /**
     * Part of the protocol for the {@link Parcelable} interface.
     */
    public static final Parcelable.Creator<Query> CREATOR =
        new Parcelable.Creator<Query>()
    {
        public Query createFromParcel(Parcel source) {
            ClassLoader myClassLoader = getClass().getClassLoader();
            NamedLocation from =
                (NamedLocation)source.readParcelable(myClassLoader);
            NamedLocation to =
                (NamedLocation)source.readParcelable(myClassLoader);

            long time = source.readLong();
            Calendar when = null;
            if (time != PARCEL_TIME_NOW) {
                when = new GregorianCalendar();
                when.setTimeInMillis(time);
            }

            boolean booleanArray[] = new boolean[1];
            source.readBooleanArray(booleanArray);
            boolean whenIsDeparture = booleanArray[0];

            Query query = new Query(from, to);
            query.setWhen(when, whenIsDeparture);

            return query;
        }

        public Query[] newArray(int size) {
            return new Query[size];
        }
    };

    @Override
    public boolean equals(Object o) {
        Query theOtherGuy = (Query)o;

        if (!theOtherGuy.from.equals(from)) {
            return false;
        }

        if (!theOtherGuy.to.equals(to)) {
            return false;
        }

        // Compare times
        boolean imNow = (when == null);
        boolean theOtherGuyIsNow = (theOtherGuy.when == null);
        if (imNow && theOtherGuyIsNow) {
            return true;
        }
        if (imNow != theOtherGuyIsNow) {
            return false;
        }
        if (!when.equals(theOtherGuy.when)) {
            return false;
        }
        if (whenIsDeparture != theOtherGuy.whenIsDeparture) {
            return false;
        }

        return true;
    }

    @Override
    public int hashCode() {
        return from.hashCode() ^ to.hashCode() ^ when.hashCode();
    }

    /**
     * Where are we coming from?
     *
     * @return Where we are coming from.
     */
    public NamedLocation getFrom() {
        return from;
    }

    /**
     * Where are we going to?
     *
     * @return To where we are going.
     */
    public NamedLocation getTo() {
        return to;
    }
}
