/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl;

import java.util.Calendar;
import java.util.HashSet;
import java.util.Set;

import net.launchpad.sl.location.LocationsDbAdapter;
import net.launchpad.sl.location.NamedLocation;
import net.launchpad.sl.location.PhoneLocator;
import net.launchpad.sl.location.ScoredLocation;
import net.launchpad.sl.resultsdisplay.DisplayResult;

import android.app.Activity;
import android.app.DatePickerDialog;
import android.app.ProgressDialog;
import android.app.TimePickerDialog;
import android.content.Intent;
import android.graphics.Color;
import android.location.LocationManager;
import android.os.AsyncTask;
import android.os.Bundle;
import android.util.Log;
import android.view.View;
import android.view.View.OnClickListener;
import android.widget.AdapterView;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.DatePicker;
import android.widget.Spinner;
import android.widget.TextView;
import android.widget.TimePicker;

/**
 * Query http://reseplanerare.sl.se for how to get from here to there.
 *
 * @author johan.walles@gmail.com
 */
public class SL extends Activity {
    /**
     * Tag to pass to {@link Log} calls.
     */
    public final static String TAG = "SL";

    /**
     * The name of our locations database.
     */
    public final static String DB_NAME = "locations";

    /**
     * {@link #whenSpinner} position meaning the user wants to go now.
     */
    private static final int WHEN_SPINNER_NOW = 0;

    /**
     * {@link #whenSpinner} position meaning the user wants to arrive
     * {@link #when}.
     */
    private static final int WHEN_SPINNER_ARRIVAL_TIME = 1;

    /**
     * {@link #whenSpinner} position meaning the user wants to leave
     * {@link #when}.
     */
    private static final int WHEN_SPINNER_DEPARTURE_TIME = 2;

    /**
     * Spinner for selecting starting location.
     */
    private Button fromButton;

    /**
     * Spinner for selecting goal location.
     */
    private Button toButton;

    /**
     * Where do we want to start our journey?
     */
    private NamedLocation fromLocation;

    /**
     * Where does our journey end?
     */
    private NamedLocation toLocation;

    /**
     * The button used for launching a search.
     */
    private Button searchButton;

    /**
     * A text displayed above the search button.
     */
    private TextView searchButtonHint;

    /**
     * The spinner where the user can select when to go.
     */
    private Spinner whenSpinner;

    /**
     * The button displaying the time to go.
     */
    private Button timeButton;

    /**
     * The button describing the date to go.
     */
    private Button dateButton;

    /**
     * A date specification to complement the when spinner and its friends.
     *
     * @see #getWhen()
     */
    private Calendar when;

    /**
     * Our connection to persistent storage.
     */
    private LocationsDbAdapter db;

    /**
     * Used for warming up the positioning system before the user hit "search".
     */
    private PhoneLocator phoneLocator;

    /**
     * Create a result that can be returned to SL using
     * {@link #setResult(int, Intent)}.
     *
     * @param location The location to return
     *
     * @param locationSelectorId The location selector we're returning a
     * location for
     *
     * @return A result that can be returned to {@link SL} using
     * {@link #setResult(int, Intent)}.
     */
    public static Intent createLocationResult(NamedLocation location, int locationSelectorId) {
        Intent result = new Intent();
        result.putExtra(RequestKeys.LOCATION, location);
        result.putExtra(RequestKeys.LOCATION_SELECTOR_ID, locationSelectorId);
        return result;
    }

    /**
     * Return a non-null time greater than or equal to the current time.
     *
     * @return a non-null time greater than or equal to the current time.
     */
    private Calendar getWhen() {
        Calendar now = Calendar.getInstance();
        if (when == null) {
            // Set time to five minutes from now, but zero out the seconds
            when = now;
            when.add(Calendar.MINUTE, 5);
            when.set(Calendar.SECOND, 0);
        }

        if (!when.after(now)) {
            when = now;
            if (whenSpinner.getSelectedItemPosition() != WHEN_SPINNER_NOW) {
                whenSpinner.setSelection(WHEN_SPINNER_NOW);
            }
        }

        return when;
    }

    /**
     * Make sure the when selection stuff is up to date.
     */
    private void updateWhen() {
        timeButton.setText(String.format("%02d:%02d",
            getWhen().get(Calendar.HOUR_OF_DAY),
            getWhen().get(Calendar.MINUTE)));
        dateButton.setText(CalendarUtil.formatDate(getWhen()));

        int buttonsVisibility = View.VISIBLE;
        if (whenSpinner.getSelectedItemPosition() == WHEN_SPINNER_NOW) {
            buttonsVisibility = View.GONE;
        }
        timeButton.setVisibility(buttonsVisibility);
        dateButton.setVisibility(buttonsVisibility);
    }

    /**
     * Make sure the database is up to date, then call
     * {@link #finishOnCreate()}.
     */
    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        setTitle(R.string.app_long_name);

        // Set up our database connection
        db = new LocationsDbAdapter(this, DB_NAME);
        if (!db.needsUpgrade()) {
            // No database upgrade needed, keep going
            finishOnCreate();
            return;
        }

        // Upgrade the database, then finish onCreate
        new AsyncTask<Object, Integer, Object>() {
            private ProgressDialog progress;

            @Override
            protected void onPostExecute(Object result) {
                if (progress.isShowing()) {
                    progress.dismiss();
                    progress = null;
                }
                finishOnCreate();
                onResume();
            }

            @Override
            protected void onPreExecute() {
                progress = new ProgressDialog(SL.this);
                progress.setTitle("Uppdaterar databasen");
                progress.setMessage("Uppdaterar platsdatabasen...");

                progress.setIndeterminate(false);
                progress.setProgressStyle(ProgressDialog.STYLE_HORIZONTAL);
                progress.setProgress(0);
                progress.setMax(100);

                progress.setCancelable(false);

                progress.show();
            }

            @Override
            protected void onProgressUpdate(Integer... percentages) {
                progress.setProgress(percentages[0]);
            }

            @Override
            protected Object doInBackground(Object... params) {
                db.upgradeDb(new LocationsDbAdapter.Progress() {
                    public void onProgress(int percent) {
                        publishProgress(percent);
                    }
                });

                return null;
            }
        }.execute();
    }

    /**
     * Finish {@link #onCreate(Bundle)} after any database upgrades are done.
     */
    private void finishOnCreate() {
        // Get ourselves a location manager
        LocationManager locationManager =
            (LocationManager)getSystemService(LOCATION_SERVICE);
        phoneLocator = new PhoneLocator(locationManager);

        // Set up the search button
        setContentView(R.layout.main);
        searchButtonHint = (TextView)findViewById(R.id.searchButtonHint);
        searchButton = (Button)findViewById(R.id.searchTripButton);
        searchButton.setOnClickListener(new View.OnClickListener() {
            public void onClick(View v) {
                db.increaseScores(fromLocation, toLocation);
                Query query = new Query(fromLocation, toLocation);

                // Set trip time
                switch (whenSpinner.getSelectedItemPosition()) {
                case WHEN_SPINNER_ARRIVAL_TIME:
                    query.setWhen(getWhen(), false);
                    break;
                case WHEN_SPINNER_DEPARTURE_TIME:
                    query.setWhen(getWhen(), true);
                    break;
                }

                DisplayResult.answerQuery(SL.this, query);
            }
        });

        Button quitButton = (Button)findViewById(R.id.quitButton);
        quitButton.setOnClickListener(new View.OnClickListener() {
            public void onClick(View v) {
                finish();
            }
        });

        Button returnTripButton = (Button)findViewById(R.id.returnTripButton);
        returnTripButton.setOnClickListener(new View.OnClickListener() {
            public void onClick(View v) {
                NamedLocation oldFromLocation = fromLocation;
                fromLocation = toLocation;
                toLocation = oldFromLocation;
                updateEndpoints();
            }
        });

        //  Set up the location selectors
        fromButton = (Button)findViewById(R.id.fromButton);
        fromButton.setOnClickListener(new OnClickListener() {
            public void onClick(View v) {
                SelectLocationActivity.launch(SL.this, fromButton.getId(), fromLocation);
            }
        });
        toButton = (Button)findViewById(R.id.toButton);
        toButton.setOnClickListener(new OnClickListener() {
            public void onClick(View v) {
                SelectLocationActivity.launch(SL.this, toButton.getId(), toLocation);
            }
        });

        whenSpinner = (Spinner)findViewById(R.id.whenSpinner);
        whenSpinner.setAdapter(new ArrayAdapter<String>(this,
            android.R.layout.simple_spinner_item,
            new String[] {
            "Nu",
            "Framme senast...",
            "Åka tidigast..."
        }));
        whenSpinner.setOnItemSelectedListener(new AdapterView.OnItemSelectedListener() {
            public void onItemSelected(AdapterView<?> parent, View view,
                int position, long id)
            {
                if (whenSpinner.getSelectedItemPosition() != WHEN_SPINNER_NOW) {
                    // We want a time in the future
                    Calendar in5minutes = Calendar.getInstance();
                    in5minutes.add(Calendar.MINUTE, 5);
                    in5minutes.set(Calendar.SECOND, 0);
                    if (when == null || !when.after(in5minutes)) {
                        // Set time to five minutes from now, but zero out the seconds
                        when = in5minutes;
                    }
                }

                updateWhen();
            }

            public void onNothingSelected(AdapterView<?> parent) {
                // This method intentionally left blank
            }
        });

        timeButton = (Button)findViewById(R.id.whenTime);
        timeButton.setVisibility(View.GONE);
        timeButton.setOnClickListener(new OnClickListener() {
            /**
             * Demands time from the user.
             */
            private TimePickerDialog timePicker;

            public void onClick(View v) {
                timePicker =
                    new TimePickerDialog(SL.this,
                        new TimePickerDialog.OnTimeSetListener() {
                        public void onTimeSet(TimePicker view, int hourOfDay, int minute) {
                            getWhen().set(Calendar.HOUR_OF_DAY, hourOfDay);
                            getWhen().set(Calendar.MINUTE, minute);
                            timePicker.dismiss();
                            timePicker = null;
                            updateWhen();
                        }
                    },
                    getWhen().get(Calendar.HOUR_OF_DAY),
                    getWhen().get(Calendar.MINUTE),
                    true);
                timePicker.show();
            }
        });

        dateButton = (Button)findViewById(R.id.whenDate);
        dateButton.setVisibility(View.GONE);
        dateButton.setOnClickListener(new OnClickListener() {
            /**
             * Demands date from the user.
             */
            private DatePickerDialog datePicker;

            public void onClick(View v) {
                datePicker =
                    new DatePickerDialog(SL.this,
                        new DatePickerDialog.OnDateSetListener() {
                        public void onDateSet(DatePicker view, int year,
                            int monthOfYear, int dayOfMonth)
                        {
                            getWhen().set(Calendar.YEAR, year);
                            getWhen().set(Calendar.MONTH, monthOfYear);
                            getWhen().set(Calendar.DAY_OF_MONTH, dayOfMonth);
                            datePicker.dismiss();
                            datePicker = null;
                            updateWhen();
                        }
                    },
                    getWhen().get(Calendar.YEAR),
                    getWhen().get(Calendar.MONTH),
                    getWhen().get(Calendar.DAY_OF_MONTH));
                datePicker.show();
            }
        });
    }

    @Override
    protected void onResume() {
        super.onResume();

        if (phoneLocator == null) {
            // Do nothing if we haven't gotten going properly yet
            return;
        }

        // Trigger location updates because the user is likely to want them RSN.
        phoneLocator.refresh();

        updateEndpoints();
        updateWhen();
    }

    /**
     * Make sure the data displayed for the journey endpoints is up to date, and
     * that the search button has the correct state.
     */
    private void updateEndpoints() {
        Set<ScoredLocation> scoredLocations = db.fetchLocations();
        Set<NamedLocation> namedLocations =
            new HashSet<NamedLocation>(scoredLocations.size());
        for (ScoredLocation location : scoredLocations) {
            namedLocations.add(location.getLocation());
        }

        if (fromLocation != null && !namedLocations.contains(fromLocation)) {
            // We have an unknown from location
            fromLocation = null;
        }
        if (toLocation != null && !namedLocations.contains(toLocation)) {
            // We have an unknown to location
            toLocation = null;
        }

        if (fromLocation == null) {
            // Select the most popular from location
            int bestFromScore = -1;
            for (ScoredLocation location : scoredLocations) {
                if (location.getFromScore() > bestFromScore) {
                    bestFromScore = location.getFromScore();
                    fromLocation = location.getLocation();
                }
            }
        }
        if (toLocation == null) {
            // Select the most popular to location
            int bestToScore = -1;
            for (ScoredLocation location : scoredLocations) {
                if (location.getToScore() > bestToScore) {
                    bestToScore = location.getToScore();
                    toLocation = location.getLocation();
                }
            }
        }

        if (fromLocation.samePlaceAs(toLocation)) {
            searchButton.setEnabled(false);
            searchButtonHint.setTextColor(Color.RED);
            searchButtonHint.setText("Välj olika platser för från och till!");
        } else {
            searchButton.setEnabled(true);
            searchButtonHint.setText("");
        }

        fromButton.setText(fromLocation.getAlias());
        toButton.setText(toLocation.getAlias());
    }

    @Override
    protected void onActivityResult(int requestCode, int resultCode, Intent data)
    {
        if (resultCode != RESULT_OK) {
            Log.w(TAG, "Ignoring non-OK result code " + resultCode
                  + " for request " + requestCode);
            return;
        }

        switch (requestCode) {
        case RequestCodes.CREATE_LOCATION:
            handleLocationSelection(data);
            break;
        case RequestCodes.SELECT_LOCATION:
            handleLocationSelection(data);
            break;
        default:
            Log.w(TAG, "Got activity result with unknown request code "
                  + requestCode);
        }
    }

    /**
     * A new location has been selected.  Deal with it.
     *
     * @param data Location selection results.
     */
    private void handleLocationSelection(Intent data) {
        NamedLocation namedLocation =
            data.getParcelableExtra(RequestKeys.LOCATION);

        // Change the to or from location selection
        int buttonId = data.getIntExtra(RequestKeys.LOCATION_SELECTOR_ID, -1);
        switch (buttonId) {
        case R.id.fromButton:
            fromLocation = namedLocation;
            break;

        case R.id.toButton:
            toLocation = namedLocation;
            break;

        default:
            Log.w(TAG, "Got new location for unknown button " + buttonId
                  + ", not changing any selection");
        }

        updateEndpoints();
    }

    @Override
    protected void onPause() {
        super.onPause();

        if (phoneLocator == null) {
            // Do nothing if we haven't gotten going properly yet
            return;
        }
        phoneLocator.cancelRefresh();
    }
}

