/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.launchpad.sl;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import net.launchpad.sl.location.LocationNameWatcher;
import net.launchpad.sl.location.LocationsDbAdapter;
import net.launchpad.sl.location.NamedLocation;
import net.launchpad.sl.location.ScoredLocation;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.DialogInterface;
import android.content.Intent;
import android.os.Bundle;
import android.text.TextWatcher;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.View.OnLongClickListener;
import android.widget.Button;
import android.widget.EditText;
import android.widget.RadioButton;
import android.widget.RadioGroup;
import android.widget.TextView;
import android.widget.RadioGroup.LayoutParams;
import android.widget.RadioGroup.OnCheckedChangeListener;

/**
 * List locations, let the user choose one of them or edit / delete existing
 * locations.
 *
 * @author johan.walles@gmail.com
 */
public class SelectLocationActivity extends Activity {
    /**
     * Locations are selected from this database.
     */
    private LocationsDbAdapter db;

    /**
     * The locations listed on screen.
     */
    private RadioGroup locationsList;

    /**
     * The available locations from the database.
     */
    private List<ScoredLocation> locations;

    /**
     * Accept the current choice in the locations list.
     */
    private Button locationSelectionOkButton;

    /**
     * The currently selected location.
     */
    private NamedLocation currentSelection;

    /**
     * Returns the currently selected location to the launcher of this activity.
     */
    private OnClickListener returnSelectedLocation = new OnClickListener() {
        public void onClick(View ignored) {
            ScoredLocation selection =
                locations.get(locationsList.getCheckedRadioButtonId());
            Log.d(SL.TAG,
                "Location selector returning "
                + selection.getLocation().getAlias()
                + " for spinner " + getLocationSelectorId()
                + "...");
            setResult(RESULT_OK,
                      SL.createLocationResult(selection.getLocation(),
                                              getLocationSelectorId()));
            finish();
        }
    };

    /**
     * Handles location long clicks, allowing the user to delete and edit
     * locations.
     */
    private class LocationLongClickListener
    implements OnLongClickListener
    {
        /**
         * This is the location we're supposed to operate upon.
         */
        private ScoredLocation victim;

        /**
         * Create a location long click handler.
         *
         * @param victim This is the location we will be operating upon.
         */
        public LocationLongClickListener(ScoredLocation victim) {
            this.victim = victim;
        }

        public boolean onLongClick(View longClickedView) {
            // Pop up a dialog for our ScoredLocation
            AlertDialog.Builder editLocationbuilder =
                new AlertDialog.Builder(SelectLocationActivity.this);
            LayoutInflater inflater =
                (LayoutInflater)getSystemService(LAYOUT_INFLATER_SERVICE);
            View editNameView = inflater.inflate(R.layout.edit_location, null);
            editLocationbuilder.setTitle("Redigera plats");
            editLocationbuilder.setView(editNameView);

            // Fill in the current name in the edit field
            final EditText nameEditor =
                (EditText)editNameView.findViewById(R.id.editName);
            nameEditor.setText(victim.getLocation().getAlias());

            // Make sure the user doesn't enter an existing name
            Button newNameOkButton =
                (Button)editNameView.findViewById(R.id.okButton);
            TextWatcher textWatcher =
                new LocationNameWatcher(db, nameEditor, newNameOkButton);
            nameEditor.addTextChangedListener(textWatcher);

            // Handle the user clicking the OK button
            final AlertDialog editLocationDialog = editLocationbuilder.create();
            newNameOkButton.setOnClickListener(new OnClickListener() {
                public void onClick(View v) {
                    String oldName = victim.getLocation().getAlias();
                    String newName = nameEditor.getText().toString();
                    if (!oldName.equals(newName)) {
                        db.renameLocation(oldName, newName);
                        refreshLocations();
                    }
                    editLocationDialog.dismiss();
                }
            });

            // Handle the user clicking the remove button
            Button removeButton =
                (Button)editNameView.findViewById(R.id.removeButton);
            removeButton.setOnClickListener(new OnClickListener() {
                public void onClick(View v) {
                    AlertDialog.Builder confirmRemoveBuilder =
                        new AlertDialog.Builder(SelectLocationActivity.this);
                    confirmRemoveBuilder.setTitle("Ta bort plats?");
                    confirmRemoveBuilder.setMessage("Är du säker på att du vill ta bort "
                        + victim.getLocation().getAlias()
                        + "?");
                    confirmRemoveBuilder.setIcon(android.R.drawable.ic_dialog_alert);

                    confirmRemoveBuilder.setPositiveButton("Ja",
                        new DialogInterface.OnClickListener()
                    {
                        public void onClick(DialogInterface dialog, int which) {
                            db.removeLocation(victim.getLocation().getAlias());
                            refreshLocations();
                            dialog.dismiss();
                            editLocationDialog.dismiss();
                        }
                    });

                    confirmRemoveBuilder.setNegativeButton("Nej",
                        new DialogInterface.OnClickListener() {
                            public void onClick(DialogInterface dialog, int which) {
                                dialog.dismiss();
                            }
                    });

                    confirmRemoveBuilder.show();
                }
            });

            editLocationDialog.show();

            return true;
        }
    }

    /**
     * Launch the location selector.
     *
     * @param launcher The activity that is launching the selector (usually
     * <i>this</i>).
     *
     * @param locationSelectorId The location selector we're selecting locations
     * for.  Can be R.id.fromButton or R.idToButton.
     *
     * @param currentSelection The currently selected item (if any)
     */
    public static void launch(Activity launcher,
            int locationSelectorId, NamedLocation currentSelection) {
        Intent intent =
            new Intent(launcher, SelectLocationActivity.class);
        intent.putExtra(RequestKeys.LOCATION_SELECTOR_ID, locationSelectorId);
        intent.putExtra(RequestKeys.LOCATION, currentSelection);
        launcher.startActivityForResult(intent, RequestCodes.SELECT_LOCATION);
    }

    /**
     * Give us the sort order for our locations.
     *
     * @return A comparator that can be used for sorting our locations.
     */
    private Comparator<ScoredLocation> getLocationComparator() {
        switch (getLocationSelectorId()) {
        case R.id.fromButton:
            return ScoredLocation.FROM_COMPARATOR;
        case R.id.toButton:
            return ScoredLocation.TO_COMPARATOR;
        default:
            throw new IllegalArgumentException("Spinner must be either fromSpinner or toSpinner");
        }
    }

    /**
     * Give us a header text.
     *
     * @return A header text that can be printed above the selection list.
     */
    private String getHeaderText() {
        switch (getLocationSelectorId()) {
        case R.id.fromButton:
            return "Välj varifrån du vill åka, eller långklicka på en plats för att ändra eller ta bort den:";
        case R.id.toButton:
            return "Välj vart du vill åka, eller långklicka på en plats för att ändra eller ta bort den:";
        default:
            throw new IllegalArgumentException("Spinner must be either fromSpinner or toSpinner");
        }
    }

    /**
     * Which location selector are we selecting locations for?
     *
     * @return The location selector we're selecting locations for.
     */
    private int getLocationSelectorId() {
        return getIntent().getIntExtra(RequestKeys.LOCATION_SELECTOR_ID, -1);
    }

    /**
     * Refresh the on-screen locations list.
     */
    private void refreshLocations() {
        // Fetch existing locations from disk
        Set<ScoredLocation> scoredLocations = db.fetchLocations();
        SortedSet<ScoredLocation> sortedLocations =
            new TreeSet<ScoredLocation>(getLocationComparator());
        sortedLocations.addAll(scoredLocations);
        locations = new ArrayList<ScoredLocation>(sortedLocations);

        // Make sure our on-screen locations list is virgin
        locationsList.removeAllViews();
        locationsList.setOnCheckedChangeListener(null);

        // List locations on screen
        RadioGroup.LayoutParams wideAndLow =
            new LayoutParams(
                android.view.ViewGroup.LayoutParams.FILL_PARENT,
                android.view.ViewGroup.LayoutParams.WRAP_CONTENT);
        boolean validSelection = false;
        for (int i = 0; i < locations.size(); i++) {
            ScoredLocation scoredLocation = locations.get(i);
            NamedLocation namedLocation = scoredLocation.getLocation();

            final RadioButton radioButton = new RadioButton(this);
            radioButton.setText(scoredLocation.getLocation().getAlias());
            radioButton.setId(i);
            radioButton.setLayoutParams(wideAndLow);

            if (namedLocation.samePlaceAs(currentSelection)) {
                // Check the currently selected location
                validSelection = true;
                radioButton.setChecked(true);

                // We need to special-case the user clicking on the already
                // selected radio button.  The change listener doesn't react
                // to that.
                radioButton.setOnClickListener(returnSelectedLocation);
            }

            if (!namedLocation.isHere()) {
                // Allow the user to edit / delete all locations except for the
                // "here" location
                radioButton.setOnLongClickListener(
                    new LocationLongClickListener(scoredLocation));
            }

            locationsList.addView(radioButton);
        }

        // Register a click listener that returns the selected location to our
        // caller.
        locationsList.setOnCheckedChangeListener(new OnCheckedChangeListener() {
            public void onCheckedChanged(RadioGroup group, int checkedId) {
                returnSelectedLocation.onClick(null);
            }
        });

        if (validSelection) {
            locationSelectionOkButton.setEnabled(true);
        } else {
            // No location is selected, cannot OK this
            locationSelectionOkButton.setEnabled(false);
        }
    }

    @Override
    protected void onResume() {
        super.onResume();

        refreshLocations();
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        currentSelection = getIntent().getParcelableExtra(RequestKeys.LOCATION);

        db = new LocationsDbAdapter(this, SL.DB_NAME);
        setContentView(R.layout.select_location);
        locationsList = (RadioGroup)findViewById(R.id.locationsList);

        // Update the header text
        TextView headerText = (TextView)findViewById(R.id.headerText);
        headerText.setText(getHeaderText());

        // Register a click listener for the add-new-location button
        Button addLocationButton = (Button)findViewById(R.id.addLocationButton);
        addLocationButton.setOnClickListener(new OnClickListener() {
            public void onClick(View v) {
                CreateLocationActivity.launch(SelectLocationActivity.this,
                    getLocationSelectorId());
            }
        });

        // Register a click listener for the OK button
        locationSelectionOkButton = (Button)findViewById(R.id.locationSelectOk);
        locationSelectionOkButton.setOnClickListener(returnSelectedLocation);
    }

    @Override
    protected void onActivityResult(int requestCode, int resultCode, Intent data)
    {
        if (resultCode != RESULT_OK) {
            Log.w(SL.TAG, "Ignoring non-OK result code " + resultCode
                  + " for request " + requestCode);
            return;
        }

        switch (requestCode) {
        case RequestCodes.CREATE_LOCATION:
            NamedLocation newLocation =
                data.getParcelableExtra(RequestKeys.LOCATION);
            currentSelection = newLocation;
            refreshLocations();
            break;

        default:
            Log.w(SL.TAG, "Got activity result with unknown request code "
                  + requestCode);
        }
    }
}
