/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.launchpad.sl.location;

import java.util.HashSet;
import java.util.Set;

import android.text.Editable;
import android.text.TextWatcher;
import android.widget.Button;
import android.widget.EditText;

/**
 * Make sure that the user enters a non-empty name that doesn't already exist.
 *
 * @author johan.walles@gmail.com
 */
public class LocationNameWatcher implements TextWatcher {
    /**
     * Names that we can't use.
     */
    private Set<String> existingLocationNames;

    /**
     * The name editor that we're watching.
     */
    private EditText nameEditor;

    /**
     * An OK button that we will enable / disable.
     */
    private Button okButton;

    /**
     * Create a new location name watcher.
     *
     * @param db The database in which to look for duplicate names.
     *
     * @param nameEditor The name editor that we're watching.
     *
     * @param okButton An OK button that we will enable / disable.
     */
    public LocationNameWatcher(LocationsDbAdapter db, EditText nameEditor,
        Button okButton)
    {
        this.nameEditor = nameEditor;
        this.okButton = okButton;

        // Set up existingLocationNames from the db
        Set<ScoredLocation> scoredLocations = db.fetchLocations();
        existingLocationNames = new HashSet<String>(scoredLocations.size());
        for (ScoredLocation scoredLocation : scoredLocations) {
            NamedLocation namedLocation = scoredLocation.getLocation();
            existingLocationNames.add(namedLocation.getAlias().toLowerCase());
        }

        // Make sure we accept the initial contents of the name editor in case
        // it is non-empty
        String initialName = nameEditor.getText().toString();
        if (initialName.length() > 0) {
            existingLocationNames.remove(initialName.toLowerCase());
        }

        // Make sure we start out good
        afterTextChanged(nameEditor.getText());
    }

    public void onTextChanged(CharSequence s, int start, int before, int count)
    {
        // This method intentionally left blank
    }

    public void beforeTextChanged(CharSequence s, int start, int count,
        int after)
    {
        // This method intentionally left blank
    }

    public void afterTextChanged(Editable s) {
        if (s.length() == 0) {
            // Location name is empty
            enableAccept(false, null);
        } else if (existingLocationNames.contains(s.toString().toLowerCase())) {
            // We have an existing location name
            enableAccept(false, "Det finns redan en plats som heter \"" + s + "\".");
        } else {
            // This is a new location name
            enableAccept(true, null);
        }
    }

    /**
     * Enable or disable the user's possibility of accepting the current text
     * as the name of this location.
     *
     * @param enabled True to enable acceptance, false otherwise.
     *
     * @param message A message describing why the button is disabled, or null
     * for no message.
     */
    private void enableAccept(boolean enabled, String message) {
        okButton.setEnabled(enabled);
        if (enabled) {
            nameEditor.setError(null);
        } else {
            nameEditor.setError(message);
        }

        // FIXME: We would like to disable the virtual keyboard's DONE button
        // here, but I can't come up with anything suitable to pass to
        // nameEditor.setImeOptions() for that button to become disabled.
    }
}
