/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl.location;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;

import net.launchpad.sl.SL;

import android.content.ContentValues;
import android.content.Context;
import android.database.Cursor;
import android.database.SQLException;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteException;
import android.location.Address;
import android.location.Geocoder;
import android.location.Location;
import android.util.Log;

/**
 * Store and retrieve {@link ScoredLocation}s from disk.
 */
public class LocationsDbAdapter {
    /**
     * Current database schema version.
     */
    private final static int SCHEMA_VERSION = 3;

    /**
     * Used for OS access.
     */
    private Context context;

    /**
     * The name of the database we're operating upon.
     */
    private String dbPath;

    /**
     * Progress can be reported to implementors of this interface.
     */
    public interface Progress {
        /**
         * Set the current progress in percent.
         *
         * @param percent A progress percentage 0-100.
         */
        public void onProgress(int percent);
    }

    /**
     * Connect to the database.
     *
     * @param context Used for application private database access.
     *
     * @param dbName The name of the database to use for persistence.
     */
    public LocationsDbAdapter(Context context, String dbName) {
        this.context = context;
        this.dbPath = context.getDatabasePath(dbName).getAbsolutePath();
    }

    /**
     * Create a locations database with schema version 1 and one entry in it for
     * the phone's current location.
     * <p>
     * Has default protection for testing purposes.
     *
     * @param db The database to initialize.
     */
    void createBaseDb(SQLiteDatabase db) {
        final String createTable =
            "create table locations "
            + "(name text primary key, "
            + "latitude real, "
            + "longitude real, "
            + "from_score integer not null, "
            + "to_score integer not null);";

        Log.d(SL.TAG, "Creating a new locations database...");

        db.execSQL(createTable);

        // Insert the phone's current location
        ContentValues values = new ContentValues();
        values.put("name", "Telefonens nuvarande position");
        values.put("from_score", 1);
        values.put("to_score", 0);
        db.insertOrThrow("locations", null, values);

        db.setVersion(1);

        Log.d(SL.TAG, "New database successfully created");
    }

    /**
     * Open a connection to our backing store.  The database will be created
     * or upgraded as necessary for this connection to be useful.
     *
     * @param flags {@link SQLiteDatabase#OPEN_READONLY} or
     * {@link SQLiteDatabase#OPEN_READWRITE}.
     *
     * @return A database connection that can be used for locations storage.
     */
    private SQLiteDatabase getDatabase(int flags) {
        SQLiteDatabase db;

        // Create the database if necessary
        File dbFile = new File(dbPath);
        if (!dbFile.exists()) {
            // We sometimes have to create a directory to put the database file
            // in.
            dbFile.getParentFile().mkdirs();

            db = SQLiteDatabase.openOrCreateDatabase(dbPath, null);
            try {
                createBaseDb(db);
            } finally {
                db.close();
                db = null;
            }
        }

        // Upgrade the database if necessary
        if (needsUpgrade()) {
            upgradeDb(null);
        }

        return SQLiteDatabase.openDatabase(dbPath, null, flags);
    }

    /**
     * Open a read-only database connection.
     *
     * @return A read-only database connection.
     */
    private SQLiteDatabase getReadableDatabase() {
        return getDatabase(SQLiteDatabase.OPEN_READONLY);
    }

    /**
     * Open a read-write database connection.
     *
     * @return A read-write database connection.
     */
    private SQLiteDatabase getWritableDatabase() {
        return getDatabase(SQLiteDatabase.OPEN_READWRITE);
    }

    /**
     * Does the database have to be upgraded to a new schema format?
     *
     * @return True if the database needs to be upgraded to a new schema format.
     * False otherwise.
     */
    public boolean needsUpgrade() {
        SQLiteDatabase db;
        try {
            db = SQLiteDatabase.openDatabase(
                dbPath, null, SQLiteDatabase.OPEN_READONLY);
        } catch (SQLiteException e) {
            // Couldn't open database; if it needs something it's not an upgrade
            return false;
        }

        int schemaVersion = db.getVersion();
        db.close();

        if (schemaVersion == SCHEMA_VERSION) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * Upgrade the database schema.
     *
     * @param progress Upgrade progress reports go here
     */
    public void upgradeDb(Progress progress) {
        SQLiteDatabase db = SQLiteDatabase.openDatabase(
            dbPath, null, SQLiteDatabase.OPEN_READWRITE);

        try {
            db.beginTransaction();

            int currentVersion = db.getVersion();
            if (currentVersion == SCHEMA_VERSION) {
                // Already at the correct version, no need to do anything
                return;
            }

            if (currentVersion == 1) {
                upgradeDbV1ToV2(progress, db);
            }

            if (currentVersion <= 2) {
                upgradeDbV2ToV3(db);
            }

            // Store the new schema version number
            db.setVersion(SCHEMA_VERSION);
            db.setTransactionSuccessful();
        } finally {
            db.endTransaction();
            db.close();
        }
    }

    /**
     * Upgrade database schema from v2 to v3.
     *
     * @param db The database to upgrade.
     */
    private void upgradeDbV2ToV3(SQLiteDatabase db) {
        // Rename column "name" to "alias"
        // Rename column "description" to "addressOrStop"
        // Add column "isAddressOrStop"

        // From "http://www.sqlite.org/faq.html#q11":
        String newColumns = "(alias text primary key, "
            + "isAddressOrStop integer not null,"
            + "addressOrStop text,"
            + "latitude real, "
            + "longitude real, "
            + "from_score integer not null, "
            + "to_score integer not null)";
        db.execSQL("CREATE TEMPORARY TABLE temp " + newColumns + ";");
        db.execSQL("INSERT INTO temp SELECT "
            + "name, 0, description, latitude, longitude, from_score, to_score "
            + "FROM locations;");
        db.execSQL("DROP TABLE locations;");
        db.execSQL("CREATE TABLE locations" + newColumns + ";");
        db.execSQL("INSERT INTO locations SELECT "
            + "alias, isAddressOrStop, addressOrStop, latitude, longitude, from_score, to_score "
            + "FROM temp;");
        db.execSQL("DROP TABLE temp;");
    }

    /**
     * Upgrade database schema from v1 to v2.
     *
     * @param progress If non-null, progress updates will be sent here.
     *
     * @param db The database to upgrade.
     */
    void upgradeDbV1ToV2(Progress progress, SQLiteDatabase db) {
        // Add the description column
        final String addColumn =
            "alter table locations "
            + "add description text;";
        db.execSQL(addColumn);

        // Populate the description column based on the locations
        Geocoder geocoder = LocationUtil.createGeocoder(context);
        Cursor cursor =
            db.query("locations",
                new String[] { "name", "latitude", "longitude" },
                "latitude is not null", null, null, null, null);
        int rows = cursor.getCount();
        try {
            if (progress != null) {
                progress.onProgress(rows == 0 ? 100 : 0);
            }

            Pattern containsDigit = Pattern.compile("^.*[0-9].*$");
            while (cursor.moveToNext()) {
                String name = cursor.getString(0);
                double latitude = cursor.getDouble(1);
                double longitude = cursor.getDouble(2);
                CharSequence description = name;

                if (containsDigit.matcher(name).matches()) {
                    // The name contains digits, meaning that it's probably
                    // already an address.  No need to screw it up by
                    // looking up its coordinates and possibly finding some
                    // other address.
                } else {
                    try {
                        // Look up the address of this location
                        List<Address> descriptions =
                            geocoder.getFromLocation(latitude, longitude, 1);
                        if (descriptions.size() == 1) {
                            CharSequence generatedDescription =
                                LocationUtil.getSingleLineDescription(
                                    descriptions.get(0));
                            if (generatedDescription.length() > 0) {
                                description = generatedDescription;
                            }
                        }
                    } catch (IOException e) {
                        // Generating a proper description failed, do nothing in
                        // this case and just stick to using the name
                    }
                }

                ContentValues values = new ContentValues();
                values.put("description", description.toString());
                db.update("locations",
                    values,
                    "name = ?", new String[] { name });

                if (progress != null) {
                    int nextRow = cursor.getPosition() + 1;
                    progress.onProgress((nextRow * 100) / rows);
                }
            }
        } finally {
            cursor.close();
        }
        db.setVersion(2);
    }

    /**
     * Fetch the scored locations from the database.
     *
     * @return The scored locations.
     */
    public Set<ScoredLocation> fetchLocations() {
        Log.d(SL.TAG, "Reading locations from database...");

        SQLiteDatabase db = getReadableDatabase();

        try {
            Cursor cursor =
                db.query("locations",
                         new String[] { "addressOrStop",
                                        "latitude", "longitude",
                                        "alias",
                                        "from_score", "to_score",
                                        "isAddressOrStop"
                         },
                         null, null, null, null, null);
            Log.d(SL.TAG,
                  "  Database contains " + cursor.getCount() + " locations");

            Set<ScoredLocation> locations = new HashSet<ScoredLocation>();
            while (cursor.moveToNext()) {
                String addressOrStop = cursor.getString(0);
                Double latitude = null;
                Double longitude = null;
                if (!cursor.isNull(1)) {
                    latitude = cursor.getDouble(1);
                    longitude = cursor.getDouble(2);
                }
                String alias = cursor.getString(3);
                int fromScore = cursor.getInt(4);
                int toScore = cursor.getInt(5);
                boolean isAddressOrStop = (cursor.getInt(6) != 0);

                NamedLocation namedLocation = null;
                if (isAddressOrStop) {
                    // This is an address / stop
                    namedLocation = new NamedLocation(addressOrStop, alias);
                } else if (latitude == null) {
                    // This is a here location
                    namedLocation = new NamedLocation(alias);
                } else {
                    // This is a coordinate
                    namedLocation =
                        new NamedLocation(addressOrStop, latitude, longitude, alias);
                }

                ScoredLocation scoredLocation =
                    new ScoredLocation(namedLocation, fromScore, toScore);
                Log.d(SL.TAG, "  Location read from database: " + scoredLocation);
                locations.add(scoredLocation);
            }
            cursor.close();

            Log.d(SL.TAG,
                  "  " + locations.size() + " locations successfully read from database");

            return locations;
        } finally {
            db.close();
        }
    }

    /**
     * Fetch a scored location from the database.
     *
     * @param name The name of the scored location to fetch.
     *
     * @return The scored location, or null if it wasn't found.
     */
    public ScoredLocation fetchByName(String name) {
        Set<ScoredLocation> locations = fetchLocations();
        for (ScoredLocation location : locations) {
            if (name.equals(location.getLocation().getAlias())) {
                return location;
            }
        }

        return null;
    }

    /**
     * Insert a location into the database.
     *
     * @param db The database to insert the location into.
     *
     * @param insertMe The location to insert into the database.
     */
    private void insert(SQLiteDatabase db, ScoredLocation insertMe) {
        NamedLocation namedLocation = insertMe.getLocation();
        ContentValues values = new ContentValues();
        values.put("addressOrStop", namedLocation.getAddressOrStop());
        values.put("from_score", insertMe.getFromScore());
        values.put("to_score", insertMe.getToScore());
        values.put("alias", namedLocation.getAlias());
        values.put("isAddressOrStop", namedLocation.isAddressOrStop());

        boolean hasCoordinates = !insertMe.getLocation().isAddressOrStop();
        if (hasCoordinates && !insertMe.getLocation().isHere()) {
            // Set latitude and longitude for not-here locations
            Location location = namedLocation.getCoordinate();

            values.put("latitude", location.getLatitude());
            values.put("longitude", location.getLongitude());
        }

        db.insertOrThrow("locations", null, values);
    }

    /**
     * Store locations to database.  Calling this method will overwrite the
     * current contents of the database, and upon return only the locations
     * passed to here will be in the database.
     *
     * @param locations The locations to store.
     */
    public void storeLocations(Collection<ScoredLocation> locations) {
        Log.d(SL.TAG,
              "Storing " + locations.size() + " locations to database...");

        SQLiteDatabase db = getWritableDatabase();

        try {
            db.beginTransaction();

            db.delete("locations", null, null);
            for (ScoredLocation insertMe : locations) {
                Log.d(SL.TAG, "  Location written to database: " + insertMe);
                insert(db, insertMe);
            }

            db.setTransactionSuccessful();

            Log.d(SL.TAG,
                  "  " + locations.size() + " locations successfully stored to database");
        } catch(SQLException e) {
            Log.e(SL.TAG, "Failed storing locations to database", e);
        } finally {
            db.endTransaction();
            db.close();
        }
    }

    /**
     * A user has searched for a trip between from and to.  Update their
     * respective from and to scores in the database.
     *
     * @param from Where the user started her trip.
     *
     * @param to Where the user is going.
     */
    public void increaseScores(NamedLocation from, NamedLocation to) {
        SQLiteDatabase db = getWritableDatabase();

        try {
            db.beginTransaction();

            // Update the starting location's from score
            Cursor result = db.query("locations", new String[] {"from_score"},
                "alias = ?", new String[] { from.getAlias() },
                null, null, null);
            if (result.getCount() != 1) {
                throw new SQLException("Expected one result for <"
                    + from.getAlias()
                    + ", got >" + result.getCount());
            }
            result.moveToFirst();
            int oldScore = result.getInt(0);
            ContentValues newScore = new ContentValues();
            newScore.put("from_score", oldScore + 1);
            result.close();
            int rowsUpdated = db.update("locations",
                newScore,
                "alias = ?", new String[] { from.getAlias() });
            if (rowsUpdated != 1) {
                throw new SQLException("Should have updated one row, did " + rowsUpdated);
            }

            // Update the destination location's to score
            result = db.query("locations", new String[] {"to_score"},
                              "alias = ?", new String[] { to.getAlias() },
                              null, null, null);
            if (result.getCount() != 1) {
                throw new SQLException("Expected one result for <"
                                       + from.getAlias()
                                       + ", got >" + result.getCount());
            }
            result.moveToFirst();
            oldScore = result.getInt(0);
            newScore = new ContentValues();
            newScore.put("to_score", oldScore + 1);
            result.close();
            rowsUpdated = db.update("locations",
                newScore,
                "alias = ?", new String[] { to.getAlias() });
            if (rowsUpdated != 1) {
                throw new SQLException("Should have updated one row, did " + rowsUpdated);
            }

            db.setTransactionSuccessful();
        } catch(SQLException e) {
            Log.e(SL.TAG, "Failed increasing scores of " + from + " and " + to, e);
        } finally {
            db.endTransaction();
            db.close();
        }
    }

    /**
     * Change the name of a location.
     *
     * @param oldAlias The location's current name.
     *
     * @param newAlias The name we want the location to have.
     */
    public void renameLocation(String oldAlias, String newAlias) {
        SQLiteDatabase db = getWritableDatabase();

        try {
            db.beginTransaction();

            ContentValues newAliasValue = new ContentValues();
            newAliasValue.put("alias", newAlias);
            int rowsUpdated = db.update("locations",
                newAliasValue,
                "alias = ?", new String[] { oldAlias });
            if (rowsUpdated != 1) {
                throw new SQLException("Should have updated one row, did " + rowsUpdated);
            }

            db.setTransactionSuccessful();
        } catch(SQLException e) {
            Log.e(SL.TAG, "Failed to change name of <" + oldAlias + "> to <" + newAlias + ">", e);
        } finally {
            db.endTransaction();
            db.close();
        }
    }

    /**
     * Remove a location from the database.
     *
     * @param locationAlias The name of the location to remove.
     */
    public void removeLocation(String locationAlias) {
        SQLiteDatabase db = getWritableDatabase();

        try {
            int rowsDeleted = db.delete("locations",
                "alias = ?", new String[] { locationAlias });
            if (rowsDeleted != 1) {
                throw new SQLException("Should have deleted one row, did " + rowsDeleted);
            }
        } catch(SQLException e) {
            Log.e(SL.TAG, "Failed to remove <" + locationAlias + "> from the database", e);
        } finally {
            db.close();
        }
    }
}
