/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl.location;

import android.location.Location;
import android.os.Parcel;
import android.os.Parcelable;

/**
 * This class represents a named location.
 *
 * @author johan.walles@gmail.com
 */
public class NamedLocation implements Parcelable {
    /**
     * Where is this?
     */
    private Location coordinate;

    /**
     * What's this place called?
     */
    private String alias;

    /**
     * Where is this place?  This description can end up as a map marker, and
     * is probably a street address or the name of a train / bus stop.
     * <p>
     * If the this field is null, it won't be marked on any map.
     */
    private CharSequence addressOrStop;

    /**
     * True if this is an address or stop.  False if this is a named coordinate.
     */
    private boolean isAddressOrStop;

    /**
     * Create a new named location.
     *
     * @param addressOrStop An address or the name of a train / bus stop.  Will
     * be written on map markers. If this is null, this location will not be
     * marked on any maps.
     *
     * @param coordinate The coordinate we represent.
     *
     * @param alias The alias name of this location, commonly used for naming
     * location bookmarks.
     */
    public NamedLocation(CharSequence addressOrStop, Location coordinate,
        String alias)
    {
        if (alias == null) {
            throw new NullPointerException("The alias cannot be null");
        }

        this.alias = alias;
        this.addressOrStop = addressOrStop;
        this.coordinate = coordinate;
    }

    /**
     * Create a new named location at the given coordinates.
     *
     * @param addressOrStop An address or the name of a train / bus stop.  Will
     * be written on map markers. If this is null, this location will not be
     * marked on any maps.
     *
     * @param latitude The latitude of this location.
     *
     * @param longitude The longitude of this location.
     *
     * @param alias The alias name of this location, commonly used for naming
     * location bookmarks.
     */
    public NamedLocation(CharSequence addressOrStop,
        double latitude, double longitude,
        String alias)
    {
        if (alias == null) {
            throw new NullPointerException("The alias cannot be null");
        }

        Location newCoordinate = new Location("Johan");
        newCoordinate.setLatitude(latitude);
        newCoordinate.setLongitude(longitude);

        this.addressOrStop = addressOrStop;
        this.coordinate = newCoordinate;
        this.alias = alias;
    }

    /**
     * Create a new named location representing the phone's current location.
     *
     * @param alias The alias name of this location, commonly used for naming
     * location bookmarks.
     */
    public NamedLocation(String alias) {
        if (alias == null) {
            throw new NullPointerException("The alias cannot be null");
        }

        this.alias = alias;
    }

    /**
     * Create a new named location representing an address or bus / train stop.
     *
     * @param addressOrStop The address or bus / train stop to name.
     *
     * @param alias The alias name for this address or bus / train stop.  Can
     * be null.
     */
    public NamedLocation(CharSequence addressOrStop, String alias) {
        if (addressOrStop == null) {
            throw new NullPointerException("Address / stop cannot be null");
        }

        this.addressOrStop = addressOrStop;
        if (alias != null) {
            this.alias = alias;
        } else {
            this.alias = addressOrStop.toString();
        }
        isAddressOrStop = true;
    }

    /**
     * Does this location represent the phone's current location?
     *
     * @return True if this location represents the phone's current location.
     * False otherwise.
     *
     * @throws IllegalStateException if I'm an address or stop.  Only coordinate
     * locations can potentially be *here*.
     *
     * @see #isAddressOrStop()
     */
    public boolean isHere() {
        if (isAddressOrStop()) {
            throw new IllegalStateException("I'm an address / stop, can't be *here*");
        }

        return coordinate == null;
    }

    /**
     * Does this location represent an address or bus / train stop?  The other
     * alternative is that we represent a coordinate.
     *
     * @return True if this is an address or stop.  False if this is a
     * coordinate.
     */
    public boolean isAddressOrStop() {
        return isAddressOrStop;
    }

    /**
     * Get an alias name for this location.  The alias name can be used as the
     * name of a bookmark for example.
     *
     * @return The alias name of this location.
     */
    public String getAlias() {
        return alias;
    }

    /**
     * Get a description of where this location is.
     *
     * @return A description of where this location is, or null if this location
     * should not be marked on any map.  Could be a street address or the name
     * of a train / bus stop.
     */
    public String getAddressOrStop() {
        if (addressOrStop == null) {
            return null;
        }

        return addressOrStop.toString();
    }

    /**
     * The coordinate we're naming.
     *
     * @return The coordinate we're naming, or null for "the phone's current
     * position".
     *
     * @throws IllegalStateException if the coordinates are unknown.
     */
    public Location getCoordinate() {
        if (isHere()) {
            throw new IllegalStateException("I'm *here*, go ask the PhoneLocator if you want coordinates");
        }
        return coordinate;
    }

    public int describeContents() {
        return 0;
    }

    public void writeToParcel(Parcel dest, int flags) {
        dest.writeBooleanArray(new boolean[] { isAddressOrStop });
        dest.writeString(getAddressOrStop());
        dest.writeParcelable(coordinate, 0);
        dest.writeString(alias);
    }

    /**
     * Part of the protocol for the {@link Parcelable} interface.
     */
    public static final Parcelable.Creator<NamedLocation> CREATOR =
        new Parcelable.Creator<NamedLocation>()
    {
        public NamedLocation createFromParcel(Parcel source) {
            boolean[] booleans = source.createBooleanArray();
            boolean isAddressOrStop = booleans[0];
            String addressOrStop = source.readString();
            Location coordinate = (Location)source.readParcelable(null);
            String alias = source.readString();
            if (isAddressOrStop) {
                return new NamedLocation(addressOrStop, alias);
            } else {
                return new NamedLocation(addressOrStop, coordinate, alias);
            }
        }

        public NamedLocation[] newArray(int size) {
            return new NamedLocation[size];
        }
    };

    /**
     * Does these two locations represent the same place?
     *
     * @param maybeTheSamePlace The other location to test for sameness.
     *
     * @return True if the two locations represent the same place.  False
     * otherwise.
     */
    public boolean samePlaceAs(NamedLocation maybeTheSamePlace) {
        if (isAddressOrStop()) {
            if (!maybeTheSamePlace.isAddressOrStop()) {
                return false;
            }
            return getAddressOrStop().equals(maybeTheSamePlace.getAddressOrStop());
        }

        boolean imHere = isHere();
        boolean theOtherGuyIsHere = maybeTheSamePlace.isHere();
        if (imHere != theOtherGuyIsHere) {
            return false;
        }
        if (imHere && theOtherGuyIsHere) {
            // We're both here
            return true;
        }

        // We only care about latitude and longitude
        if (coordinate.getLatitude() != maybeTheSamePlace.coordinate.getLatitude()) {
            return false;
        }
        if (coordinate.getLongitude() != maybeTheSamePlace.coordinate.getLongitude()) {
            return false;
        }

        return true;
    }

    @Override
    public boolean equals(Object o) {
        NamedLocation theOtherGuy = (NamedLocation)o;

        if (isAddressOrStop != theOtherGuy.isAddressOrStop) {
            return false;
        }

        if (!alias.equals(theOtherGuy.alias)) {
            return false;
        }

        if ((addressOrStop == null) != (theOtherGuy.addressOrStop == null)) {
            return false;
        }
        if (addressOrStop != null) {
            if (!addressOrStop.equals(theOtherGuy.addressOrStop)) {
                return false;
            }
        }

        return samePlaceAs(theOtherGuy);
    }

    @Override
    public int hashCode() {
        int hashCode = alias.hashCode();
        if (addressOrStop != null) {
            hashCode ^= addressOrStop.hashCode();
        }
        if (coordinate != null) {
            hashCode ^= new Double(coordinate.getLatitude()).hashCode();
            hashCode ^= new Double(coordinate.getLongitude()).hashCode();
        }
        if (isAddressOrStop) {
            hashCode ^= 0x800000;
        }
        return hashCode;
    }

    @Override
    public String toString() {
        if (addressOrStop == null) {
            return alias;
        }
        if (alias == null) {
            return addressOrStop.toString();
        }
        if (addressOrStop.toString().equals(alias)) {
            return alias;
        }
        return addressOrStop + " (" + alias + ")";
    }
}
