/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.launchpad.sl.location;

import net.launchpad.sl.SL;
import android.location.GpsSatellite;
import android.location.GpsStatus;
import android.location.Location;
import android.location.LocationListener;
import android.location.LocationManager;
import android.location.LocationProvider;
import android.os.Build;
import android.os.Bundle;
import android.util.Log;

/**
 * Make best effort attempts at retrieving the phone's current position.
 *
 * @author johan.walles@gmail.com
 */
public class PhoneLocator {
    /**
     * How many meters off is it OK for the positioning to be?
     *
     * @see #isPrecisionOk()
     */
    private static final int MAXIMUM_ERROR_METERS = 100;

    /**
     * When comparing two locations of different age and with different
     * precision, how much do we consider the error in meters to increase per
     * second of age?
     */
    private final static int ERROR_METERS_PER_SECOND = 1;

    /**
     * Locations older than this are considered too old.
     */
    public final static int MAX_LOCATION_AGE_MS =
        (1000 * MAXIMUM_ERROR_METERS) / ERROR_METERS_PER_SECOND;

    /**
     * This is where we get positioning data from.
     */
    private LocationManager locationManager;

    /**
     * How many GPS satellites are currently visible?
     */
    private int gpsVisibleSatellites = 0;

    /**
     * How many GPS satellites were used to get the last fix?
     */
    private int gpsUsedSatellites = 0;

    /**
     * Listens and reacts to GPS location updates.
     */
    private SlLocationListener gpsLocationListener;

    /**
     * Listens and reacts to network location updates.
     */
    private SlLocationListener networkLocationListener;

    /**
     * Listens and reacts to GPS status changes..
     */
    private SlGpsStatusListener gpsStatusListener;

    /**
     * The user has told us where (s)he is.
     */
    private Location manualLocation;

    /**
     * Listens for positioning updates.
     *
     * @author johan.walles@gmail.com
     */
    public interface PositioningListener {
        /**
         * Something about the phone's positioning system has changed.
         */
        public void positioningUpdated();
    }

    /**
     * Listens to and reacts to GPS status changes.
     */
    private class SlGpsStatusListener implements GpsStatus.Listener {
        /**
         * The last received GPS status.
         */
        private GpsStatus gpsStatus;

        /**
         * This listener will be notified on GPS status updates.
         */
        private PositioningListener positioningListener;

        /**
         * Create a new GPS status listener that passes events on to a
         * positioning listener.
         *
         * @param positioningListener The positioning listener to pass on events
         * to.
         */
        public SlGpsStatusListener(PositioningListener positioningListener) {
            this.positioningListener = positioningListener;
        }

        public void onGpsStatusChanged(int event) {
            gpsStatus = locationManager.getGpsStatus(gpsStatus);

            // Count satellites
            gpsVisibleSatellites = 0;
            gpsUsedSatellites = 0;
            for (GpsSatellite satellite : gpsStatus.getSatellites()) {
                gpsVisibleSatellites++;
                if (satellite.usedInFix()) {
                    gpsUsedSatellites++;
                }
            }

            if (event == GpsStatus.GPS_EVENT_STOPPED) {
                gpsVisibleSatellites = 0;
                gpsUsedSatellites = 0;
            }

            Log.d(SL.TAG, "Got GPS status update, "
                + gpsUsedSatellites + "/" + gpsVisibleSatellites
                + " visible satellites used for positioning");

            positioningListener.positioningUpdated();
        }
    }

    /**
     * Listens to location updates and keeps the warning UI up to date.
     */
    private class SlLocationListener implements LocationListener {
        /**
         * This listener will be notified on location updates.
         */
        private PositioningListener positioningListener;

        /**
         * Create a new location listener that passes events on to a positioning
         * listener.
         *
         * @param positioningListener The positioning listener to pass on events
         * to.
         */
        public SlLocationListener(PositioningListener positioningListener) {
            this.positioningListener = positioningListener;
        }

        public void onStatusChanged(String provider, int status, Bundle extras) {
            Log.d(SL.TAG, provider + " status changed to " + status);
            positioningListener.positioningUpdated();
        }

        public void onProviderEnabled(String provider) {
            Log.d(SL.TAG, "Location provider enabled: " + provider);
            positioningListener.positioningUpdated();
        }

        public void onProviderDisabled(String provider) {
            Log.d(SL.TAG, "Location provider disabled: " + provider);
            positioningListener.positioningUpdated();
        }

        public void onLocationChanged(Location location) {
            Log.d(SL.TAG, "Got new " + location.getProvider() + " location, " +
                    "error is " + location.getAccuracy() + "m");
            positioningListener.positioningUpdated();
        }
    }

    /**
     * This location listener will just remove itself as soon as it receives the
     * first update.  Can be used to trigger exactly one update from some
     * location provider.
     *
     * @see LocationManager#requestLocationUpdates(String, long, float, LocationListener)
     * @see #refresh()
     */
    private class RefreshListener implements LocationListener {
        /**
         * What kind of refreshes are we listening for?
         */
        private String kind;

        /**
         * Create a new refresh listener for a certain kind of refreshes.
         *
         * @param kind What kind of refresh are we waiting for?  Probably "GPS"
         * or "network".
         */
        public RefreshListener(String kind) {
            this.kind = kind;
        }

        public void onLocationChanged(Location location) {
            Log.d(SL.TAG, kind + " location successfully refreshed");
            locationManager.removeUpdates(this);
        }

        public void onProviderDisabled(String provider) {
            Log.d(SL.TAG, kind + " positioning disabled, giving up on refresh");
            locationManager.removeUpdates(this);
        }

        public void onProviderEnabled(String provider) {
            // Ignore this and keep waiting for a location update
        }

        public void onStatusChanged(String provider, int status, Bundle extras)
        {
            if (status == LocationProvider.AVAILABLE) {
                // Ignore this and keep waiting for a location update
                return;
            }
            if (status == LocationProvider.TEMPORARILY_UNAVAILABLE) {
                // Keep waiting for the location provider to come on line
                Log.d(SL.TAG, kind +
                    " refresh waiting for location provider to come online");
                return;
            }

            // Provider is unavailable and we won't get any location update,
            // give up
            Log.d(SL.TAG, kind + " positioning unavailable (status " + status
                + "), giving up on refresh");
            locationManager.removeUpdates(this);
        }
    }

    /**
     * Listens for new GPS positions.
     */
    private RefreshListener gpsRefreshListener = new RefreshListener("GPS");

    /**
     * Listens for new network positions.
     */
    private RefreshListener netRefreshListener = new RefreshListener("network");

    /**
     * Create a new positioning helper.
     *
     * @param locationManager The location manager to use for positioning.
     */
    public PhoneLocator(LocationManager locationManager) {
        this.locationManager = locationManager;
    }

    /**
     * Disable positioning change listeners.
     */
    public void disableLocationListeners() {
        if (gpsLocationListener != null) {
            locationManager.removeUpdates(gpsLocationListener);
            gpsLocationListener = null;
        }
        if (networkLocationListener != null) {
            locationManager.removeUpdates(networkLocationListener);
            networkLocationListener = null;
        }
        if (gpsStatusListener != null) {
            locationManager.removeGpsStatusListener(gpsStatusListener);
            gpsStatusListener = null;
        }
    }

    /**
     * Enable positioning listeners.
     *
     * @param positioningListener This listener will be notified on positioning
     * events.
     */
    public void enableLocationListeners(PositioningListener positioningListener)
    {
        if (gpsLocationListener == null) {
            gpsLocationListener = new SlLocationListener(positioningListener);
            locationManager.requestLocationUpdates(
                LocationManager.GPS_PROVIDER,
                1000, MAXIMUM_ERROR_METERS / 2,
                gpsLocationListener);
        }
        if (networkLocationListener == null) {
            networkLocationListener = new SlLocationListener(positioningListener);
            locationManager.requestLocationUpdates(
                LocationManager.NETWORK_PROVIDER,
                15000, MAXIMUM_ERROR_METERS / 2,
                networkLocationListener);
        }
        if (gpsStatusListener == null) {
            gpsStatusListener = new SlGpsStatusListener(positioningListener);
            locationManager.addGpsStatusListener(gpsStatusListener);
        }
    }

    /**
     * Find out if we're running on emulated hardware.
     *
     * @return true if we're running on the emulator, false otherwise
     */
    private boolean isRunningOnEmulator() {
        return "unknown".equals(Build.BOARD);
    }

    /**
     * Get the phone's current location from:<ul>
     * <li>GPS</li>
     * <li>Network</li>
     * <li>Hard-coded constant if running on an emulator</li>
     * </ul>
     *
     * @return The phone's current location, or null if the phone doesn't know
     * its location right now.
     *
     * @see #getLastKnownLocation()
     */
    public Location getCurrentLocation() {
        Location here  = getLastKnownLocation();
        if (here == null) {
            return null;
        }

        if (getErrorMeters(here) <= MAXIMUM_ERROR_METERS) {
            return here;
        } else {
            return null;
        }
    }

    /**
     * Computes the error score for a location using
     * {@link #ERROR_METERS_PER_SECOND}.
     *
     * @param location The location to compute the error for.
     *
     * @return The error for that location in meters.
     */
    public int getErrorMeters(Location location) {
        int locationAgeSeconds =
            (int)((System.currentTimeMillis() - location.getTime()) / 1000);
        return (int)(
            location.getAccuracy()
            + locationAgeSeconds * ERROR_METERS_PER_SECOND);
    }

    /**
     * Get the phone's last known location from:<ul>
     * <li>GPS</li>
     * <li>Network</li>
     * <li>Hard-coded constant if running on an emulator</li>
     * </ul>
     *
     * @return The phone's last known location.
     */
    public Location getLastKnownLocation() {
        Location here = null;

        // Fetch GPS and network location
        Location gpsLocation =
            locationManager.getLastKnownLocation(LocationManager.GPS_PROVIDER);
        Location networkLocation =
            locationManager.getLastKnownLocation(LocationManager.NETWORK_PROVIDER);

        // Select from available locations
        int bestErrorMeters = 0;
        if (manualLocation != null) {
            int error = getErrorMeters(manualLocation);
            if (here == null || (error < bestErrorMeters)) {
                here = manualLocation;
                bestErrorMeters = error;
            }
        }
        if (gpsLocation != null) {
            int error = getErrorMeters(gpsLocation);
            if (here == null || (error < bestErrorMeters)) {
                here = gpsLocation;
                bestErrorMeters = error;
            }
        }
        if (networkLocation != null) {
            int error = getErrorMeters(networkLocation);
            if (here == null || (error < bestErrorMeters)) {
                here = networkLocation;
                bestErrorMeters = error;
            }
        }

        if (here == null && isRunningOnEmulator()) {
            // We're on the emulator, and there's no emulated position.
            // Fake one.  This simplifies testing.
            here = new Location("johan");

            // An arbitrary spot in Stockholm
            here.setLatitude(59.4);
            here.setLongitude(17.8);
            here.setAccuracy(500);
            here.setTime(System.currentTimeMillis());
        }

        return here;
    }

    /**
     * Get the current GPS location.
     *
     * @return The current GPS location, or null if the GPS doesn't currently
     * know where it is.
     */
    public Location getCurrentGpsLocation() {
        Location location =
            locationManager.getLastKnownLocation(LocationManager.GPS_PROVIDER);
        if (location ==  null) {
            return null;
        }

        long locationAge = System.currentTimeMillis() - location.getTime();
        if (locationAge > MAX_LOCATION_AGE_MS) {
            return null;
        }

        return location;
    }

    /**
     * Is our current-location measurement good enough for us to show it to
     * the user?
     *
     * @return True if our calculated location is at most
     * {@link #MAXIMUM_ERROR_METERS} from our real position.  False otherwise.
     */
    public boolean isPrecisionOk() {
        return getCurrentLocation() != null;
    }

    /**
     * Request a refresh of the phone's most recently known position.  The
     * refresh is initiated by this method, but when or if it will finish is
     * undefined.
     *
     * @see #cancelRefresh()
     */
    public void refresh() {
        locationManager.requestLocationUpdates(
            LocationManager.GPS_PROVIDER, 0, 0, gpsRefreshListener);
        locationManager.requestLocationUpdates(
            LocationManager.NETWORK_PROVIDER, 0, 0, netRefreshListener);
    }

    /**
     * Cancels any outstanding refresh request.
     *
     * @see #refresh()
     */
    public void cancelRefresh() {
        locationManager.removeUpdates(netRefreshListener);
        locationManager.removeUpdates(gpsRefreshListener);
    }

    /**
     * Is the GPS enabled?
     *
     * @return True if the GPS is enabled.  False otherwise.
     */
    public boolean isGpsEnabled() {
        return locationManager.isProviderEnabled(LocationManager.GPS_PROVIDER);
    }

    /**
     * How many GPS satellites can we see?
     *
     * @return The number of GPS satellites visible to the phone.
     */
    public int getVisibleSatellites() {
        return gpsVisibleSatellites;
    }

    /**
     * How many GPS satellites are we using for positioning?
     *
     * @return The number of GPS satellites currently used for determining the
     * phone's location.
     */
    public int getUsedSatellites() {
        return gpsUsedSatellites;
    }

    /**
     * Use this if you know where you are.
     *
     * @param here Where you are.  Can be null for "don't know".
     */
    public void setLocation(Location here) {
        manualLocation = here;
    }
}
