/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl.location;

import java.util.Comparator;
import java.util.Date;

/**
 * A {@link NamedLocation} with a going-to score and a going-from score.
 *
 * @author johan.walles@gmail.com
 */
public class ScoredLocation {
    /**
     * Sort scored locations by from score, highest score first, most recently
     * used first.
     */
    public static final Comparator<ScoredLocation> FROM_COMPARATOR = new Comparator<ScoredLocation>()
    {
        public int compare(ScoredLocation a, ScoredLocation b) {
            Integer aScore = a.getFromScore();
            Integer bScore = b.getFromScore();
            int result = -aScore.compareTo(bScore);
            if (result != 0) {
                return result;
            }

            Date aUse = a.getFromLastUse();
            Date bUse = b.getFromLastUse();
            if (aUse == null && bUse != null) {
                return 1;
            }
            if (aUse != null && bUse == null) {
                return -1;
            }
            if (aUse != null && bUse != null) {
                result = -aUse.compareTo(bUse);
                if (result != 0) {
                    return result;
                }
            }

            if (result == 0) {
                result =
                    a.getLocation().getAlias().compareTo(b.getLocation().getAlias());
            }

            return result;
        }
    };

    /**
     * Sort scored locations by to score, highest score first.
     */
    public static final Comparator<ScoredLocation> TO_COMPARATOR = new Comparator<ScoredLocation>()
    {
        public int compare(ScoredLocation a, ScoredLocation b) {
            Integer aScore = a.getToScore();
            Integer bScore = b.getToScore();
            int result = -aScore.compareTo(bScore);
            if (result != 0) {
                return result;
            }

            Date aUse = a.getToLastUse();
            Date bUse = b.getToLastUse();
            if (aUse == null && bUse != null) {
                return 1;
            }
            if (aUse != null && bUse == null) {
                return -1;
            }
            if (aUse != null && bUse != null) {
                result = -aUse.compareTo(bUse);
                if (result != 0) {
                    return result;
                }
            }

            if (result == 0) {
                result =
                    a.getLocation().getAlias().compareTo(b.getLocation().getAlias());
            }

            return result;
        }
    };

    /**
     * How many times has this location been searched for as a from location?
     */
    private int fromScore;

    /**
     * How many times has this location been searched for as a to location?
     */
    private int toScore;

    /**
     * When was this location last used as a starting location?
     */
    private Date fromLastUse;

    /**
     * When was this location last used for destination?
     */
    private Date toLastUse;

    /**
     * The location we're keeping scores for.
     */
    private NamedLocation location;

    /**
     * Create a new scored location
     * @param location The location to keep score for
     * @param fromScore How many times has this location been searched for as
     * the start of a trip?
     * @param toScore How many times has this location been searched for as
     * the end of a trip?
     */
    public ScoredLocation(NamedLocation location, int fromScore, int toScore) {
        this.location = location;
        this.fromScore = fromScore;
        this.toScore = toScore;
    }

    /**
     * How many times has this location been searched for as the start of a
     * trip?
     * @return The number of times has this location been searched for
     * as the start of a trip
     */
    public int getFromScore() {
        return fromScore;
    }

    /**
     * Set the number of times this location has been searched for as the start
     * of a trip.
     * @param fromScore The number of times has this location been searched for
     * as the start of a trip
     */
    public void setFromScore(int fromScore) {
        this.fromScore = fromScore;
    }

    /**
     * How many times has this location been searched for as the end of a trip?
     * @return The number of times has this location been searched for
     * as the end of a trip
     */
    public int getToScore() {
        return toScore;
    }

    /**
     * Set the number of times this location has been searched for as the end
     * of a trip.
     * @param toScore The number of times has this location been searched for
     * as the end of a trip
     */
    public void setToScore(int toScore) {
        this.toScore = toScore;
    }

    /**
     * Fetch the location we're keeping score for.
     * @return the location we're keeping score for
     */
    public NamedLocation getLocation() {
        return location;
    }

    @Override
    public String toString() {
        return location.toString();
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result
            + ((fromLastUse == null) ? 0 : fromLastUse.hashCode());
        result = prime * result + fromScore;
        result = prime * result
            + ((location == null) ? 0 : location.hashCode());
        result = prime * result
            + ((toLastUse == null) ? 0 : toLastUse.hashCode());
        result = prime * result + toScore;
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof ScoredLocation)) {
            return false;
        }
        ScoredLocation other = (ScoredLocation) obj;
        if (fromLastUse == null) {
            if (other.fromLastUse != null) {
                return false;
            }
        } else if (!fromLastUse.equals(other.fromLastUse)) {
            return false;
        }
        if (fromScore != other.fromScore) {
            return false;
        }
        if (location == null) {
            if (other.location != null) {
                return false;
            }
        } else if (!location.equals(other.location)) {
            return false;
        }
        if (toLastUse == null) {
            if (other.toLastUse != null) {
                return false;
            }
        } else if (!toLastUse.equals(other.toLastUse)) {
            return false;
        }
        if (toScore != other.toScore) {
            return false;
        }
        return true;
    }

    /**
     * When was this location last used as a starting point?
     *
     * @return When this location was last used as a starting point, or null if
     * that never happened.
     */
    public Date getFromLastUse() {
        return fromLastUse;
    }

    /**
     * When was this location last used as a destination?
     *
     * @return When this location was last used for destination, or null if
     * that never happened.
     */
    public Date getToLastUse() {
        return toLastUse;
    }

    /**
     * When was this location last used, either as starting point or
     * destination?
     *
     * @return When this location was last used as either end of a trip, or
     * null if it has never been used.
     */
    public Date getLastUse() {
        if (fromLastUse == null && toLastUse == null) {
            return null;
        }
        if (fromLastUse == null) {
            return toLastUse;
        }
        if (toLastUse == null) {
            return fromLastUse;
        }
        if (fromLastUse.after(toLastUse)) {
            return fromLastUse;
        } else {
            return toLastUse;
        }
    }

    /**
     * Set when this location was last used as a starting point.
     *
     * @param lastUse When this location was last used as a starting point.
     */
    public void setFromLastUse(Date lastUse) {
        this.fromLastUse =  lastUse;
    }

    /**
     * Mark this location as having been used as a starting point right now.
     */
    public void setFromLastUse() {
        setFromLastUse(new Date());
    }

    /**
     * Set when this location was last used as a destination.
     *
     * @param lastUse When this location was last used as a destination.
     */
    public void setToLastUse(Date lastUse) {
        this.toLastUse =  lastUse;
    }

    /**
     * Mark this location as having been used as a destination right now.
     */
    public void setToLastUse() {
        setToLastUse(new Date());
    }
}
