/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.launchpad.sl.resultsdisplay;

import net.launchpad.sl.Query;
import net.launchpad.sl.R;
import net.launchpad.sl.RequestCodes;
import net.launchpad.sl.RequestKeys;
import net.launchpad.sl.location.PhoneLocator;
import android.app.TabActivity;
import android.content.Context;
import android.content.Intent;
import android.location.Location;
import android.location.LocationManager;
import android.os.Bundle;
import android.view.View;
import android.widget.Button;
import android.widget.TabHost;

/**
 * Display a query result, giving the user a choice of result display options.
 *
 * @author johan.walles@gmail.com
 */
public class DisplayResult extends TabActivity {
    /**
     * The location that we last used as a point to search from.
     */
    private Location location;

    /**
     * The query we're supposed to answer.
     */
    private Query query;

    /**
     * Used for finding the phone's current location.
     */
    private PhoneLocator phoneLocator;

    /**
     * Create an intent to display the answer to a given query.
     *
     * @param context The application context.
     *
     * @param query The query to display the answer for.
     */
    public static void answerQuery(Context context, Query query) {
        Intent intent = new Intent(context, DisplayResult.class);
        intent.putExtra(RequestKeys.QUERY, query);
        context.startActivity(intent);
    }

    /**
     * Do we need positioning services to answer this query?
     *
     * @return True if positioning services are needed.  False otherwise.
     */
    private boolean isPositioningNeeded() {
        return query.getFrom().isHere() || query.getTo().isHere();
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.display_result);

        Button goBack = (Button)findViewById(R.id.goBack);
        goBack.setOnClickListener(new View.OnClickListener() {
            public void onClick(View v) {
                finish();
            }
        });

        query = (Query)getIntent().getParcelableExtra(RequestKeys.QUERY);

        LocationManager locationManager =
            (LocationManager)getSystemService(Context.LOCATION_SERVICE);
        phoneLocator = new PhoneLocator(locationManager);

        phoneLocator.setLocation(location);
        if (isPositioningNeeded() && !phoneLocator.isPrecisionOk()) {
            // Positioning is needed and we don't know where we are
            PrecisionWarning.complain(this, query.getFrom().isHere());
            return;
        }

        displayResult(phoneLocator.getCurrentLocation());
    }

    /**
     * Update result display(s) to reflect results for a given location.
     *
     * @param currentLocation The location to use as the current location.
     */
    private void displayResult(Location currentLocation) {
        TabHost tabHost = getTabHost();
        tabHost.clearAllTabs();

        // Update the parsed display
        TabHost.TabSpec parsedDisplay = tabHost.newTabSpec("trips");
        parsedDisplay.setIndicator("Reseförslag");
        parsedDisplay.setContent(DisplayResultOverview.createDisplayIntent(
            this, query, currentLocation));
        tabHost.addTab(parsedDisplay);

        // Update the web display
        TabHost.TabSpec webDisplay = tabHost.newTabSpec("browser");
        webDisplay.setIndicator("www.sl.se");
        webDisplay.setContent(DisplayResultWeb.createDisplayIntent(
            this, query, currentLocation));
        tabHost.addTab(webDisplay);
    }

    @Override
    protected void onActivityResult(int requestCode, int resultCode, Intent data)
    {
        if (requestCode != RequestCodes.BAD_POSITIONING) {
            // We only know one thing, and this wasn't it.  Ignore it.
            return;
        }
        if (resultCode != RESULT_OK) {
            // User gave up on positioning, and without a position we can't do
            // anything.  Bail out.
            finish();
            return;
        }

        // User was happy with this location, display an answer for it
        location = data.getParcelableExtra(RequestKeys.LOCATION);
        displayResult(location);
    }
}
