/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.launchpad.sl.resultsdisplay;

import java.util.LinkedList;
import java.util.List;
import net.launchpad.sl.R;
import net.launchpad.sl.RequestKeys;
import net.launchpad.sl.location.PhoneLocator;
import net.launchpad.sl.location.PhoneLocator.PositioningListener;
import net.launchpad.sl.trip.SlMapData;
import net.launchpad.sl.trip.Trip;
import net.launchpad.sl.trip.TripException;
import net.launchpad.sl.trip.TripPart;
import android.app.AlertDialog;
import android.app.ProgressDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.graphics.Path;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.drawable.BitmapDrawable;
import android.location.Location;
import android.location.LocationManager;
import android.os.AsyncTask;
import android.os.Bundle;
import android.view.View;
import android.widget.Button;
import android.widget.RadioGroup;
import android.widget.TextView;
import com.google.android.maps.GeoPoint;
import com.google.android.maps.ItemizedOverlay;
import com.google.android.maps.MapActivity;
import com.google.android.maps.MapController;
import com.google.android.maps.MapView;
import com.google.android.maps.MyLocationOverlay;
import com.google.android.maps.OverlayItem;

/**
 * Display a map of either the starting point or the destination of a trip.
 *
 * @author johan.walles@gmail.com
 */
public class DisplayResultMap extends MapActivity {
    /**
     * The trip we're displaying maps for.
     */
    private Trip trip;

    /**
     * The map showing the start or end of the trip.
     */
    private MapView map;

    /**
     * Controller for our {@link #map}.
     */
    private MapController mapController;

    /**
     * Shows the best known location on the map.
     */
    private MyLocationOverlay myLocationOverlay;

    /**
     * Keeps track of where the phone is.
     */
    private PhoneLocator phoneLocator;

    /**
     * Map data for the start of the trip.
     */
    private SlMapData from;

    /**
     * Map data for the end of the trip.
     */
    private SlMapData to;

    /**
     * Downloads map data from reseplanerare.sl.se.
     */
    private class SlMapDownloader
    extends AsyncTask<Boolean, Integer, SlMapDownloader>
    {
        /**
         * The query result.
         */
        private SlMapData mapData;

        /**
         * If non-null, the query failed with this exception.
         */
        private TripException exception;

        /**
         * Shows a spinner while we're waiting for sl.se.
         */
        private ProgressDialog progress;

        @Override
        protected void onPostExecute(SlMapDownloader result) {
            if (progress.isShowing()) {
                progress.dismiss();
            }
            progress = null;

            if (exception != null) {
                showDialog("Kunde inte hämta kartdata",
                    exception.getMessage(),
                    true);

                return;
            }

            if (slOverlay != null) {
                map.getOverlays().remove(slOverlay);
            }
            slOverlay = new SlOverlay();
            map.getOverlays().add(slOverlay);

            showMap(result.mapData);
        }

        @Override
        protected SlMapDownloader doInBackground(Boolean... isStart) {
            try {
                if (isStart[0]) {
                    mapData.populate(trip.getStartMapUrl(), true);
                    from = mapData;
                } else {
                    mapData.populate(trip.getEndMapUrl(), false);
                    to = mapData;
                }
            } catch (TripException e) {
                exception = e;
            }

            return this;
        }

        @Override
        protected void onPreExecute() {
            mapData = new SlMapData(DisplayResultMap.this);
            progress =
                ProgressDialog.show(DisplayResultMap.this,
                    "Hämtar kartdata", "Hämtar kartinformation från reseplanerare.sl.se",
                    true, true, new DialogInterface.OnCancelListener()
                {
                    public void onCancel(DialogInterface dialog) {
                        mapData.interruptDownload();
                    }
                });
        }
    }

    /**
     * An overlay showing data about the trip start and end.
     */
    private class SlOverlay extends ItemizedOverlay<OverlayItem> {
        /**
         * Create a new SL overlay.
         */
        public SlOverlay() {
            super(null);

            populate();
        }

        /**
         * A bunch of locations, with a boolean that is true if this location is
         * an SL stop, false otherwise.
         */
        private List<OverlayItem> items;

        @Override
        protected OverlayItem createItem(int i) {
            return items.get(i);
        }

        /**
         * Create an overlay item for a titled location.
         *
         * @param location Where to put this overlay item.
         *
         * @param title The title of this overlay item.
         *
         * @return An overlay item matching the specifications.
         */
        private OverlayItem createItem(Location location, String title)
        {
            int latE6 = (int)(location.getLatitude() * 1e6);
            int lonE6 = (int)(location.getLongitude() * 1e6);

            OverlayItem item =
                new OverlayItem(new GeoPoint(latE6, lonE6), title, null);

            // Measure the title size in pixels
            Paint paint = new Paint();
            Rect textSize = new Rect();
            paint.getTextBounds(title, 0, title.length(), textSize);

            // Create a bitmap to draw on
            int arrowSize = 10;

            // Width of the black bubble outline
            int edgeWidth = 2;

            // The bubble's interior margin
            int mariginWidth = 4;
            Bitmap bitmap =
                Bitmap.createBitmap(
                    textSize.width() + 2 * edgeWidth + 2 * mariginWidth,
                    textSize.height() + 2 * edgeWidth + 2 * mariginWidth + arrowSize,
                    Bitmap.Config.ARGB_4444);
            paint.setAntiAlias(true);

            // Create the bubble outline
            int top = edgeWidth;
            int left = edgeWidth;
            int bottom = edgeWidth + 2 * mariginWidth + textSize.height();
            int right = edgeWidth + 2 * mariginWidth + textSize.width();
            int arrowLeft = bitmap.getWidth() / 3;
            int arrowRight = arrowLeft + arrowSize;
            int arrowMiddle = (arrowLeft + arrowRight) / 2;
            int arrowBottom = bottom + arrowSize;
            Path bubblePath = new Path();
            bubblePath.moveTo(left + mariginWidth, top);

            // Top + right arc
            bubblePath.arcTo(
                new RectF(right - mariginWidth, top, right, top + mariginWidth),
                270, 90);

            // Right + bottom arc
            bubblePath.arcTo(
                new RectF(right - mariginWidth, bottom - mariginWidth, right, bottom),
                0, 90);

            // Arrow
            bubblePath.lineTo(arrowRight, bottom);
            bubblePath.lineTo(arrowMiddle, arrowBottom);
            bubblePath.lineTo(arrowLeft, bottom);

            // Bottom + left arc
            bubblePath.arcTo(
                new RectF(left, bottom - mariginWidth, left + mariginWidth, bottom),
                90, 90);

            // Left + top arc
            bubblePath.arcTo(
                new RectF(left, top, left + mariginWidth, top + mariginWidth),
                180, 90);

            // Close the path
            bubblePath.close();

            // Draw a white background
            Canvas canvas = new Canvas(bitmap);
            paint.setColor(0xffffffff);
            paint.setStyle(Paint.Style.FILL);
            canvas.drawPath(bubblePath, paint);

            // Stroke the background edge with black
            paint.setColor(0xff000000);
            paint.setStyle(Paint.Style.STROKE);
            paint.setStrokeWidth(edgeWidth);
            canvas.drawPath(bubblePath, paint);

            // Draw the text on the background
            paint.setStrokeWidth(0);
            canvas.drawText(title,
                left + mariginWidth,
                left + mariginWidth - paint.getFontMetricsInt().ascent - 1,
                paint);

            // Align the marker so that the hot spot is at the pointer tip
            BitmapDrawable bubble = new BitmapDrawable(bitmap);
            int width = bitmap.getWidth();
            int height = bitmap.getHeight();
            int hotX = arrowMiddle;
            int hotY = bitmap.getHeight();
            bubble.setBounds(-hotX, -hotY, width - hotX, height - hotY);

            item.setMarker(bubble);

            return item;
        }

        @Override
        public int size() {
            if (items == null) {
                items = new LinkedList<OverlayItem>();

                if (from != null) {
                    TripPart firstPart = trip.getParts().get(0);
                    if (from.getStopLocation() != null) {
                        items.add(createItem(from.getStopLocation(),
                            firstPart.getToStop()));
                    }
                    if (from.getLocation() != null) {
                        if (!trip.getQuery().getFrom().isHere()) {
                            items.add(createItem(from.getLocation(),
                                firstPart.getFromStop()));
                        }
                    }
                }

                if (to != null) {
                    TripPart lastPart =
                        trip.getParts().get(trip.getParts().size() - 1);
                    if (to.getStopLocation() != null) {
                        items.add(createItem(to.getStopLocation(),
                            lastPart.getFromStop()));
                    }
                    if (to.getLocation() != null) {
                        if (!trip.getQuery().getTo().isHere()) {
                            items.add(createItem(to.getLocation(),
                                lastPart.getToStop()));
                        }
                    }
                }
            }

            return items.size();
        }
    }

    /**
     * Shows data about trip start and end.
     */
    private SlOverlay slOverlay;

    /**
     * Create an intent to display the details of a given trip.
     *
     * @param context The application context.
     *
     * @param trip The trip to show details for.
     *
     * @return An intent to show the details of the given trip.
     */
    public static Intent createDisplayIntent(Context context, Trip trip) {
        Intent intent = new Intent(context, DisplayResultMap.class);
        intent.putExtra(RequestKeys.TRIP, trip);

        return intent;
    }
    /**
     * Ask for a trouble dialog to be shown.
     *
     * @param title The title of the dialog.
     *
     * @param message The message displayed by the dialog.
     *
     * @param finishOnClick True to finish this activity on dialog
     * acceptance.  False to just dismiss the dialog itself.
     */
    private void showDialog(CharSequence title, CharSequence message,
        final boolean finishOnClick) {
        AlertDialog.Builder builder =
            new AlertDialog.Builder(this);
        builder.setTitle(title);
        builder.setMessage(message);
        builder.setIcon(android.R.drawable.ic_dialog_alert);

        builder.setPositiveButton("OK", new DialogInterface.OnClickListener() {
            public void onClick(DialogInterface dialog, int which) {
                dialog.dismiss();
                if (finishOnClick) {
                    finish();
                }
            }
        });

        builder.show();
    }

    /**
     * Show one end of the trip.
     *
     * @param mapData Describes the end of the trip that we should show.
     */
    private void showMap(SlMapData mapData) {
        // Position the map
        int stopLatE6 = (int)(mapData.getStopLocation().getLatitude() * 1e6);
        int stopLonE6 = (int)(mapData.getStopLocation().getLongitude() * 1e6);
        int latitudeSpan = 0;
        int longitudeSpan = 0;

        int centerLatE6 = stopLatE6;
        int centerLonE6 = stopLonE6;
        if (mapData.getLocation() != null) {
            int locLatE6 = (int)(mapData.getLocation().getLatitude() * 1e6);
            int locLonE6 = (int)(mapData.getLocation().getLongitude() * 1e6);

            centerLatE6 = (centerLatE6 + locLatE6) / 2;
            centerLonE6 = (centerLonE6 + locLonE6) / 2;

            latitudeSpan = Math.abs(stopLatE6 - locLatE6);
            longitudeSpan = Math.abs(stopLonE6 - locLonE6);
        }
        mapController.animateTo(new GeoPoint(centerLatE6, centerLonE6));

        int widthPixels = map.getWidth();
        int heightPixels = map.getHeight();
        int shortestPixels = Math.min(widthPixels, heightPixels);
        if (shortestPixels == 0) {
            // We don't know the display size, can't compute any zoom
            mapController.setZoom(MapUtil.DEFAULT_MAP_ZOOM);
            return;
        }

        // Zoom in enough to give us a ~ flat surface
        int zoom = 13;
        mapController.setZoom(zoom);

        // Measure the map
        float metersPerPixel = MapUtil.getMetersPerPixel(map);
        int shortestMeters = (int)(shortestPixels * metersPerPixel);
        int mapLatitudeSpan = map.getLatitudeSpan();
        int mapLongitudeSpan = map.getLongitudeSpan();

        // Find the right zoom level
        while (shortestMeters > 400) {
            if (mapLatitudeSpan < 2 * latitudeSpan) {
                // We've zoomed too far to be able to cover enough latitude
                break;
            }
            if (mapLongitudeSpan < 2 * longitudeSpan) {
                // We've zoomed too far to be able to cover enough longitude
                break;
            }

            zoom++;
            shortestMeters /= 2;
            mapLatitudeSpan /= 2;
            mapLongitudeSpan /= 2;
        }
        mapController.setZoom(zoom);
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        trip = (Trip)getIntent().getExtras().get(RequestKeys.TRIP);
        LocationManager locationManager =
            (LocationManager)getSystemService(LOCATION_SERVICE);
        phoneLocator = new PhoneLocator(locationManager);

        setContentView(R.layout.travel_map);
        map = (MapView)findViewById(R.id.tripMap);
        mapController = map.getController();
        TextView mapDescription = (TextView)findViewById(R.id.mapDescription);
        RadioGroup startEndButtons =
            (RadioGroup)findViewById(R.id.startEndRadio);
        Button quitButton = (Button)findViewById(R.id.showTimesButton);

        // Center the map
        mapController.setCenter(new GeoPoint(
            MapUtil.SL_CENTER_LATITUDE_E6, MapUtil.SL_CENTER_LONGITUDE_E6));
        mapController.setZoom(MapUtil.DEFAULT_MAP_ZOOM);

        // Configure the map
        myLocationOverlay = new MyLocationOverlay(this, map) {
            @Override
            public synchronized void onLocationChanged(Location location) {
                // Use the phone locator's opinion and nobody else's
                super.onLocationChanged(phoneLocator.getLastKnownLocation());
            }
        };
        map.getOverlays().add(myLocationOverlay);
        map.setBuiltInZoomControls(true);

        // Configure the buttons and the heading text
        mapDescription.setText("");
        quitButton.setOnClickListener(new View.OnClickListener() {
            public void onClick(View v) {
                finish();
            }
        });
        startEndButtons.setOnCheckedChangeListener(new RadioGroup.OnCheckedChangeListener() {
            public void onCheckedChanged(RadioGroup group, int checkedId) {
                if (checkedId == R.id.showStart) {
                    if (from == null) {
                        // From is already loading, and the map will be updated
                        // when it's done
                        return;
                    }
                    showMap(from);
                } else {
                    if (to != null) {
                        showMap(to);
                        return;
                    }
                    new SlMapDownloader().execute(false);
                }
            }
        });

        // Start loading map data for the trip's starting point.
        new SlMapDownloader().execute(true);
    }

    @Override
    protected boolean isRouteDisplayed() {
        return false;
    }

    @Override
    protected boolean isLocationDisplayed() {
        return true;
    }

    @Override
    protected void onPause() {
        super.onPause();

        phoneLocator.disableLocationListeners();
        myLocationOverlay.disableMyLocation();
        myLocationOverlay.disableCompass();
    }

    @Override
    protected void onResume() {
        super.onResume();

        myLocationOverlay.enableCompass();
        myLocationOverlay.enableMyLocation();
        phoneLocator.enableLocationListeners(new PositioningListener() {
            public void positioningUpdated() {
                myLocationOverlay.onLocationChanged(
                    phoneLocator.getLastKnownLocation());
            }
        });
    }
}
