/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.launchpad.sl.resultsdisplay;

import net.launchpad.sl.Query;
import net.launchpad.sl.R;
import net.launchpad.sl.RequestKeys;
import net.launchpad.sl.location.PhoneLocator;
import net.launchpad.sl.trip.Trip;
import net.launchpad.sl.trip.TripException;
import net.launchpad.sl.trip.TripList;
import android.app.AlertDialog;
import android.app.ListActivity;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.location.Location;
import android.location.LocationManager;
import android.os.Bundle;
import android.os.Handler;
import android.view.View;
import android.widget.AdapterView;

/**
 * Display an overview of the available trips.
 *
 * @author johan.walles@gmail.com
 */
public class DisplayResultOverview extends ListActivity {
    /**
     * Keeps track of our trips.
     */
    private TripListAdapter adapter;

    /**
     * Used for posting events to this thread.
     */
    private Handler handler = new Handler();

    /**
     * Displays a warning dialog.
     */
    private class DialogDisplayer implements Runnable {
        /**
         * The dialog title.
         */
        private CharSequence title;

        /**
         * The dialog message.
         */
        private CharSequence message;

        /**
         * True to finish this activity on dialog acceptance.  False to just
         * dismiss the dialog.
         */
        private boolean finishOnClick;

        /**
         * Ask for a trouble dialog to be shown.
         *
         * @param title The title of the dialog.
         *
         * @param message The message displayed by the dialog.
         *
         * @param finishOnClick True to finish this activity on dialog
         * acceptance.  False to just dismiss the dialog itself.
         */
        public DialogDisplayer(CharSequence title, CharSequence message,
            final boolean finishOnClick)
        {
            this.title = title;
            this.message = message;
            this.finishOnClick = finishOnClick;
        }

        public void run() {
            AlertDialog.Builder builder =
                new AlertDialog.Builder(DisplayResultOverview.this);
            builder.setTitle(title);
            builder.setMessage(message);
            builder.setIcon(android.R.drawable.ic_dialog_alert);

            builder.setPositiveButton("OK", new DialogInterface.OnClickListener() {
                public void onClick(DialogInterface dialog, int which) {
                    dialog.dismiss();
                    if (finishOnClick) {
                        finish();
                    }
                }
            });

            builder.show();
        }
    }

    /**
     * Create an intent to display the answer to a given query.
     *
     * @param context The application context.
     *
     * @param query The query to display the answer for.
     *
     * @param location An optional location to use as the current location.
     *
     * @return An intent to show the result for the given query.
     */
    public static Intent createDisplayIntent(Context context,
        Query query, Location location)
    {
        Intent intent = new Intent(context, DisplayResultOverview.class);
        intent.putExtra(RequestKeys.QUERY, query);
        if (location != null) {
            intent.putExtra(RequestKeys.LOCATION, location);
        }

        return intent;
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        setTitle(R.string.app_long_name);

        final Query query =
            (Query)getIntent().getParcelableExtra(RequestKeys.QUERY);

        LocationManager locationManager =
            (LocationManager)getSystemService(LOCATION_SERVICE);
        PhoneLocator phoneLocator = new PhoneLocator(locationManager);

        Location location =
            (Location)getIntent().getParcelableExtra(RequestKeys.LOCATION);
        phoneLocator.setLocation(location);

        adapter = new TripListAdapter(this);
        setListAdapter(adapter);

        TripListLoader.loadTripList(this, query,
            query.toUriString(location, true),
            new TripListLoader.Receiver()
        {
            @Override
            public void failure(TripException e) {
                showDialog("Sökningen misslyckades", e.getMessage(), true);
            }

            @Override
            public void done(TripList tripList) {
                adapter.setTripList(tripList);
            }
        });

        getListView().setOnItemClickListener(new AdapterView.OnItemClickListener() {
            public void onItemClick(AdapterView<?> parent,
                View view, int position, long id)
            {
                final Trip trip = adapter.getItem(position);
                if (trip == null) {
                    String url = null;
                    if (position == 0) {
                        // The user wants to see earlier trips
                        url = adapter.getTripList().getGoEarlierUrl();
                    } else if (position == adapter.getCount() - 1) {
                        // The user wants to see later trips
                        url = adapter.getTripList().getGoLaterUrl();
                    }

                    TripListLoader.loadTripList(DisplayResultOverview.this,
                        query,
                        url, new TripListLoader.Receiver() {
                            @Override
                            public void failure(TripException e) {
                                showDialog("Hittade inga fler resor", e.getMessage(), false);
                            }

                            @Override
                            public void done(TripList tripList) {
                                adapter.setTripList(tripList);
                            }
                    });

                    return;
                }

                if (trip.getParts() != null) {
                    showDetails(trip);
                    return;
                }

                TripListLoader.loadTripList(DisplayResultOverview.this, query,
                    adapter.getTripList().getGetDetailsUrl(),
                    new TripListLoader.Receiver() {
                        @Override
                        public void failure(TripException e) {
                            showDialog("Hittade inga detaljer", e.getMessage(), true);
                        }

                        @Override
                        public void done(TripList tripList) {
                            adapter.setTripList(tripList);
                            Trip tripWithDetails =
                                tripList.getById(trip.getId());
                            showDetails(tripWithDetails);
                        }
                });
            }

            private void showDetails(Trip trip) {
                Intent detailsIntent =
                    DisplayResultDetails.createDisplayIntent(
                        DisplayResultOverview.this, trip);
                startActivity(detailsIntent);
            }
        });
    }

    /**
     * Show a trouble dialog.
     *
     * @param title The title of the dialog.
     *
     * @param message The message displayed by the dialog.
     *
     * @param finishOnClick True to finish this activity on dialog acceptance.
     * False to just dismiss the dialog itself.
     */
    private void showDialog(CharSequence title, CharSequence message,
        final boolean finishOnClick)
    {
        handler.post(new DialogDisplayer(title, message, finishOnClick));
    }
}
