/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.launchpad.sl.resultsdisplay;

import net.launchpad.sl.CreateLocationActivity;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.location.Location;
import android.net.Uri;

import com.google.android.maps.GeoPoint;
import com.google.android.maps.MapView;

/**
 * Utility functions for dealing with {@link MapView}.
 *
 * @author johan.walles@gmail.com
 */
public class MapUtil {
    /**
     * If we don't know for some reason, zoom the map to this level.
     */
    public final static int DEFAULT_MAP_ZOOM = 8;

    /**
     * The center latitude of the SL area.
     */
    public static final int SL_CENTER_LATITUDE_E6 =
        (int)(1e6
            * (CreateLocationActivity.SL_LOWER_LEFT_LAT
                + CreateLocationActivity.SL_UPPER_RIGHT_LAT) / 2.0);
    /**
     * The center longitude of the SL area.
     */
    public static final int SL_CENTER_LONGITUDE_E6 =
        (int)(1e6
            * (CreateLocationActivity.SL_LOWER_LEFT_LON
            + CreateLocationActivity.SL_UPPER_RIGHT_LON) / 2.0);

    /**
     * Compute the number of meters per pixel at the map's current zoom level.
     * The return method from this method is only useful if the map is zoomed in
     * enough that the area it covers can be considered flat.
     *
     * @param map The map to compute the number of meters per pixel for.
     *
     * @return The number of meters per pixel for the map along the X axis.
     */
    public static float getMetersPerPixel(MapView map) {
        // Find a location for the map center
        GeoPoint mapCenterPoint = map.getMapCenter();
        Location centerLocation = getLocation(mapCenterPoint);

        int longitudeSpanE6 = map.getLongitudeSpan();
        int edgeLongitudeE6 =
            mapCenterPoint.getLongitudeE6() + longitudeSpanE6 / 2;
        Location edgeLocation = new Location(centerLocation);
        edgeLocation.setLongitude(edgeLongitudeE6 / 1e6f);

        float metersToEdge = centerLocation.distanceTo(edgeLocation);
        float pixelsToEdge = map.getWidth() / 2f;
        float metersPerPixel = metersToEdge / pixelsToEdge;
        return metersPerPixel;
    }

    /**
     * Convert a GeoPoint to a Location.
     *
     * @param geoPoint The point to convert to a Location
     *
     * @return A location describing the same place as the point
     */
    public static Location getLocation(GeoPoint geoPoint) {
        Location location = new Location("johan");
        location.setTime(System.currentTimeMillis());
        location.setLatitude(geoPoint.getLatitudeE6() / 1e6f);
        location.setLongitude(geoPoint.getLongitudeE6() / 1e6f);
        return location;
    }

    /**
     * Launch Google Maps showing a walking route between two coordinates.
     * @param context An OS context, needed for launching Google Maps
     * @param from Where you are going from
     * @param to The destination
     */
    public static void showRoute(Context context, Location from, Location to) {
        Intent intent = new Intent();
        String uriString =
            "http://maps.google.com/maps?saddr="
            + from.getLatitude() + "," + from.getLongitude()
            + "&daddr="
            + to.getLatitude() + "," + to.getLongitude()
            + "&dirflg=w";
        intent.setData(Uri.parse(uriString));
        intent.setAction(Intent.ACTION_VIEW);
        intent.setComponent(new ComponentName("com.google.android.apps.maps",
            "com.google.android.maps.MapsActivity"));

        // FIXME: We need to do something more to show the map by default,
        // currently the default is to show the text representation of the
        // walking directions.
        //
        // Get inspired here: http://mapki.com/wiki/Google_Map_Parameters

        context.startActivity(intent);
    }
}
