/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl.resultsdisplay;

import net.launchpad.sl.CalendarUtil;
import net.launchpad.sl.R;
import net.launchpad.sl.RequestCodes;
import net.launchpad.sl.RequestKeys;
import net.launchpad.sl.location.PhoneLocator;

import com.google.android.maps.GeoPoint;
import com.google.android.maps.MapActivity;
import com.google.android.maps.MapController;
import com.google.android.maps.MapView;
import com.google.android.maps.MyLocationOverlay;

import android.app.Activity;
import android.content.Intent;
import android.location.Location;
import android.location.LocationManager;
import android.os.Bundle;
import android.provider.Settings;
import android.view.View;
import android.widget.Button;
import android.widget.CompoundButton;
import android.widget.TextView;
import android.widget.ToggleButton;
import android.widget.CompoundButton.OnCheckedChangeListener;

/**
 * Warn the user that the locationing precision is bad.
 *
 * @author johan.walles@gmail.com
 */
public class PrecisionWarning extends MapActivity {
    /**
     * This tells us where we are.
     */
    private PhoneLocator phoneLocator;

    /**
     * The enable / disable GPS toggle button.
     */
    private ToggleButton gpsButton;

    /**
     * The GPS status text next to the {@link #gpsButton}.
     */
    private TextView gpsText;

    /**
     * The search-anyway button.
     */
    private Button searchAnyway;

    /**
     * This button makes the map go back to showing the last known location,
     * ignoring any panning done by the user.
     */
    private Button resetMap;

    /**
     * The map showing our current positioning precision.
     */
    private MapView map;

    /**
     * Controller for our {@link #map}.
     */
    private MapController mapController;

    /**
     * Shows the best known location on the map.
     */
    private MyLocationOverlay myLocationOverlay;

    /**
     * The location that we last displayed on the map.
     */
    private Location lastShownLocation;

    /**
     * True if we're complaining about the starting location, false if we're
     * complaining about the destination.
     */
    private boolean aboutFromLocation;

    /**
     * Show a given location on map, zoomed to a level matching the location's
     * accuracy.
     *
     * @param location The location to show on the map.
     */
    private void showOnMap(Location location) {
        if (location == null) {
            mapController.setCenter(new GeoPoint(
                MapUtil.SL_CENTER_LATITUDE_E6, MapUtil.SL_CENTER_LONGITUDE_E6));
            mapController.setZoom(MapUtil.DEFAULT_MAP_ZOOM);

            lastShownLocation = null;
            return;
        }

        if (lastShownLocation != null) {
            if (lastShownLocation.getLatitude() == location.getLatitude()
                && lastShownLocation.getLongitude() == location.getLongitude()
                && lastShownLocation.getAccuracy() == location.getAccuracy())
            {
                // We're already showing this location, do nothing
                return;
            }
        }

        myLocationOverlay.onLocationChanged(phoneLocator.getLastKnownLocation());

        mapController.animateTo(new GeoPoint(
            (int)(location.getLatitude() * 1e6),
            (int)(location.getLongitude() * 1e6)));

        // Zoom in to the current accuracy...
        float errorMeters = location.getAccuracy();

        // We should optimize for the shortest of width and height
        int lengthPixels;
        if (map.getHeight() < map.getWidth()) {
            lengthPixels = map.getHeight();
        } else {
            lengthPixels = map.getWidth();
        }
        if (lengthPixels == 0) {
            // We don't know the display size, can't compute any zoom
            mapController.setZoom(MapUtil.DEFAULT_MAP_ZOOM);
            return;
        }

        // We want to cover errorMeters * 2 in lengthPixels
        float wantedMetersPerPixel = (2.0f * errorMeters) / lengthPixels;

        // Zoom in enough to give us a ~ flat surface
        int zoom = 13;
        mapController.setZoom(zoom);

        float metersPerPixel = MapUtil.getMetersPerPixel(map);

        // Find the zoom level we want
        while (metersPerPixel >= wantedMetersPerPixel) {
            zoom++;
            metersPerPixel /= 2.0f;
        }

        // If Location.getAccuracy() returned a radius we'd have zoomed in one
        // step too far here and should zoom out one notch.  However, Google's
        // MyLocationOverlay uses the accuracy as a diameter, and they should
        // know, so we'll leave the zoom at this value.
        //
        // Bug report requesting clarification:
        // http://code.google.com/p/android/issues/detail?id=4531

        // Actually zoom the map
        mapController.setZoom(zoom);

        lastShownLocation = new Location(location);
    }

    /**
     * Update the information in the UI.
     */
    private void updateUi() {
        if (phoneLocator.isPrecisionOk()) {
            // Precision has become OK, we're done
            returnResult(phoneLocator.getLastKnownLocation());
            return;
        }

        showOnMap(phoneLocator.getLastKnownLocation());

        final boolean gpsEnabled = phoneLocator.isGpsEnabled();
        gpsButton.setChecked(gpsEnabled);

        Location gpsLocation = null;
        if (gpsEnabled) {
            gpsLocation = phoneLocator.getCurrentGpsLocation();
        }

        int visibleSatellites = phoneLocator.getVisibleSatellites();
        int usedSatellites = phoneLocator.getUsedSatellites();
        if (!gpsEnabled) {
            gpsText.setText("Av, tryck på knappen för att slå på");
        } else if (usedSatellites > 0) {
            if (gpsLocation != null) {
                gpsText.setText(String.format(
                    "%dm osäkerhet, %d satelliter synliga varav %d används för "
                    + "positionering.",
                    (int)gpsLocation.getAccuracy(),
                    visibleSatellites, usedSatellites));
            } else {
                gpsText.setText(String.format("%d satelliter synliga varav %d " +
                		"används för positionering.",
                		visibleSatellites, usedSatellites));
            }
        } else {
            if (gpsLocation != null) {
                gpsText.setText(String.format(
                    "%dm osäkerhet, %d satelliter synliga, se till att telefonen "
                    + "kan se mer av himlen.",
                    (int)gpsLocation.getAccuracy(), visibleSatellites));
            } else {
                gpsText.setText(String.format(
                    "%d satelliter synliga, se till att telefonen kan se mer av himlen.",
                    visibleSatellites));
            }
        }

        TextView lastFixDescription =
            (TextView)findViewById(R.id.lastFixDescription);
        Location bestLocation = phoneLocator.getLastKnownLocation();
        if (bestLocation == null) {
            lastFixDescription.setVisibility(View.GONE);
        } else {
            lastFixDescription.setText(String.format(
                "Senast kända position är %s och hade %dm felmarginal.",
                CalendarUtil.describeTime(bestLocation.getTime()),
                (int)bestLocation.getAccuracy()));
            lastFixDescription.setVisibility(View.VISIBLE);
        }
    }

    /**
     * Complain about that positioning is bad.
     *
     * @param launcher The activity that wants to complain about positioning.
     *
     * @param from True if we're complaining about the departure location.
     * False if we're complaining about the destination.
     */
    public static void complain(Activity launcher, boolean from) {
        Intent intent = new Intent(launcher, PrecisionWarning.class);
        intent.putExtra(RequestKeys.ABOUT_FROM_LOCATION, from);
        launcher.startActivityForResult(intent, RequestCodes.BAD_POSITIONING);
    }

    @Override
    protected void onResume() {
        super.onResume();

        phoneLocator.enableLocationListeners(new PhoneLocator.PositioningListener() {
            public void positioningUpdated() {
                updateUi();
            }
        });
        myLocationOverlay.enableMyLocation();

        updateUi();
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        LocationManager locationManager =
            (LocationManager)getSystemService(LOCATION_SERVICE);
        phoneLocator = new PhoneLocator(locationManager);
        aboutFromLocation =
            getIntent().getBooleanExtra(RequestKeys.ABOUT_FROM_LOCATION, true);

        setContentView(R.layout.bad_positioning_warning);
        gpsButton = (ToggleButton)findViewById(R.id.gpsButton);
        gpsText = (TextView)findViewById(R.id.gpsText);
        searchAnyway = (Button)findViewById(R.id.searchAnywayButton);
        resetMap = (Button)findViewById(R.id.resetMap);
        map = (MapView)findViewById(R.id.badPositioningMap);
        mapController = map.getController();

        // Toggling the GPS button should take us to the GPS control system
        // preferences.  Applications are not allowed to toggle the GPS state by
        // themselves.
        gpsButton.setOnCheckedChangeListener(new OnCheckedChangeListener() {
            public void onCheckedChanged(CompoundButton button, boolean isChecked) {
                if (isChecked != phoneLocator.isGpsEnabled()) {
                    startActivity(new Intent(Settings.ACTION_LOCATION_SOURCE_SETTINGS));
                }
            }
        });

        if (aboutFromLocation) {
            searchAnyway.setText("Sök resa från kartans mitt");
        } else {
            searchAnyway.setText("Sök resa till kartans mitt");
        }
        searchAnyway.setOnClickListener(new View.OnClickListener() {
            public void onClick(View v) {
                returnResult(MapUtil.getLocation(map.getMapCenter()));
            }
        });

        resetMap.setOnClickListener(new View.OnClickListener() {
            public void onClick(View v) {
                // This will make us re-position the map to the last known
                // location.
                lastShownLocation = null;
                showOnMap(phoneLocator.getLastKnownLocation());
            }
        });

        myLocationOverlay = new MyLocationOverlay(this, map) {
            @Override
            public synchronized void onLocationChanged(Location location) {
                // Use the phone locator's opinion and nobody else's
                super.onLocationChanged(phoneLocator.getLastKnownLocation());
            }
        };
        map.getOverlays().add(new CenterCrossOverlay());
        map.getOverlays().add(myLocationOverlay);
        map.setBuiltInZoomControls(true);
        mapController.setZoom(MapUtil.DEFAULT_MAP_ZOOM);
        mapController.setCenter(
            new GeoPoint(MapUtil.SL_CENTER_LATITUDE_E6, MapUtil.SL_CENTER_LONGITUDE_E6));
    }

    @Override
    protected void onPause() {
        super.onPause();

        myLocationOverlay.disableMyLocation();
        phoneLocator.disableLocationListeners();
    }

    @Override
    protected boolean isRouteDisplayed() {
        return false;
    }

    @Override
    protected boolean isLocationDisplayed() {
        // Displaying locations is what we do
        return true;
    }

    /**
     * Return a location to our caller.
     *
     * @param location The location to return to our caller.
     */
    private void returnResult(Location location) {
        if (location == null) {
            location = MapUtil.getLocation(map.getMapCenter());
        }
        Intent data = new Intent();
        data.putExtra(RequestKeys.LOCATION, location);
        setResult(Activity.RESULT_OK, data);
        finish();
    }
}
