/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.launchpad.sl.resultsdisplay;

import net.launchpad.sl.trip.Trip;
import net.launchpad.sl.trip.TripList;
import android.content.Context;
import android.graphics.Typeface;
import android.os.Handler;
import android.view.Gravity;
import android.view.View;
import android.view.ViewGroup;
import android.widget.BaseAdapter;
import android.widget.LinearLayout;
import android.widget.ListView;
import android.widget.TextView;

/**
 * Used for displaying a {@link TripList} in a {@link ListView}.
 *
 * @author johan.walles@gmail.com
 */
public class TripListAdapter extends BaseAdapter {
    /**
     * This is the trip list we're displaying.  Accesses to this field must be
     * synchronized.
     */
    private TripList tripList;

    /**
     * Our OS context, used for creating views for the list.
     */
    private Context context;

    /**
     * Handles asynchronous results in the UI thread.
     */
    private Handler handler = new Handler();

    /**
     * Replaces the current trip list with a new one.
     */
    private class ListReplacer implements Runnable {
        /**
         * The replacement trip list.
         */
        private TripList newTripList;

        /**
         * Create a new trip list replacer.
         *
         * @param newTripList The replacement trip list.
         */
        public ListReplacer(TripList newTripList) {
            this.newTripList = newTripList;
        }

        public void run() {
            tripList = newTripList;
            notifyDataSetChanged();
        }
    }

    /**
     * Construct a new adapter showing an empty list.
     *
     * @param context Used for creating views for the list entries.
     *
     * @see #setTripList(TripList)
     */
    public TripListAdapter(Context context) {
        this.context = context;
    }

    /**
     * Build the trip view.
     *
     * @return A view that can display summary information about one trip.
     */
    private View createInfoView() {
        // Top linear layout
        TwoPlusOneLayout top = new TwoPlusOneLayout(context);

        // Left column
        LinearLayout leftColumn = new LinearLayout(context);
        leftColumn.setOrientation(LinearLayout.VERTICAL);
        leftColumn.setId(4);
        top.addView(leftColumn);

        // Departure time + station
        TextView text = new TextView(context);
        text.setTypeface(Typeface.DEFAULT_BOLD);
        text.setId(1);
        leftColumn.addView(text);

        // Arrival time + station
        text = new TextView(context);
        text.setId(2);
        text.setGravity(Gravity.RIGHT);
        top.addView(text);

        // Transport metadata
        text = new TextView(context);
        text.setId(3);
        leftColumn.addView(text);

        return top;
    }

    public int getCount() {
        if (tripList == null) {
            return 0;
        }

        return tripList.size() + 2;
    }

    public Trip getItem(int position) {
        if (position == 0) {
            // This is the "go earlier" item
            return null;
        }
        if (position == getCount() - 1) {
            // This is the "go later" item
            return null;
        }

        return tripList.get(position - 1);
    }

    public long getItemId(int position) {
        if (position == 0) {
            return -1;
        }
        if (position == getCount() - 1) {
            // This is the "go later" item
            return -2;
        }

        return getItem(position).getId();
    }

    public View getView(int position, View view, ViewGroup parent) {
        if (view == null) {
            view = createInfoView();
        }

        TextView fromView = (TextView)view.findViewById(1);
        TextView toView = (TextView)view.findViewById(2);
        TextView metaView = (TextView)view.findViewById(3);

        if (position == 0) {
            fromView.setText("Åk tidigare...");
            toView.setText("");
            metaView.setText("");
            return view;
        }
        if (position == getCount() - 1) {
            fromView.setText("Åk senare...");
            toView.setText("");
            metaView.setText("");
            return view;
        }

        Trip trip = getItem(position);

        fromView.setText(trip.getDepartureTime()
            + " från " + trip.getFirstStop());

        toView.setText("Till "
            + trip.getLastStop() + " " + trip.getArrivalTime());

        String timeString = null;
        int tripMinutes = trip.getDurationMinutes();
        if (tripMinutes < 60) {
            timeString = tripMinutes + "min";
        } else {
            int hours = tripMinutes / 60;
            int minutes = tripMinutes % 60;
            if (minutes != 0) {
                timeString = String.format("%dh%02dm", hours, minutes);
            } else {
                timeString = hours + "h";
            }
        }

        int changes = trip.getChanges();
        String changesText;
        if (changes == 0) {
            changesText = "inga byten";
        } else if (changes == 1) {
            changesText = "ett byte";
        } else {
            changesText = changes + " byten";
        }

        metaView.setText(timeString + ", " + changesText);

        return view;
    }

    /**
     * Set the trip list displayed by this adapter.
     *
     * @param tripList The trip list to display.
     */
    public void setTripList(TripList tripList) {
        handler.post(new ListReplacer(tripList));
    }

    /**
     * Get the trip list displayed by this adapter.
     *
     * @return The trip list displayed by this adapter.
     */
    public TripList getTripList() {
        return tripList;
    }
}
