/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.launchpad.sl.resultsdisplay;

import android.app.ProgressDialog;
import android.content.Context;
import android.content.DialogInterface;
import net.launchpad.sl.Query;
import net.launchpad.sl.trip.TripException;
import net.launchpad.sl.trip.TripList;

/**
 * Used for loading a {@link TripList} in the background while showing a
 * {@link ProgressDialog}.
 *
 * @author johan.walles@gmail.com
 */
public class TripListLoader {
    /**
     * Callback for receiving a loaded {@link TripList}.
     *
     * @author johan.walles@gmail.com
     */
    public static abstract class Receiver {
        /**
         * The {@link TripList} has finished loading, here it is!
         *
         * @param tripList The trip list that we just finished loading.
         */
        public abstract void done(TripList tripList);

        /**
         * An error occurred while loading the trip list.
         *
         * @param e The error that occurred.
         */
        public abstract void failure(TripException e);
    }

    /**
     * Load a {@link TripList} in the background, displaying a
     * {@link ProgressDialog} while loading.
     *
     * @param context Used for displaying the {@link ProgressDialog}.
     *
     * @param query The query for which we're loading the answer.
     *
     * @param url The URL to load the {@link TripList} from.
     *
     * @param receiver Will be notified when the TripList has been loaded.
     */
    public static void loadTripList(Context context, Query query,
        final String url, final Receiver receiver)
    {
        final TripList tripList = new TripList(context, query);
        final ProgressDialog progress =
            ProgressDialog.show(context,
                "Hämtar information", "Hämtar information från reseplanerare.sl.se",
                true, true, new DialogInterface.OnCancelListener() {
                    public void onCancel(DialogInterface dialog) {
                        tripList.interruptDownload();
                    }
                });
        new Thread(new Runnable() {
            public void run() {
                try {
                    tripList.populate(url);
                    progress.dismiss();
                    receiver.done(tripList);
                } catch (TripException e) {
                    progress.dismiss();
                    receiver.failure(e);
                }
            }
        }).start();
    }
}
