/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.launchpad.sl.trip;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.launchpad.sl.SL;
import net.launchpad.sl.StreamUtil;

import android.content.Context;
import android.location.Location;
import android.util.Log;

/**
 * Extracts coordinates for a stop and a location from a reseplanerare.sl.se
 * map.
 *
 * @author johan.walles@gmail.com
 */
public class SlMapData {
    /**
     * The last parsed HTML is always stored in this file for debugging
     * purposes.
     */
    private final static String LAST_PARSED = "last-parsed-map.html";

    /**
     * Should match the following line:<br/>
     * var db = ((18042090-18024120)*(716/358)-(18042090-18024120))/2;
     * <p>
     * Will extract right longitude, left longitude and pixel width or
     * top latitude, bottom latitude and pixel height.
     */
    private final static Pattern LXITUDE_DATA =
        Pattern.compile("^var d([bh]) = \\(\\(([0-9]+)-([0-9]+)\\)\\*\\([0-9]+/([0-9]+)\\).*$");

    /**
     * Should match the following line:
     * &lt;area shape="circle" coords="281,314,10"
     */
    private final static Pattern PIXEL_COORDINATES =
        Pattern.compile("^<area shape=\"circle\" coords=\"([0-9]+),([0-9]+),.*$");

    /**
     * The start or destination location.
     */
    private Location location;

    /**
     * The location of the SL stop.
     */
    private Location stopLocation;

    /**
     * Our OS context.
     */
    private Context context;

    /**
     * The stream from which we're reading our map data.
     */
    private InputStream mapDataSource;

    /**
     * Have we been interrupted by {@link #interruptDownload()}?
     */
    private boolean interrupted = false;

    /**
     * Create a map data holder.  Must be {@link #populate(String, boolean)}d
     * before it can be used.
     *
     * @param context An OS context
     */
    public SlMapData(Context context) {
        this.context = context;
    }

    /**
     * Interrupt an in-progress download from sl.se.
     */
    public void interruptDownload() {
        synchronized (this) {
            if (mapDataSource == null) {
                return;
            }

            interrupted = true;
            try {
                mapDataSource.close();
            } catch (IOException e) {
                // This is a best effort operation, errors are silently ignored
                Log.w(SL.TAG, "Interrupting map data download failed", e);
            }
        }
    }

    /**
     * Read map data from an input stream.
     * <p>
     * This method has default protection for testing purposes.
     *
     * @param html A HTML stream containing a map page from sl.se
     *
     * @param startingPoint true if this is the trip's starting point.  False
     * if this is the trip's endpoint.
     *
     * @throws IOException if accessing the map data fails
     *
     * @throws TripException if there's a problem with getting map data
     */
    public void populate(InputStream html, boolean startingPoint)
    throws TripException, IOException
    {
        synchronized (this) {
            mapDataSource = html;
        }
        try {
            StreamUtil.download(context, html, LAST_PARSED);
            parseHtml(startingPoint);
        } catch (IOException e) {
            if (interrupted) {
                throw new IOException("Nedladdningen avbröts");
            } else {
                throw e;
            }
        } finally {
            synchronized (this) {
                mapDataSource = null;
            }
        }
    }

    /**
     * Fetch map data from sl.se.
     *
     * @param url Map data will be loaded from this URL
     *
     * @param startingPoint true if this is the trip's starting point.  False
     * if this is the trip's endpoint.
     *
     * @throws TripException if there's a problem with getting map data
     *
     * @see Trip#getStartMapUrl()
     * @see Trip#getEndMapUrl()
     */
    public void populate(String url, boolean startingPoint)
    throws TripException {
        try {
            populate(new URL(url).openStream(), startingPoint);
        } catch (IOException e) {
            throw new TripException("Misslyckades med att hämta kartdata från sl.se: "
                + e.getMessage(), e);
        }
    }

    /**
     * Parse map information from sl.se.
     *
     * @param startingPoint true if this is the trip's starting point.  False
     * if this is the trip's endpoint.
     *
     * @throws TripException if there's a problem with getting map data
     *
     * @throws IOException if there's a problem reading map data
     */
    private void parseHtml(boolean startingPoint)
    throws TripException, IOException {
        BufferedReader input = null;
        try {
            // 8kB == 2 blocks (probably) == a reasonable buffer size for
            // reading files of about 32kB
            input = new BufferedReader(new InputStreamReader(
                context.openFileInput(LAST_PARSED)), 8192);

            // Map dimensions
            int leftLon = 0;
            int rightLon = 0;
            int topLat = 0;
            int bottomLat = 0;
            int width = 0;
            int height = 0;
            int nCoordinates = 0;

            // Point pixel coordinates
            int x[] = new int[2];
            int y[] = new int[2];

            for (String line = input.readLine(); line != null; line = input.readLine()) {
                Matcher matcher = LXITUDE_DATA.matcher(line);
                if (matcher.matches()) {
                    boolean isLongitudeAndWidth = matcher.group(1).equals("b");
                    if (isLongitudeAndWidth) {
                        rightLon = Integer.parseInt(matcher.group(2));
                        leftLon = Integer.parseInt(matcher.group(3));
                        width = Integer.parseInt(matcher.group(4));
                    } else {
                        topLat = Integer.parseInt(matcher.group(2));
                        bottomLat = Integer.parseInt(matcher.group(3));
                        height = Integer.parseInt(matcher.group(4));
                    }
                }

                matcher = PIXEL_COORDINATES.matcher(line);
                if (matcher.matches()) {
                    x[nCoordinates] = Integer.parseInt(matcher.group(1));
                    y[nCoordinates] = Integer.parseInt(matcher.group(2));
                    nCoordinates++;

                    if (nCoordinates >= 2) {
                        // Both coordinates found, we're done!
                        break;
                    }
                }
            }

            switch (nCoordinates) {
            case 0:
                throw new TripException("Hittade inga platser på kartan, försök igen");

            case 1:
                stopLocation =
                    toLocation(leftLon, topLat, rightLon, bottomLat,
                        x[0], y[0], width, height);
                break;

            case 2:
                // If this is the trip start, the stop location comes last...
                int locationIndex = 0;
                int stopLocationIndex = 1;
                if (!startingPoint) {
                    // ... otherwise the stop location comes first.
                    locationIndex = 1;
                    stopLocationIndex = 0;
                }

                location =
                    toLocation(leftLon, topLat, rightLon, bottomLat,
                        x[locationIndex], y[locationIndex], width, height);
                stopLocation =
                    toLocation(leftLon, topLat, rightLon, bottomLat,
                        x[stopLocationIndex], y[stopLocationIndex],
                        width, height);
                break;
            }
        } finally {
            try {
                if (input != null) {
                    input.close();
                    input = null;
                }
            } catch (IOException e) {
                // Closing the stream is a best-effort operation, so we
                // intentionally ignore any trouble here.
            }
        }
    }

    /**
     * Compute a Location from map data.
     *
     * @param leftLon The left longitude of the map.
     * @param topLat The top latitude of the map.
     * @param rightLon The right longitude of the map.
     * @param bottomLat The bottom latitude of the map.
     * @param x The x coordinate of the location.
     * @param y The y coordinate of the location.
     * @param width The width of the map picture.
     * @param height The height of the map picture.
     *
     * @return A Location corresponding to the location marked on the map.
     */
    private Location toLocation(int leftLon, int topLat,
        int rightLon, int bottomLat,
        int x, int y, int width, int height)
    {
        double xFraction = ((double)x) / (width - 1);
        double yFraction = ((double)y) / (height - 1);

        int widthLon = rightLon - leftLon;
        int heightLat = topLat - bottomLat;

        double lat = (topLat - (heightLat * yFraction)) / 1e6;
        double lon = (leftLon + (widthLon * xFraction)) / 1e6;

        Location result = new Location("johan");
        result.setLatitude(lat);
        result.setLongitude(lon);
        return result;
    }

    /**
     * Returns the position of the trip start or end point.
     *
     * @return The position of the trip start or end point.
     */
    public Location getLocation() {
        return location;
    }

    /**
     * Returns the position of the stop.
     *
     * @return The stop position from the map.
     */
    public Location getStopLocation() {
        return stopLocation;
    }
}
