/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl.trip;

import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.launchpad.sl.Query;

import android.os.Parcel;
import android.os.Parcelable;

/**
 * A trip that you can take with SL.  This is just a summary of the trip, if
 * you want details you have to fetch its parts.
 *
 * @author johan.walles@gmail.com
 */
public class Trip implements Parcelable {
    /**
     * Parser for extracting the numbers from a trip ID string (like "C1-1").
     */
    private final static Pattern IDSTRING =
        Pattern.compile("^C([0-9]+)-([0-9]+)$");

    /**
     * Required by implementors of {@link Parcelable}.
     */
    public static final Parcelable.Creator<Trip> CREATOR =
        new Parcelable.Creator<Trip>()
    {
        public Trip createFromParcel(Parcel source) {
            Query query = source.readParcelable(getClass().getClassLoader());
            String pArrivalTime = source.readString();
            String pBaseUrl = source.readString();
            int pChanges = source.readInt();
            String pDepartureTime = source.readString();
            int pDurationMinutes = source.readInt();
            String pFirstStop = source.readString();
            String pIdString = source.readString();
            String pLastStop = source.readString();

            List<TripPart> pParts = new LinkedList<TripPart>();
            source.readTypedList(pParts, TripPart.CREATOR);

            List<String> pTransports = new LinkedList<String>();
            source.readList(pTransports, null);

            Trip trip =
                new Trip(query,
                    pFirstStop, pLastStop,
                    pDepartureTime, pArrivalTime,
                    pDurationMinutes, pChanges,
                    pTransports,
                    pIdString, pBaseUrl);
            trip.setParts(pParts);

            return trip;
        }

        public Trip[] newArray(int size) {
            return new Trip[size];
        }
    };

    /**
     * The query that this trip is the answer to.
     */
    private Query query;

    /**
     * The trip starts at this stop.
     */
    private String firstStop;

    /**
     * The trip ends at this stop.
     */
    private String lastStop;

    /**
     * The departure time in the form "07:22".
     */
    private String departureTime;

    /**
     * The arrival time in the form "13:43".
     */
    private String arrivalTime;

    /**
     * The duration of the trip in minutes.
     */
    private int durationMinutes;

    /**
     * The number of times you have to change your means of travel during this
     * trip.
     */
    private int changes;

    /**
     * The different means of travel involved in this trip.
     */
    private List<String> transports;

    /**
     * Details for the different parts of this trip.
     */
    private List<TripPart> parts;

    /**
     * A unique identifier for this trip.
     */
    private int id;

    /**
     * Used for constructing map URLs.
     */
    private String idString;

    /**
     * Used for constructing map URLs.
     */
    private String baseUrl;

    /**
     * Remove cruft from a time string.
     *
     * @param time A time description from the big bad Internet.
     *
     * @return A time string on the format "07:33".
     */
    private String trimTime(String time) {
        if (time == null) {
            return "";
        }
        switch (time.length()) {
        case 0:
            return time;
        case 4:
            return "0" + time;
        case 5:
            return time;
        default:
            throw new IllegalArgumentException("Unable to parse time string: " + time);
        }
    }

    /**
     * Create a new trip.
     *
     * @param query the query that this trip is the answer to
     *
     * @param from Where the trip starts
     *
     * @param to Where the trip ends
     *
     * @param departureTime When the trip starts in the form "07:40".  null or
     * the empty string means unknown.
     *
     * @param arrivalTime When the trip ends in the form "07:40".  null or
     * the empty string means unknown.
     *
     * @param durationMinutes The duration of this trip in minutes.
     *
     * @param changes The number of changes involved in this trip.
     *
     * @param transports The different kinds of transports involved in this trip.
     *
     * @param idString a string uniquely identifying this trip, on a format that
     * can be matched by {@link #IDSTRING}.
     *
     * @param baseUrl is the base URL string of the form from which this trip
     * was read.
     */
    public Trip(Query query,
        String from, String to,
        String departureTime, String arrivalTime,
        int durationMinutes,
        int changes,
        List<String> transports,
        String idString,
        String baseUrl)
    {
        this.query = query;

        this.firstStop = from;
        this.lastStop = to;

        this.departureTime = trimTime(departureTime);
        this.arrivalTime = trimTime(arrivalTime);
        this.durationMinutes = durationMinutes;

        this.changes = changes;

        this.transports = transports;

        // Prepare for creating map URLs
        this.idString = idString;
        this.baseUrl = baseUrl;

        // Convert the ID string into an ID number
        Matcher matcher = IDSTRING.matcher(idString);
        if (!matcher.matches()) {
            throw new IllegalArgumentException("ID string doesn't match "
                + IDSTRING.pattern() + ": " + idString);
        }
        int n1 = Integer.valueOf(matcher.group(1));
        int n2 = Integer.valueOf(matcher.group(2));
        id = (n1 << 16) + n2;
    }

    /**
     * Return a unique identifier for this trip.
     *
     * @return A unique identifier for this trip.
     */
    public int getId() {
        return id;
    }

    /**
     * Get the name of the stop where this trip starts.
     * @return The name of the stop where this trip starts.
     */
    public String getFirstStop() {
        return firstStop;
    }

    /**
     * Get the name of the stop where this trip ends.
     * @return The name of the stop where this trip ends.
     */
    public String getLastStop() {
        return lastStop;
    }

    /**
     * Get the trip departure time.
     *
     * @return The trip departure time in the form "06:30".
     */
    public String getDepartureTime() {
        return departureTime;
    }

    /**
     * Get the trip arrival time.
     *
     * @return The trip arrival time in the form "18:39".
     */
    public String getArrivalTime() {
        return arrivalTime;
    }

    /**
     * Get the trip duration in minutes.
     *
     * @return The trip duration in minutes.
     */
    public int getDurationMinutes() {
        return durationMinutes;
    }

    /**
     * Get the number of times you have to change your means of travel during
     * this trip.
     *
     * @return The number of times you have to change your means of travel
     * during this trip.
     */
    public int getChanges() {
        return changes;
    }

    /**
     * Get the different kinds of transports involved in this trip.
     *
     * @return The different kinds of transports involved in this trip.
     */
    public List<String> getTransports() {
        return transports;
    }

    /**
     * Get a list of all the parts of this trip.
     *
     * @return A list of all the parts of this trip.
     */
    public List<TripPart> getParts() {
        return parts;
    }

    /**
     * Get the query that this trip is one answer to.
     *
     * @return The query that this trip is one answer to.
     */
    public Query getQuery() {
        return query;
    }

    /**
     * Get an URL that will give you a map of where this trip is starting.
     *
     * @return an URL that will give you a map of where this trip is starting.
     */
    public String getStartMapUrl() {
        return (getBaseMapUrl()
            + "&REQMapConnectionMarkerTransferLocation=STATION"
            + "&REQMapConnectionSection=START"
            + "&SetGlobalOptionGO_activeMapSlider=START"
            + "&REQMapConnectionMarkerFirstLocation=ADDRESS"
            + "&REQMapConnectionMarkerLastLocation=STATION");
    }

    /**
     * Get an URL that will give you a map of where this trip ends.
     *
     * @return an URL that will give you a map of where this trip ends.
     */
    public String getEndMapUrl() {
        return (getBaseMapUrl()
            + "&REQMapConnectionMarkerTransferLocation=STATION"
            + "&REQMapConnectionSection=END"
            + "&SetGlobalOptionGO_activeMapSlider=END"
            + "&REQMapConnectionMarkerFirstLocation=STATION"
            + "&REQMapConnectionMarkerLastLocation=ADDRESS");
    }

    /**
     * Get an URL on which we can build map URLs for this trip.
     *
     * @return an URL on which we can build map URLs for this trip.
     */
    private String getBaseMapUrl() {
        String baseMapUrl = baseUrl
                    + "&performGis=1"
                    + "&REQMapConnectionId=" + idString
                    + "&REQMapWidth=358"
                    + "&REQMapHeight=358"
                    + "&application=DIRECTIONS"
                    + "&guiVCtrl_connection_directions_onlyElement_" + idString
                    + "&SetGlobalOptionGO_callMapFromPosition=tpDetailsRouteComplete"
                    + "&REQMapConnectionLocationNameHandling=2";
        return baseMapUrl;
    }

    /**
     * Describe the parts that make up this trip.
     *
     * @param parts The parts that make up this trip.
     */
    public void setParts(List<TripPart> parts) {
        if (this.parts != null) {
            throw new IllegalStateException("Trip parts already set");
        }
        this.parts = parts;
    }

    @Override
    public String toString() {
        StringBuilder builder = new StringBuilder();
        builder.append(String.format("From %s at %s to %s at %s in %d minutes, %d changes and using:",
            firstStop, departureTime, lastStop, arrivalTime,
            durationMinutes, changes));
        for (String transport : transports) {
            builder.append(transport);
        }
        return builder.toString();
    }

    public int describeContents() {
        return 0;
    }

    public void writeToParcel(Parcel dest, int flags) {
        dest.writeParcelable(query, 0);
        dest.writeString(arrivalTime);
        dest.writeString(baseUrl);
        dest.writeInt(changes);
        dest.writeString(departureTime);
        dest.writeInt(durationMinutes);
        dest.writeString(firstStop);
        dest.writeString(idString);
        dest.writeString(lastStop);
        dest.writeTypedList(parts);
        dest.writeList(transports);
    }
}
