/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl.trip;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import android.content.Context;
import android.util.Log;

import net.launchpad.sl.Query;
import net.launchpad.sl.SL;
import net.launchpad.sl.StreamUtil;

/**
 * A list of trips.  This information represents the answer to a {@link Query},
 * and is basically an interpretation of "http://reseplanerare.sl.se".
 *
 * @author johan.walles@gmail.com
 */
public class TripList implements Iterable<Trip> {
    /**
     * The last parsed HTML is always stored in this file for debugging
     * purposes.
     */
    private final static String LAST_PARSED = "last-parsed.html";

    /**
     * Marks the beginning of a trip description.
     */
    private final static Pattern TRIP_START =
        Pattern.compile("^.*name=\"(guiVCtrl_connection_detailsOut_select_([^\"]*))\".*$");

    /**
     * Marks the end of a trip description.
     */
    private final static Pattern TRIP_END =
        Pattern.compile("^</tr>$");

    /**
     * Describes where a trip starts and ends.
     */
    private final static Pattern TRIP_FROM_TO =
        Pattern.compile("^<td headers=\"header3\">(.*)<br />(.*)</td><td align=\"right\" headers=\"header4\">$");

    /**
     * Find a time string.
     */
    private final static Pattern TIME =
        Pattern.compile("^[^0-9]*([0-2]?[0-9]:[0-5][0-9]).*$");

    /**
     * A line break.
     */
    private final static Pattern BR = Pattern.compile("<br */>");

    /**
     * A HTML tag.
     */
    private final static Pattern HTML_TAG = Pattern.compile("<[^>]*>");

    /**
     * Marks the start of a detailed trip description.
     */
    private final static Pattern DETAILS_START =
        Pattern.compile("^.*hafasDetailTable.*summary.*$");

    /**
     * Recognize the form posting URL.
     */
    private final static Pattern FORM_URL =
        Pattern.compile("^<form name=\"tp_results_form\" action=\"(http://reseplanerare.sl.se/bin/query.exe/.*)&OK#focus\" method=\"post\" style=\"display:inline\">$");

    /**
     * Our application context, for file system access.
     */
    private Context context;

    /**
     * The query this trip list is the answer to.
     */
    private Query query;

    /**
     * Our list of suggested trips.
     */
    private List<Trip> trips = new LinkedList<Trip>();

    /**
     * These trips are checked, used to match the trip overviews with trip
     * details.  The string starts with an &amp;.
     */
    private List<Trip> checkedTrips = new LinkedList<Trip>();

    /**
     * The base URL for {@link #getGetDetailsUrl()}, {@link #getGoEarlierUrl()}
     * and {@link #getGoLaterUrl()}.
     */
    private String baseUrl;

    /**
     * Add this to the URL to request details for all trips currently visible
     * on this page.
     */
    // 150 = room for three checkboxes which is the default query result.  They
    // are about 45 chars each.
    private StringBuilder allCheckboxes = new StringBuilder(150);

    /**
     * Trip data is read from this stream.
     */
    private InputStream inputStream;

    /**
     * Has our HTML download been interrupted?
     */
    private boolean downloadInterrupted = false;

    /**
     * Decode åäöÅÄÖ from http://sl.se.
     *
     * @param html A string from http://sl.se
     *
     * @return The same string but with correct åäöÅÄÖ
     */
    private String decode(String html) {
        html = html.replace("&nbsp;", " ");

        // Reference: http://demo.nickname.net/demo/testpak/encode.pl
        html = html.replace("&#229;", "å");
        html = html.replace("&#228;", "ä");
        html = html.replace("&#246;", "ö");
        html = html.replace("&#197;", "Å");
        html = html.replace("&#196;", "Ä");
        html = html.replace("&#214;", "Ö");
        html = html.replace("&#233;", "é");

        return html;
    }

    /**
     * Trim a time like 07:45 from a string like "ca&nbsp;07:40<br />".
     *
     * @param timeWithCruft A string with a time somewhere in it.
     *
     * @return The time string.
     */
    private String trimTime(String timeWithCruft) {
        if ("&nbsp;".equals(timeWithCruft)) {
            return "";
        }

        Matcher matcher = TIME.matcher(timeWithCruft);
        if (!matcher.matches()) {
            throw new IllegalArgumentException("String contains no time: " + timeWithCruft);
        }
        String time = matcher.group(1);
        if (time.length() == 5) {
            return time;
        } else {
            return "0" + time;
        }
    }

    /**
     * Parse trip information from sl.se.
     *
     * @param html The HTML to parse
     *
     * @param id The ID of this trip, "C1-5" for example
     *
     * @return The trip information
     *
     * @throws IOException if reading the HTML stream fails
     */
    private Trip parseTrip(BufferedReader html, String id) throws IOException {
        String from = null;
        String to = null;
        String departureTime = null;
        String arrivalTime = null;
        int durationMinutes = -1;
        int changes = -1;
        List<String> transports = null;
        for (String line = html.readLine(); line != null; line = html.readLine()) {
            if (TRIP_END.matcher(line).matches()) {
                throw new IOException("Ofullständig resebeskrivning");
            }

            if (from == null) {
                Matcher fromTo = TRIP_FROM_TO.matcher(line);
                if (fromTo.matches()) {
                    from = fromTo.group(1);
                    to = fromTo.group(2);
                }
            } else if (departureTime == null) {
                departureTime = trimTime(line);
            } else if (arrivalTime == null) {
                arrivalTime = trimTime(line);
            } else if (durationMinutes == -1) {
                Matcher time = TIME.matcher(line);
                if (time.matches()) {
                    StringTokenizer tokenizer = new StringTokenizer(line, ":");
                    String hourString = tokenizer.nextToken();
                    String minuteString = tokenizer.nextToken();
                    durationMinutes =
                        Integer.valueOf(minuteString)
                        + 60 * Integer.valueOf(hourString);
                }
            } else if (changes == -1) {
                try {
                    changes = Integer.valueOf(line);
                } catch (NumberFormatException e) {
                    // Nope, this wasn't the number of changes.  Ignore this
                    // line and hope for the number of changes to appear on the
                    // next one.
                }
            } else if (transports == null) {
                if (BR.matcher(line).find()) {
                    transports = new LinkedList<String>();
                    for (String transport : BR.split(line)) {
                        transports.add(decode(transport.trim()));
                    }

                    // The list of transports is the last thing we're looking
                    // for, so now we're done!
                    return new Trip(query,
                        decode(from), decode(to),
                        departureTime, arrivalTime, durationMinutes,
                        changes, transports, id, baseUrl);
                }
            }
        }
        throw new IOException("Sidan tog slut mitt i en resebeskrivning");
    }

    /**
     * Parse a trip part from http://reseplanerare.sl.se
     *
     * @param html The HTML code to parse
     *
     * @return A trip part parsed from the given HTML code
     *
     * @throws IOException if the HTML code cannot be accessed
     */
    private TripPart parseTripPart(BufferedReader html) throws IOException {
        String from = null;
        String to = null;
        String fromTime = null;
        String toTime = null;
        String transport = null;
        String details = null;
        int rowEnds = 0;
        boolean toSeen = false;
        for (String line = html.readLine(); line != null; line = html.readLine()) {
            if ("</tr>".equals(line)) {
                rowEnds++;
                if (rowEnds >= 2) {
                    return new TripPart(from, to, fromTime, toTime, transport, details);
                }
            }

            if (from == null) {
                if (!HTML_TAG.matcher(line).find()) {
                    from = decode(line);
                }
            } else if (fromTime == null) {
                if (!HTML_TAG.matcher(line).find()) {
                    fromTime = trimTime(line);
                }
            } else if (transport == null) {
                if (!line.startsWith("<")) {
                    transport = line;
                    int firstTagIndex = line.indexOf('<');
                    if (firstTagIndex >= 0) {
                        transport = line.substring(0, firstTagIndex);
                    }
                    transport = decode(transport);
                }
            } else if (details == null) {
                if (!HTML_TAG.matcher(line).find()) {
                    details = decode(line);
                }
            } else if (!toSeen) {
                if ("Till".equals(line)) {
                    toSeen = true;
                }
            } else if (to == null) {
                if (!HTML_TAG.matcher(line).find()) {
                    to = decode(line);
                }
            } else if (toTime == null) {
                if (!HTML_TAG.matcher(line).find()) {
                    toTime = trimTime(line);
                }
            }
        }

        throw new IOException("Sidan tog slut mitt i en detaljbeskrivning");
    }

    /**
     * Parse trip details from http://sl.se
     *
     * @param html HTML code to parse
     *
     * @return A list with the parts that make up this trip
     *
     * @throws IOException if the HTML code cannot be accessed
     */
    private List<TripPart> parseDetails(BufferedReader html)
    throws IOException
    {
        List<TripPart> parts = new LinkedList<TripPart>();
        for (String line = html.readLine(); line != null; line = html.readLine()) {
            if ("Fr&#229;n".equals(line)) {
                parts.add(parseTripPart(html));
            }
            if ("</table>".equals(line)) {
                return parts;
            }
        }
        throw new IOException("Sidan tog slut mitt ibland detaljbeskrivningarna");
    }

    /**
     * Parse an error message from http://reseplanerare.sl.se
     *
     * @param html The HTML code to parse
     * @return An error exception from the HTML code
     * @throws IOException If reading the error message fails
     */
    private TripException parseFailure(BufferedReader html) throws IOException {
        for (String line = html.readLine(); line != null; line = html.readLine()) {
            if (!HTML_TAG.matcher(line).find()) {
                String message = decode(line);
                int dotIndex = message.indexOf('.');
                if (dotIndex > 0) {
                    message = message.substring(0, dotIndex + 1);
                }
                return new TripException(message);
            }
        }
        throw new IOException("Sidan tog slut mitt i ett felmeddelande");
    }

    /**
     * Construct an empty trip list.
     * <p>
     * Call {@link #populate(String)} to populate the trip list.
     *
     * @param context An OS context.
     *
     * @param query The query this trip list is the answer to.
     */
    public TripList(Context context, Query query) {
        this.context = context;
        this.query = query;
    }

    /**
     * Load a trip list from a given SL URL.
     *
     * @param urlString An SL URL. Probably from {@link #getGetDetailsUrl()},
     * {@link #getGoEarlierUrl()} or {@link #getGoLaterUrl()}.
     *
     * @throws TripException if a route couldn't be found
     */
    public void populate(String urlString) throws TripException
    {
        try {
            // Download the HTML result so we have it locally in case of trouble
            populate(new URL(urlString).openStream());
        } catch (IOException e) {
            throw new TripException("Misslyckades med att få svar från reseplaneraren: "
                + e.getLocalizedMessage(), e);
        }
    }

    /**
     * Populate a trip list from a HTML stream coming from
     * "http://reseplanerare.sl.se".
     * <p>
     * Has default protection for testing purposes, use
     * {@link #populate(String)} together with
     * {@link Query#toUriString(android.location.Location, boolean)} if you want
     * a query answered.
     *
     * @param html A HTML input stream with information from
     * "http://reseplanerare.sl.se".  The stream will be closed by this method.
     *
     * @throws TripException if the travel planner reports an error
     *
     * @throws IOException on IO trouble
     */
    void populate(InputStream html)
    throws TripException, IOException
    {
        synchronized (this) {
            inputStream = html;
        }

        try {
            long t0 = System.currentTimeMillis();

            StreamUtil.download(context, html, LAST_PARSED);
            long t1 = System.currentTimeMillis();

            parseHtml();
            long t2 = System.currentTimeMillis();

            Log.d(SL.TAG, "Downloaded web page in " + (t1 - t0) + "ms");
            Log.d(SL.TAG, "Parsed web page in " + (t2 - t1) + "ms");
        } catch (IOException e) {
            if (downloadInterrupted) {
                IOException interrupted =
                    new IOException("Nedladdningen avbröts");
                interrupted.initCause(e);
                throw interrupted;
            }
        } finally {
            inputStream = null;
        }
    }

    /**
     * Interrupt an in-progress download.
     */
    public void interruptDownload() {
        synchronized (this) {
            if (inputStream != null) {
                try {
                    inputStream.close();
                } catch (IOException e) {
                    Log.w(SL.TAG, "Download interrupting failed", e);
                }
                downloadInterrupted = true;
            }
        }
    }

    /**
     * Initialize ourselves from {@link #LAST_PARSED}.
     *
     * @throws TripException if the travel planner reports an error.
     *
     * @throws IOException on trouble accessing the travel planner HTML
     */
    private void parseHtml() throws TripException, IOException {
        BufferedReader input = null;
        try {
            // 8kB == 2 blocks (probably) == a reasonable buffer size for
            // reading files of about 20kB and up
            input = new BufferedReader(new InputStreamReader(
                context.openFileInput(LAST_PARSED)), 8192);

            Iterator<Trip> checkedIterator = null;
            for (String line = input.readLine(); line != null; line = input.readLine()) {
                Matcher baseUrlMatcher = FORM_URL.matcher(line);
                if (baseUrlMatcher.matches()) {
                    baseUrl = baseUrlMatcher.group(1);
                    continue;
                }

                Matcher tripStartMatcher = TRIP_START.matcher(line);
                if (tripStartMatcher.matches()) {
                    allCheckboxes.append('&');
                    allCheckboxes.append(tripStartMatcher.group(1));
                    allCheckboxes.append("=yes");
                    Trip trip = parseTrip(input, tripStartMatcher.group(2));
                    trips.add(trip);
                    if (line.contains("checked")) {
                        checkedTrips.add(trip);
                    }
                } else if (DETAILS_START.matcher(line).matches()) {
                    if (checkedIterator == null) {
                        checkedIterator = checkedTrips.iterator();
                    }
                    checkedIterator.next().setParts(parseDetails(input));
                } else if ("<div class=\"ErrorBoxWide\">".equals(line)) {
                    throw parseFailure(input);
                }
            }
        } finally {
            try {
                if (input != null) {
                    input.close();
                    input = null;
                }
            } catch (IOException e) {
                // Closing the stream is a best-effort operation, so we
                // intentionally ignore any trouble here.
            }
        }

        if (trips.size() == 0) {
            throw new TripException("Hittade inga resor!");
        }
    }

    public Iterator<Trip> iterator() {
        return trips.iterator();
    }

    /**
     * Return the number of trips in the list.
     *
     * @return The number of trips in the list.
     */
    public int size() {
        return trips.size();
    }

    /**
     * Get a trip from the list.
     *
     * @param index Which trip to retrieve.
     *
     * @return The trip at the given index.
     *
     * @throws IndexOutOfBoundsException If the index is less than 0 or greater
     * than or equal to {@link #size()}.
     */
    public Trip get(int index) throws IndexOutOfBoundsException {
        return trips.get(index);
    }

    /**
     * Retrieve a trip by its ID.
     *
     * @param id The trip ID.
     *
     * @return The trip with the given ID, or null if no trip with that ID could
     * be found.
     */
    public Trip getById(int id) {
        for (Trip trip : trips) {
            if (trip.getId() == id) {
                return trip;
            }
        }

        return null;
    }

    /**
     * Get an URL that will give us some more trips before the current ones.
     *
     * @return an URL that will give us some more trips before the current ones.
     */
    public String getGoEarlierUrl() {
        return getGetDetailsUrl() + "&REQ0HafasScrollDir=2";
    }

    /**
     * Get an URL that will give us some more trips after the current ones.
     *
     * @return an URL that will give us some more trips after the current ones.
     */
    public String getGoLaterUrl() {
        return getGetDetailsUrl() + "&REQ0HafasScrollDir=1";
    }

    /**
     * Get an URL that will add details for all trips on this page.
     *
     * @return an URL that will add details for all trips on this page.
     */
    public String getGetDetailsUrl() {
        return (baseUrl
            + allCheckboxes
            + "&guiVCtrl_connection_detailsOut_add_selection=yes");
    }
}
