/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl.trip;

import android.os.Parcel;
import android.os.Parcelable;

/**
 * This is one part of a trip, like taking a certain subway line from one
 * station to another.
 *
 * @author johan.walles@gmail.com
 */
public class TripPart implements Parcelable {
    /**
     * Our departure stop.
     */
    private String from;

    /**
     * Our arrival stop.
     */
    private String to;

    /**
     * Our departure time in the form "07:44".
     */
    private String fromTime;

    /**
     * Our arrival time in the form "18:15".
     */
    private String toTime;

    /**
     * What kind of transport is used for this part of the trip.
     */
    private String transport;

    /**
     * Details about this part of the trip.
     */
    private String details;

    /**
     * Required by the {@link Parcelable} interface.
     */
    public static final Parcelable.Creator<TripPart> CREATOR =
        new Parcelable.Creator<TripPart>()
    {
        public TripPart createFromParcel(Parcel source) {
            String from = source.readString();
            String fromTime = source.readString();

            String to = source.readString();
            String toTime = source.readString();

            String transport = source.readString();
            String details = source.readString();

            return new TripPart(from, to, fromTime, toTime, transport, details);
        }

        public TripPart[] newArray(int size) {
            return new TripPart[size];
        }
    };

    /**
     * Create a new trip part.
     *
     * @param from The departure stop.
     * @param to The arrival stop.
     * @param fromTime The departure time.
     * @param toTime The arrival time.
     * @param transport The transport for this part of the trip.
     * @param details Details about this part of the trip.
     */
    public TripPart(String from, String to,
        String fromTime, String toTime,
        String transport, String details)
    {
        this.from = from;
        this.fromTime = fromTime;

        this.to = to;
        this.toTime = toTime;

        this.transport = transport;
        this.details = details;
    }

    /**
     * At what stop does this trip part start?
     * @return The name of the starting stop for this trip part.
     */
    public String getFromStop() {
        return from;
    }

    /**
     * When does this transport leave?
     *
     * @return The departure time, in the form "18:36", or "" if the departure
     * time is unknown.
     */
    public String getFromTime() {
        return fromTime;
    }

    /**
     * At what stop does this trip part end?
     * @return The name of the destination stop for this trip part.
     */
    public String getToStop() {
        return to;
    }

    /**
     * When does this transport arrive?
     *
     * @return The arrival time, in the form "07:33", or "" if the arrival time
     * is unknown.
     */
    public String getToTime() {
        return toTime;
    }

    /**
     * What kind of transport are we using ("subway line 5").
     *
     * @return A description of the transport used for this part of the trip.
     */
    public String getTransport() {
        return transport;
    }

    /**
     * Details about our transport, like which direction it is going or how far
     * we'll have to walk.
     *
     * @return Transport details.
     */
    public String getTransportDetails() {
        return details;
    }

    /**
     * Combine the results of {@link #getTransport()} and
     * {@link #getTransportDetails()} into a single descriptive string.
     *
     * @return A string describing the means for doing this part of the trip.
     */
    public CharSequence describeTransport() {
        StringBuilder description = new StringBuilder(30);

        description.append(Character.toUpperCase(transport.charAt(0)));
        description.append(transport.substring(1));

        description.append(' ');

        String censoredDetails = details.replace(":", "");
        description.append(Character.toLowerCase(censoredDetails.charAt(0)));
        description.append(censoredDetails.substring(1));

        return description;
    }

    @Override
    public String toString() {
        return String.format("From %s at %s to %s at %s using %s (%s)",
            from, fromTime, to, toTime, transport, details);
    }

    public int describeContents() {
        return 0;
    }

    public void writeToParcel(Parcel dest, int flags) {
        dest.writeString(from);
        dest.writeString(fromTime);

        dest.writeString(to);
        dest.writeString(toTime);

        dest.writeString(transport);
        dest.writeString(details);
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((details == null) ? 0 : details.hashCode());
        result = prime * result + ((from == null) ? 0 : from.hashCode());
        result = prime * result
            + ((fromTime == null) ? 0 : fromTime.hashCode());
        result = prime * result + ((to == null) ? 0 : to.hashCode());
        result = prime * result + ((toTime == null) ? 0 : toTime.hashCode());
        result = prime * result
            + ((transport == null) ? 0 : transport.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        TripPart other = (TripPart) obj;
        if (details == null) {
            if (other.details != null) {
                return false;
            }
        } else if (!details.equals(other.details)) {
            return false;
        }
        if (from == null) {
            if (other.from != null) {
                return false;
            }
        } else if (!from.equals(other.from)) {
            return false;
        }
        if (fromTime == null) {
            if (other.fromTime != null) {
                return false;
            }
        } else if (!fromTime.equals(other.fromTime)) {
            return false;
        }
        if (to == null) {
            if (other.to != null) {
                return false;
            }
        } else if (!to.equals(other.to)) {
            return false;
        }
        if (toTime == null) {
            if (other.toTime != null) {
                return false;
            }
        } else if (!toTime.equals(other.toTime)) {
            return false;
        }
        if (transport == null) {
            if (other.transport != null) {
                return false;
            }
        } else if (!transport.equals(other.transport)) {
            return false;
        }
        return true;
    }
}
