/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.launchpad.sl;

import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.regex.Pattern;

import junit.framework.TestCase;

/**
 * Verify the functionality of the SL class.
 *
 * @author johan.walles@gmail.com
 */
public class TestCalendarUtil extends TestCase {
    /**
     * Assert that a string matches a given regexp.
     *
     * @param regexp The regexp that the string should match.
     *
     * @param shouldMatch The string that should match the regexp.
     */
    private void assertMatches(String regexp, CharSequence shouldMatch) {
        if (shouldMatch == null) {
            fail("String must not be null");
        }
        if (Pattern.matches(regexp, shouldMatch)) {
            return;
        }

        fail("<" + shouldMatch + "> should have matched <" + regexp + ">");
    }

    /**
     * Verify that date formatting works.
     */
    public void testFormatDate() {
        Calendar date = Calendar.getInstance();
        CharSequence todayString = CalendarUtil.formatDate(date);

        date.add(Calendar.DAY_OF_MONTH, 1);
        CharSequence tomorrowString = CalendarUtil.formatDate(date);

        date.add(Calendar.DAY_OF_MONTH, 1);
        CharSequence dayAfterTomorrowString = CalendarUtil.formatDate(date);

        date.add(Calendar.DAY_OF_MONTH, 1);
        CharSequence threeDaysFromNowString = CalendarUtil.formatDate(date);

        date.add(Calendar.YEAR, 1);
        CharSequence nextYearString = CalendarUtil.formatDate(date);

        assertMatches("^I dag [^ ]+ [0-9]+ [^ ]+$", todayString);
        assertMatches("^I morgon [^ ]+ [0-9]+ [^ ]+( 20[0-9][0-9])?$", tomorrowString);
        assertMatches("^I övermorgon [^ ]+ [0-9]+ [^ ]+( 20[0-9][0-9])?$", dayAfterTomorrowString);
        assertMatches("^[^ ]+ [0-9]+ [^ ]+( 20[0-9][0-9])?$", threeDaysFromNowString);
        assertMatches("^[^ ]+ [0-9]+ [^ ]+ 20[0-9][0-9]$", nextYearString);

        date.set(1973, Calendar.DECEMBER, 9);
        assertEquals("Söndag 9 december 1973",
            CalendarUtil.formatDate(date).toString());
    }

    /**
     * Return a time a given number of seconds ago.
     *
     * @param seconds The number of seconds to go back in time.
     *
     * @return A time since the epoch representing that number of seconds ago.
     */
    private CharSequence describeSecondsAgo(long seconds) {
        return CalendarUtil.describeTime(
            System.currentTimeMillis() - seconds * 1000L);
    }

    /**
     * Verify {@link CalendarUtil#describeTime(long)}
     */
    public void testDescribeTime() {
        assertEquals("mindre än en minut gammal", describeSecondsAgo(0));
        assertEquals("mindre än en minut gammal", describeSecondsAgo(59));
        assertEquals("en minut gammal", describeSecondsAgo(60));
        assertEquals("en minut gammal", describeSecondsAgo(60 + 59));
        assertEquals("2 minuter gammal", describeSecondsAgo(120));
        assertEquals("59 minuter gammal", describeSecondsAgo(3599));
        assertEquals("en timme gammal", describeSecondsAgo(3600));
        assertEquals("en timme gammal", describeSecondsAgo(3600 + 3599));
        assertEquals("2 timmar gammal", describeSecondsAgo(7200));
        assertEquals("23 timmar gammal", describeSecondsAgo(24 * 3600 - 1));
        assertTrue(describeSecondsAgo(24 * 3600).toString().startsWith("från "));

        Calendar date = new GregorianCalendar(1973, Calendar.DECEMBER, 9);
        assertEquals("från 9 december 1973",
            CalendarUtil.describeTime(date.getTimeInMillis()).toString());
    }
}
