/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl;

import net.launchpad.sl.location.NamedLocation;
import android.location.Location;
import android.os.Parcel;
import junit.framework.TestCase;

/**
 * Verify the workings of {@link NamedLocation}.
 *
 * @author johan.walles@gmail.com
 */
public class TestNamedLocation extends TestCase {
    /**
     * Verify parcelling of the given named location.
     *
     * @param parcelMe The named location we should try to parcel.
     */
    private void testParceling(NamedLocation parcelMe) {
        // Marshall
        Parcel parcel = Parcel.obtain();
        parcelMe.writeToParcel(parcel, 0);
        byte marshalled[] = parcel.marshall();
        parcel.recycle();

        // Unmarshall
        parcel = Parcel.obtain();
        parcel.unmarshall(marshalled, 0, marshalled.length);
        parcel.setDataPosition(0);
        NamedLocation parcelled = NamedLocation.CREATOR.createFromParcel(parcel);
        parcel.recycle();

        assertEquals(parcelMe, parcelled);
    }

    /**
     * Verify location parcelling.
     */
    public void testParceling() {
        testParceling(new NamedLocation("Apstigen 5", 42, 45, "Skansen"));
        testParceling(new NamedLocation("storgatan 5", "mammas gata"));
        testParceling(new NamedLocation("apa"));
    }

    /**
     * Verify that toString() works as intended.
     */
    public void testToString() {
        NamedLocation namedLocation = new NamedLocation("johan");
        assertEquals("johan", namedLocation.toString());

        namedLocation = new NamedLocation("storgatan 5", "mammas gata");
        assertEquals("storgatan 5 (mammas gata)", namedLocation.toString());

        namedLocation = new NamedLocation("storgatan 5", null);
        assertEquals("storgatan 5", namedLocation.toString());

        namedLocation = new NamedLocation("storgatan 5", null);
        assertEquals("storgatan 5", namedLocation.toString());

        namedLocation =
            new NamedLocation("storgatan 5", new Location("hej"), "mammas gata");
        assertEquals("storgatan 5 (mammas gata)", namedLocation.toString());
    }

    /**
     * Verify that the NamedLocation can keep track of what kind of location
     * it is.
     */
    public void testStates() {
        NamedLocation namedLocation = new NamedLocation("johan");
        assertTrue(namedLocation.isHere());
        assertFalse(namedLocation.isAddressOrStop());
        try {
            namedLocation.getCoordinate();
            fail("Requesting the coordinate of a \"here\" location should fail");
        } catch (IllegalStateException e) {
            // Expected exception intentionally ignored
        }
        assertNull(namedLocation.getAddressOrStop());
        assertEquals("johan", namedLocation.getAlias());

        Location location = new Location("johan");
        location.setLatitude(47);
        location.setLongitude(11);
        namedLocation =
            new NamedLocation("storgatan 5", location, "mammas gata");
        assertFalse(namedLocation.isHere());
        assertFalse(namedLocation.isAddressOrStop());
        assertEquals(location, namedLocation.getCoordinate());
        assertEquals("storgatan 5", namedLocation.getAddressOrStop());
        assertEquals("mammas gata", namedLocation.getAlias());

        namedLocation =
            new NamedLocation("storgatan 5", 9.0, 10.0, "mammas gata");
        assertFalse(namedLocation.isHere());
        assertFalse(namedLocation.isAddressOrStop());
        assertEquals(9.0, namedLocation.getCoordinate().getLatitude());
        assertEquals(10.0, namedLocation.getCoordinate().getLongitude());
        assertEquals("storgatan 5", namedLocation.getAddressOrStop());
        assertEquals("mammas gata", namedLocation.getAlias());

        namedLocation =
            new NamedLocation("storgatan 5", "mammas gata");
        assertTrue(namedLocation.isAddressOrStop());
        try {
            namedLocation.isHere();
            fail("Asking an address if it's here should result in an exception");
        } catch (IllegalStateException e) {
            // Expected exception intentionally ignored
        }
        try {
            namedLocation.getCoordinate();
            fail("Asking an address for its coordinate should fail");
        } catch (IllegalStateException e) {
            // Expected exception intentionally ignored
        }
        assertEquals("storgatan 5", namedLocation.getAddressOrStop());
        assertEquals("mammas gata", namedLocation.getAlias());
    }

    /**
     * Verify constructors.
     */
    public void testConstructor() {
        try {
            new NamedLocation(null, "Mammas gata");
            fail("An address location shouldn't accept a null address");
        } catch (NullPointerException e) {
            // Expected exception intentionally ignored
        }
        assertEquals("Storgatan 5",
            new NamedLocation("Storgatan 5", null).getAlias());

        try {
            new NamedLocation(null);
            fail("A here location shouldn't accept a null alias");
        } catch (NullPointerException e) {
            // Expected exception intentionally ignored
        }

        try {
            new NamedLocation("Storgatan 5", new Location("johan"), null);
            fail("A coordinate location shouldn't accept a null alias");
        } catch (NullPointerException e) {
            // Expected exception intentionally ignored
        }

        try {
            new NamedLocation("Storgatan 5", 47, 11, null);
            fail("A coordinate location shouldn't accept a null alias");
        } catch (NullPointerException e) {
            // Expected exception intentionally ignored
        }
    }
}
