/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl;

import java.util.Calendar;
import java.util.GregorianCalendar;

import net.launchpad.sl.Query;
import net.launchpad.sl.SL;
import net.launchpad.sl.location.NamedLocation;

import android.location.Location;
import android.os.Parcel;
import android.util.Log;

import junit.framework.TestCase;

/**
 * Verify the {@link Query} functionality.
 *
 * @author johan.walles@gmail.com
 */
public class TestQuery extends TestCase {
    /**
     * An example location.
     */
    private static NamedLocation home =
        new NamedLocation("ÅÄÖåäöé", 59.312214, 18.068945, "Location with Swedish chars");

    /**
     * Another example location.
     */
    private static NamedLocation momsPlace =
        new NamedLocation("Barks väg, Bergshamra", 59.361195, 18.105469, "Mom's place");

    /**
     * Verify that we generate an acceptable query string.
     */
    public void testToUri() {
        Calendar time =
            new GregorianCalendar(2009, Calendar.SEPTEMBER, 23, 7, 35);
        Query query = new Query(home, momsPlace);
        Log.i(SL.TAG, "Generated URL: " + query.toUriString(null, false));

        query.setWhen(time, true);
        assertEquals("http://reseplanerare.sl.se/bin/query.exe/sn?queryPageDisplayed=yes&ignoreTypeCheck=yes&REQ0JourneyStopsS0A=255&REQ0JourneyStopsS0G=%C5%C4%D6%E5%E4%F6%E9&REQ0JourneyStopsSID=A%3D16%40O%3D%C5%C4%D6%E5%E4%F6%E9%40X%3D18068945%40Y%3D59312214%40&REQ0JourneyStopsZ0A=255&REQ0JourneyStopsZ0G=Barks%20v%E4g%2C%20Bergshamra&REQ0JourneyStopsZID=A%3D16%40O%3DBarks%20v%E4g%2C%20Bergshamra%40X%3D18105469%40Y%3D59361195%40&REQ0HafasSearchForw=1&REQ0JourneyDate=23.09.09&REQ0JourneyTime=07%3A35&existUnsharpSearch=yes&SetGlobalOptionGO_displayDataInTable=no&start=S%F6k%20resa#FormMainHeader",
                     query.toUriString(null, false));

        query.setWhen(time, false);
        Log.i(SL.TAG, "Generated URL: " + query.toUriString(null, true));
        assertEquals("http://reseplanerare.sl.se/bin/query.exe/sn?queryPageDisplayed=yes&ignoreTypeCheck=yes&REQ0JourneyStopsS0A=255&REQ0JourneyStopsS0G=%C5%C4%D6%E5%E4%F6%E9&REQ0JourneyStopsSID=A%3D16%40O%3D%C5%C4%D6%E5%E4%F6%E9%40X%3D18068945%40Y%3D59312214%40&REQ0JourneyStopsZ0A=255&REQ0JourneyStopsZ0G=Barks%20v%E4g%2C%20Bergshamra&REQ0JourneyStopsZID=A%3D16%40O%3DBarks%20v%E4g%2C%20Bergshamra%40X%3D18105469%40Y%3D59361195%40&REQ0HafasSearchForw=0&REQ0JourneyDate=23.09.09&REQ0JourneyTime=07%3A35&existUnsharpSearch=yes&SetGlobalOptionGO_displayDataInTable=yes&start=S%F6k%20resa#FormMainHeader",
                     query.toUriString(null, true));
    }

    /**
     * Verify that we can describe the starting point in a format acceptable to
     * sl.se.
     */
    public void testGetLocationString() {
        NamedLocation namedLocation =
            new NamedLocation("Ankeborg", 59.312214, 18.068945, "Kalle");
        Query query = new Query(null, null);
        assertEquals("A=16@O=Ankeborg@X=18068945@Y=59312214@",
                     query.getLocationString(namedLocation, null));

        NamedLocation here = new NamedLocation("here4real");
        String locationString =
            query.getLocationString(here, new Location("johan"));
        assertTrue(locationString.contains("@O=here4real@"));
    }

    /**
     * Verify that our timestamp can be properly converted into a time string.
     */
    public void testGetTimeString() {
        Calendar time =
            new GregorianCalendar(2009, Calendar.SEPTEMBER, 5, 17, 50);
        Query query = new Query(null, null);

        query.setWhen(time, false);
        assertEquals("17:50", query.getTimeString());

        time = new GregorianCalendar(2009, Calendar.SEPTEMBER, 5, 8, 05);
        query.setWhen(time, false);
        assertEquals("08:05", query.getTimeString());
    }

    /**
     * Verify that our timestamp can be properly converted into a date string.
     */
    public void testGetDateString() {
        Calendar time = new GregorianCalendar(2009, Calendar.SEPTEMBER, 5);
        Query query = new Query(null, null);
        query.setWhen(time, true);
        assertEquals("05.09.09", query.getDateString());

        time = new GregorianCalendar(2010, Calendar.NOVEMBER, 13);
        query.setWhen(time, true);
        assertEquals("13.11.10", query.getDateString());
    }

    /**
     * Verify parcelling of the given query.
     *
     * @param parcelMe The query we should try to parcel.
     */
    private void testParceling(Query parcelMe) {
        // Marshall
        Parcel parcel = Parcel.obtain();
        parcelMe.writeToParcel(parcel, 0);
        byte marshalled[] = parcel.marshall();
        parcel.recycle();

        // Unmarshall
        parcel = Parcel.obtain();
        parcel.unmarshall(marshalled, 0, marshalled.length);
        parcel.setDataPosition(0);
        Query parcelled = Query.CREATOR.createFromParcel(parcel);
        parcel.recycle();

        assertEquals(parcelMe, parcelled);
    }

    /**
     * Verify query parcelling.
     */
    public void testParceling() {
        testParceling(new Query(home, momsPlace));
        testParceling(new Query(home, new NamedLocation("here")));
        testParceling(new Query(new NamedLocation("here"), momsPlace));

        Query query = new Query(home, momsPlace);

        query.setWhen(new GregorianCalendar(), false);
        testParceling(query);

        query.setWhen(new GregorianCalendar(), true);
        testParceling(query);

        query.setWhen(null, false);
        testParceling(query);

        query.setWhen(null, true);
        testParceling(query);
    }
}
