package net.launchpad.sl;

import java.util.Date;

import net.launchpad.sl.location.NamedLocation;
import net.launchpad.sl.location.ScoredLocation;
import junit.framework.TestCase;

/**
 * Verify the functionality of {@link ScoredLocation}.
 *
 * @author johan.walles@gmail.com
 */
public class TestScoredLocation extends TestCase {
    /**
     * Verify the functionality of {@link ScoredLocation#getFromLastUse()} and
     * {@link ScoredLocation#setFromLastUse()}.
     */
    public void testFromLastUse() {
        ScoredLocation location = new ScoredLocation(null, 0, 0);
        assertNull(location.getFromLastUse());
        assertNull(location.getLastUse());

        Date someDate = new Date(4711);
        location.setFromLastUse(someDate);
        assertEquals(someDate, location.getFromLastUse());
        assertEquals(someDate, location.getLastUse());

        Date now = new Date();
        location.setFromLastUse();
        long diffMs = now.getTime() - location.getFromLastUse().getTime();
        assertTrue(Math.abs(diffMs) < 1000);
    }

    /**
     * Verify the functionality of {@link ScoredLocation#getToLastUse()} and
     * {@link ScoredLocation#setToLastUse()}.
     */
    public void testToLastUse() {
        ScoredLocation location = new ScoredLocation(null, 0, 0);
        assertNull(location.getToLastUse());
        assertNull(location.getLastUse());

        Date someDate = new Date(4711);
        location.setToLastUse(someDate);
        assertEquals(someDate, location.getToLastUse());
        assertEquals(someDate, location.getLastUse());

        Date now = new Date();
        location.setToLastUse();
        long diffMs = now.getTime() - location.getToLastUse().getTime();
        assertTrue(Math.abs(diffMs) < 1000);
    }

    /**
     * Verify the functionality of {@link ScoredLocation#getLastUse()}.
     */
    public void testLastUse() {
        ScoredLocation location = new ScoredLocation(null, 0, 0);
        assertNull(location.getLastUse());

        Date someDate = new Date(4711);
        Date someLaterDate = new Date(4711 + 4711);
        location.setFromLastUse(someLaterDate);
        assertEquals(someLaterDate, location.getLastUse());

        location.setToLastUse(someDate);
        assertEquals(someLaterDate, location.getLastUse());
    }

    /**
     * Verify the functionality of {@link ScoredLocation#equals(Object)}.
     */
    public void testEquals() {
        NamedLocation namedLocation = new NamedLocation("gris");
        ScoredLocation a = new ScoredLocation(namedLocation, 0, 0);
        assertEquals(a, a);

        ScoredLocation b = new ScoredLocation(namedLocation, 0, 0);
        b.setFromLastUse();
        assertFalse("Scored Locations with different time stamps shouldn't equals() each other",
            a.equals(b));

        ScoredLocation c = new ScoredLocation(namedLocation, 0, 0);
        c.setToLastUse();
        assertFalse("Scored Locations with different time stamps shouldn't equals() each other",
            a.equals(c));

        // Verify the from date comparison
        Date someDate = new Date(4711);
        a.setFromLastUse(someDate);
        ScoredLocation d = new ScoredLocation(namedLocation, 0, 0);
        d.setFromLastUse(someDate);
        assertEquals(a, d);

        d.setFromLastUse();
        assertFalse("Scored Locations with different time stamps shouldn't equals() each other",
            a.equals(d));
        a.setFromLastUse(null);

        // Verify the to date comparison
        a.setToLastUse(someDate);
        ScoredLocation e = new ScoredLocation(namedLocation, 0, 0);
        e.setToLastUse(someDate);
        assertEquals(a, e);

        e.setToLastUse();
        assertFalse("Scored Locations with different time stamps shouldn't equals() each other",
            a.equals(e));
        a.setFromLastUse(null);
    }

    /**
     * Verify {@link ScoredLocation#hashCode()}.
     */
    public void testHashCode() {
        ScoredLocation location =
            new ScoredLocation(new NamedLocation("gris"), 0, 0);
        int h0 = location.hashCode();

        location.setFromLastUse(new Date(4711));
        int h1 = location.hashCode();

        location.setToLastUse(new Date(4711000));
        int h2 = location.hashCode();

        assertFalse("Setting a from date should change the hash", h0 == h1);
        assertFalse("Setting a to date should change the hash", h1 == h2);
        assertFalse("Setting the dates should change the hash", h0 == h2);
    }

    /**
     * Verify {@link ScoredLocation#FROM_COMPARATOR}.
     */
    public void testFromComparator() {
        NamedLocation namedLocation = new NamedLocation("gris");

        ScoredLocation neverUsed = new ScoredLocation(namedLocation, 0, 0);

        ScoredLocation usedLongAgo = new ScoredLocation(namedLocation, 0, 0);
        usedLongAgo.setFromLastUse(new Date(0));

        ScoredLocation usedRecently = new ScoredLocation(namedLocation, 0, 0);
        usedRecently.setFromLastUse();

        assertEquals("Used locations should come before unused locations", 1,
            ScoredLocation.FROM_COMPARATOR.compare(neverUsed, usedLongAgo));
        assertEquals("Recent locations should come before old locations", -1,
            ScoredLocation.FROM_COMPARATOR.compare(usedRecently, usedLongAgo));
        assertEquals("Equal locations should compare equally", 0,
            ScoredLocation.FROM_COMPARATOR.compare(neverUsed, neverUsed));
        assertEquals("Equal locations should compare equally", 0,
            ScoredLocation.FROM_COMPARATOR.compare(usedRecently, usedRecently));
    }

    /**
     * Verify {@link ScoredLocation#TO_COMPARATOR}.
     */
    public void testToComparator() {
        NamedLocation namedLocation = new NamedLocation("gris");

        ScoredLocation neverUsed = new ScoredLocation(namedLocation, 0, 0);

        ScoredLocation usedLongAgo = new ScoredLocation(namedLocation, 0, 0);
        usedLongAgo.setToLastUse(new Date(0));

        ScoredLocation usedRecently = new ScoredLocation(namedLocation, 0, 0);
        usedRecently.setToLastUse();

        assertEquals("Used locations should come before unused locations", 1,
            ScoredLocation.TO_COMPARATOR.compare(neverUsed, usedLongAgo));
        assertEquals("Recent locations should come before old locations", -1,
            ScoredLocation.TO_COMPARATOR.compare(usedRecently, usedLongAgo));
        assertEquals("Equal locations should compare equally", 0,
            ScoredLocation.TO_COMPARATOR.compare(neverUsed, neverUsed));
        assertEquals("Equal locations should compare equally", 0,
            ScoredLocation.TO_COMPARATOR.compare(usedRecently, usedRecently));
    }
}
