/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl.location;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import net.launchpad.sl.location.LocationsDbAdapter;
import net.launchpad.sl.location.NamedLocation;
import net.launchpad.sl.location.ScoredLocation;

import android.content.ContentValues;
import android.content.ContextWrapper;
import android.database.sqlite.SQLiteDatabase;
import android.test.AndroidTestCase;

/**
 * Verify database operations.
 *
 * @author johan.walles@gmail.com
 */
public class TestLocationsDbAdapter extends AndroidTestCase {
    /**
     * The name of our test database.
     */
    private final static String DB_NAME = "test";

    /**
     * A test location.
     */
    private final ScoredLocation here =
        new ScoredLocation(new NamedLocation("here"), 1, 2);

    /**
     * A test location.
     */
    private final ScoredLocation there =
        new ScoredLocation(new NamedLocation("there address", 42.123, 24.567, "there"), 3, 4);

    /**
     * A test location.
     */
    private final ScoredLocation there2 =
        new ScoredLocation(new NamedLocation("there2 address", 12.345, 6.789, "there2"), 5, 6);

    /**
     * A test location.
     */
    private final ScoredLocation anAddress =
        new ScoredLocation(new NamedLocation("Storgatan 5", "Mammas gata"), 7, 8);

    /**
     * Verify that existing schema version 2 database can be upgraded to the
     * current database format.
     */
    public void testUpgradeFromSchemaV2() {
        // Drop the test database
        ContextWrapper contextWrapper = new ContextWrapper(getContext());
        contextWrapper.deleteDatabase(DB_NAME);

        LocationsDbAdapter adapter =
            new LocationsDbAdapter(getContext(), DB_NAME);
        assertFalse(adapter.needsUpgrade());

        // Create a new base database containing one single "current location"
        // entry
        SQLiteDatabase db =
            contextWrapper.openOrCreateDatabase(DB_NAME, 0, null);
        try {
            adapter.createBaseDb(db);
            adapter.upgradeDbV1ToV2(null, db);
            assertEquals(2, db.getVersion());
        } finally {
            db.close();
        }
        assertTrue("A v2 format database should require an upgrade",
            adapter.needsUpgrade());

        // Add a test row in the v2 database format to the database
        db = contextWrapper.openOrCreateDatabase(DB_NAME, 0, null);
        try {
            ContentValues values = new ContentValues();
            values.put("name", "Hemma");
            values.put("from_score", 2);
            values.put("to_score", 3);
            values.put("latitude", 59.312214);
            values.put("longitude", 18.068945);
            values.put("description", "Åsögatan 90");
            db.insertOrThrow("locations", null, values);

            values = new ContentValues();
            values.put("name", "Var som helst med 1 siffra i");
            values.put("from_score", 4);
            values.put("to_score", 5);
            values.put("latitude", 59.1795);
            values.put("longitude", 17.6534);
            values.put("description", "Sergels 7org");
            db.insertOrThrow("locations", null, values);

            // Upgrade the test database
            adapter.upgradeDb(new LocationsDbAdapter.Progress() {
                private int expectedProgress = 0;

                public void onProgress(int percent) {
                    assertEquals(expectedProgress, percent);

                    // Two rows need potentially time consuming upgrades
                    expectedProgress += 50;
                }
            });

            assertEquals(3, db.getVersion());
            assertFalse(adapter.needsUpgrade());
        } finally {
            db.close();
        }

        // Verify the contents of the migrated test rows
        Set<ScoredLocation> locations = adapter.fetchLocations();
        assertEquals(3, locations.size());
        assertTrue(locations.contains(new ScoredLocation(
            new NamedLocation("Telefonens nuvarande position"), 1, 0)));
        assertTrue(locations.contains(new ScoredLocation(
            new NamedLocation("Åsögatan 90", 59.312214, 18.068945, "Hemma"),
            2, 3)));
        assertTrue(locations.contains(new ScoredLocation(
            new NamedLocation("Sergels 7org",
                59.1795, 17.6534, "Var som helst med 1 siffra i"),
            4, 5)));

        assertFalse(adapter.needsUpgrade());
    }

    /**
     * Verify that existing schema version 1 database can be upgraded to the
     * current database format.
     */
    public void testUpgradeFromSchemaV1() {
        // Drop the test database
        ContextWrapper contextWrapper = new ContextWrapper(getContext());
        contextWrapper.deleteDatabase(DB_NAME);

        LocationsDbAdapter adapter =
            new LocationsDbAdapter(getContext(), DB_NAME);
        assertFalse(adapter.needsUpgrade());

        // Create a new base database containing one single "current location"
        // entry
        SQLiteDatabase db =
            contextWrapper.openOrCreateDatabase(DB_NAME, 0, null);
        try {
            adapter.createBaseDb(db);
            assertEquals(1, db.getVersion());
        } finally {
            db.close();
        }
        assertTrue("A base format database should require an upgrade",
            adapter.needsUpgrade());

        // Add a test row in the original database format to the database
        db = contextWrapper.openOrCreateDatabase(DB_NAME, 0, null);
        try {
            ContentValues values = new ContentValues();
            values.put("name", "Hemma");
            values.put("from_score", 2);
            values.put("to_score", 3);
            values.put("latitude", 59.312214);
            values.put("longitude", 18.068945);
            db.insertOrThrow("locations", null, values);

            values = new ContentValues();
            values.put("name", "Var som helst med 1 siffra i");
            values.put("from_score", 4);
            values.put("to_score", 5);
            values.put("latitude", 59.1795);
            values.put("longitude", 17.6534);
            db.insertOrThrow("locations", null, values);

            // Upgrade the test database
            adapter.upgradeDb(new LocationsDbAdapter.Progress() {
                private int expectedProgress = 0;

                public void onProgress(int percent) {
                    assertEquals(expectedProgress, percent);

                    // Two rows need potentially time consuming upgrades
                    expectedProgress += 50;
                }
            });

            assertEquals(3, db.getVersion());
            assertFalse(adapter.needsUpgrade());
        } finally {
            db.close();
        }

        // Verify the contents of the migrated test rows
        Set<ScoredLocation> locations = adapter.fetchLocations();
        assertEquals(3, locations.size());
        assertTrue(locations.contains(new ScoredLocation(
            new NamedLocation("Telefonens nuvarande position"), 1, 0)));
        assertTrue(locations.contains(new ScoredLocation(
            new NamedLocation("Åsögatan 90", 59.312214, 18.068945, "Hemma"),
            2, 3)));
        assertTrue(locations.contains(new ScoredLocation(
            new NamedLocation("Var som helst med 1 siffra i",
                59.1795, 17.6534, "Var som helst med 1 siffra i"),
            4, 5)));

        assertFalse(adapter.needsUpgrade());
    }

    /**
     * Verify that saving to and reading from the database works.
     */
    public void testPersistence() {
        // Drop the test database
        new ContextWrapper(getContext()).deleteDatabase(DB_NAME);

        LocationsDbAdapter db = new LocationsDbAdapter(getContext(), DB_NAME);
        Set<ScoredLocation> fetchedLocations = db.fetchLocations();
        assertEquals("Opening an empty database should give us one entry containing"
                     + "the phone's current location and nothing else",
                     1, fetchedLocations.size());
        ScoredLocation location = fetchedLocations.iterator().next();
        assertTrue(location.getLocation().isHere());
        assertTrue(location.getLocation().getAlias().length() > 0);
        assertEquals(1, location.getFromScore());
        assertEquals(0, location.getToScore());
        assertNull(location.getLastUse());

        // Put our test locations in the database
        List<ScoredLocation> locations = new LinkedList<ScoredLocation>();
        locations.add(here);
        locations.add(there);
        locations.add(there2);
        locations.add(anAddress);
        db.storeLocations(locations);

        // Fetch our test locations from the database
        fetchedLocations = db.fetchLocations();
        assertEquals(locations.size(), fetchedLocations.size());
        assertTrue(fetchedLocations.contains(here));
        assertTrue(fetchedLocations.contains(there));
        assertTrue(fetchedLocations.contains(there2));
        assertTrue(fetchedLocations.contains(anAddress));
    }

    /**
     * Verify that we can properly increase the scores of two locations.
     */
    // We have assertions covering what's being warned about here
    @SuppressWarnings("null")
    public void testIncreaseScore() {
        // Drop the test database
        new ContextWrapper(getContext()).deleteDatabase(DB_NAME);

        // Put two locations in the database
        LocationsDbAdapter db = new LocationsDbAdapter(getContext(), DB_NAME);
        Set<ScoredLocation> locations = new HashSet<ScoredLocation>(2);
        locations.add(here);
        locations.add(there);
        db.storeLocations(locations);

        // Increase the scores of the two locations
        db.increaseScores(here.getLocation(), there.getLocation());

        // Fetch db contents and verify that it matches expectations
        Set<ScoredLocation> fetched = db.fetchLocations();
        assertEquals(2, fetched.size());
        ScoredLocation newHere = null;
        ScoredLocation newThere = null;
        for (ScoredLocation location : fetched) {
            if (location.getLocation().getAlias().equals(here.getLocation().getAlias())) {
                newHere = location;
            }
            if (location.getLocation().getAlias().equals(there.getLocation().getAlias())) {
                newThere = location;
            }
        }
        assertNotNull(newHere);
        assertNotNull(newThere);

        assertEquals("From location's from score should have increased by 1",
                     here.getFromScore() + 1, newHere.getFromScore());
        assertEquals("From location's to score should remain unchanged",
                     here.getToScore(), newHere.getToScore());

        assertEquals("To location's from score should remain unchanged",
                     there.getFromScore(), newThere.getFromScore());
        assertEquals("To location's to score should have increased by 1",
                     there.getToScore() + 1, newThere.getToScore());
    }

    /**
     * Verify that location renaming works.
     */
    public void testRenameLocation() {
        // Drop the test database
        new ContextWrapper(getContext()).deleteDatabase(DB_NAME);

        // Set up a test location
        LocationsDbAdapter db = new LocationsDbAdapter(getContext(), DB_NAME);
        Set<ScoredLocation> locations = new HashSet<ScoredLocation>(2);
        locations.add(here);
        db.storeLocations(locations);

        db.renameLocation(here.getLocation().getAlias(), "a better here");

        // Verify that the test location is still itself, just with a new name
        Set<ScoredLocation> fetchedLocations = db.fetchLocations();
        assertEquals(1, fetchedLocations.size());

        ScoredLocation betterHere = fetchedLocations.iterator().next();
        assertEquals("a better here", betterHere.getLocation().getAlias());
        assertTrue(betterHere.getLocation().samePlaceAs(here.getLocation()));
        assertEquals(here.getFromScore(), betterHere.getFromScore());
        assertEquals(here.getToScore(), betterHere.getToScore());
    }

    /**
     * Verify storage of a single scored location in the database.
     *
     * @param location The location we should try to store.
     */
    private void testStore(ScoredLocation location) {
        new ContextWrapper(getContext()).deleteDatabase(DB_NAME);
        LocationsDbAdapter db = new LocationsDbAdapter(getContext(), DB_NAME);
        Set<ScoredLocation> locations = new HashSet<ScoredLocation>();
        locations.add(location);
        db.storeLocations(locations);
        locations = db.fetchLocations();
        assertEquals(1, locations.size());
        assertEquals(location, locations.iterator().next());
    }

    /**
     * Verify basic store and retrieve.
     */
    public void testStore() {
        testStore(here);
        testStore(there);
        testStore(anAddress);
    }

    /**
     * Verify that location removal works.
     */
    public void testRemoveLocation() {
        // Drop the test database
        new ContextWrapper(getContext()).deleteDatabase(DB_NAME);

        // Set up two test locations
        LocationsDbAdapter db = new LocationsDbAdapter(getContext(), DB_NAME);
        Set<ScoredLocation> locations = new HashSet<ScoredLocation>(2);
        locations.add(here);
        locations.add(there);
        db.storeLocations(locations);

        // Remove here...
        db.removeLocation(here.getLocation().getAlias());
        Set<ScoredLocation> fetchedLocations = db.fetchLocations();
        assertEquals(1, fetchedLocations.size());

        // And verify that there is still there...
        ScoredLocation hopefullyThere = fetchedLocations.iterator().next();
        assertEquals(there, hopefullyThere);
    }
}
