/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.launchpad.sl.trip;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

import android.location.Location;
import android.test.AndroidTestCase;

/**
 * Verify that we can parse map pages from reseplanerare.sl.se.
 *
 * @author johan.walles@gmail.com
 */
public class TestSlMapData extends AndroidTestCase {
    /**
     * How close does two locations need to be to be considered the same place?
     */
    private final static int CLOSE_ENOUGH_METERS = 10;

    /**
     * Get resource contents as an InputStream.
     *
     * @param resourceName The resource file to read from.
     *
     * @return An InputStream reading from the resource file.
     *
     * @throws Exception on trouble
     */
    private InputStream getResource(String resourceName) throws Exception {
        return getClass().getResourceAsStream(resourceName);
    }

    /**
     * Verify that a locations is at the right coordinates.
     *
     * @param latitude The expected latitude.
     *
     * @param longitude The expected longitude.
     *
     * @param location The location to check vs latitude and longitude.
     */
    private void assertCloseEnough(double latitude, double longitude,
        Location location)
    {
        Location thePlaceToBe = new Location("johan");
        thePlaceToBe.setLatitude(latitude);
        thePlaceToBe.setLongitude(longitude);

        int distance = (int)location.distanceTo(thePlaceToBe);
        if (distance <= CLOSE_ENOUGH_METERS) {
            return;
        }

        fail("Location is " + distance + "m off, won't accept more than "
            + CLOSE_ENOUGH_METERS + "m");
    }

    /**
     * Verify map parsing.
     *
     * @throws Exception when testing goes exceptionally bad
     */
    public void testParser() throws Exception {
        InputStream stream = getResource("map1.html");
        SlMapData slMapData = new SlMapData(getContext());
        slMapData.populate(stream, true);
        try {
            stream.read();
            fail("SlMapData constructor should have closed the stream");
        } catch (IOException e) {
            // Expected exception intentionally ignored
        }

        // Åsögatan 90
        assertCloseEnough(59.3122, 18.0691, slMapData.getLocation());
        // Mariatorget
        assertCloseEnough(59.3169, 18.0622, slMapData.getStopLocation());

        slMapData = new SlMapData(getContext());
        slMapData.populate(getResource("map2.html"), false);
        // Barks väg 6
        assertCloseEnough(59.3793, 18.0321, slMapData.getLocation());
        // Bergshamra
        assertCloseEnough(59.3806, 18.0370, slMapData.getStopLocation());

        slMapData = new SlMapData(getContext());
        slMapData.populate(getResource("map3.html"), false);
        // Ängbyplan
        assertCloseEnough(59.3419, 17.9070, slMapData.getStopLocation());
        assertNull(slMapData.getLocation());

        slMapData = new SlMapData(getContext());
        try {
            slMapData.populate("file:///dev/null", false);
            fail("Parsing a file with no locations should throw an exception");
        } catch (TripException e) {
            // Expected exception intentionally ignored
        }
    }

    /**
     * Verify that we can interrupt map data downloads.
     *
     * @throws Exception if testing goes exceptionally bad
     */
    public void testInterrupt() throws Exception {
        final SlMapData mapData = new SlMapData(getContext());
        final Exception exceptionHolder[] = new Exception[1];
        Thread downloader = new Thread(new Runnable() {
            public void run() {
                try {
                    mapData.populate(new FileInputStream("/dev/zero"), false);
                } catch (Exception e) {
                    exceptionHolder[0] = e;
                }
            }
        }, "Map data downloader");
        downloader.start();

        // Give the downloader time to start waiting for /dev/zero
        Thread.sleep(500);

        mapData.interruptDownload();
        downloader.join(500);
        assertFalse("interruptDownload() should terminate the downloading thread",
            downloader.isAlive());

        // Verify that we got an exception saying that the download was
        // canceled.
        assertNotNull(exceptionHolder[0]);
        assertEquals(IOException.class, exceptionHolder[0].getClass());
        assertEquals("Nedladdningen avbröts", exceptionHolder[0].getMessage());
    }
}
