/*
 * SL helps you search for trips with SL, Stockholm's public transport
 * company.  Note that the application is in Swedish.
 *
 *  Copyright (C) 2009  Johan Walles, johan.walles@gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.launchpad.sl.trip;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import android.test.AndroidTestCase;

/**
 * Verify the functionality of the {@link TripList}.
 *
 * @author johan.walles@gmail.com
 */
public class TestTripList extends AndroidTestCase {
    /**
     * Verify that a trip matches certain conditions.
     *
     * @param id The expected unique identifier for this trip
     * @param trip The trip to check
     * @param firstStop The expected first stop of the trip
     * @param lastStop The expected last stop if the trip
     * @param departureTime The departure time
     * @param arrivalTime The arrival time
     * @param durationMinutes The expected number of minutes of duration for
     * the trip.
     * @param changes The expected number of changes for this trip
     * @param transports The transports expected to be part of this trip
     */
    private void assertEquals(Trip trip, int id,
        String firstStop, String lastStop,
        String departureTime,
        String arrivalTime,
        int durationMinutes,
        int changes,
        String ... transports)
    {
        assertEquals(id, trip.getId());
        assertEquals(firstStop, trip.getFirstStop());
        assertEquals(lastStop, trip.getLastStop());
        assertEquals(departureTime, trip.getDepartureTime());
        assertEquals(arrivalTime, trip.getArrivalTime());
        assertEquals(durationMinutes, trip.getDurationMinutes());
        assertEquals(changes, trip.getChanges());
        assertNotNull(trip.getTransports());
        assertEquals("A trip must have a list of transports",
            transports.length, trip.getTransports().size());
        for (int i = 0; i < transports.length; i++) {
            assertEquals(transports[i], trip.getTransports().get(i));
        }
    }

    /**
     * Verify that a trip part fulfills certain conditions
     *
     * @param part The trip part to check.
     * @param fromStop The first stop of this start of the trip.
     * @param fromTime The departure time of this part of the trips, or null if
     * the departure time is unknown.
     * @param toStop At which stop do we get off?
     * @param toTime The arrival time of this trip, or null if the arrival time
     * is unknown.
     * @param transport The transport for this part of the trip.
     * @param transportDetails Details about this part of the trip.
     */
    private void assertEquals(TripPart part,
        String fromStop, String fromTime,
        String toStop, String toTime,
        String transport,
        String transportDetails)
    {
        assertEquals(fromStop, part.getFromStop());
        assertEquals(toStop, part.getToStop());
        assertEquals(fromTime, part.getFromTime());
        assertEquals(toTime, part.getToTime());
        assertEquals(transport, part.getTransport());
        assertEquals(transportDetails, part.getTransportDetails());
    }

    /**
     * Sanity check our parsing of a HTML file.
     *
     * @param htmlResourceName The HTML file to try.
     *
     * @throws Exception when things go exceptionally wrong.
     */
    private void testSanity(String htmlResourceName) throws Exception {
        TripList tripList = new TripList(getContext(), null);
        tripList.populate(getResource(htmlResourceName));

        Set<Integer> usedIds = new HashSet<Integer>(tripList.size());
        for (Trip trip : tripList) {
            // Make sure no ID is used twice
            Integer id = Integer.valueOf(trip.getId());
            assertFalse(usedIds.contains(id));
            usedIds.add(id);

            assertNotNull(trip.getStartMapUrl());
            assertNotNull(trip.getEndMapUrl());

            assertNotNull(trip.getFirstStop());
            assertNotNull(trip.getDepartureTime());

            assertNotNull(trip.getLastStop());
            assertNotNull(trip.getArrivalTime());

            assertEquals(trip.getTransports().size() - 1, trip.getChanges());

            assertEquals(5, trip.getDepartureTime().length());
            int h0 = Integer.valueOf(trip.getDepartureTime().subSequence(0, 2).toString());
            int m0 = Integer.valueOf(trip.getDepartureTime().subSequence(3, 5).toString());

            assertEquals(5, trip.getArrivalTime().length());
            int h1 = Integer.valueOf(trip.getArrivalTime().subSequence(0, 2).toString());
            int m1 = Integer.valueOf(trip.getArrivalTime().subSequence(3, 5).toString());
            int durationMinutes = 60 * (h1 - h0) + (m1 - m0);
            assertEquals(durationMinutes, trip.getDurationMinutes());

            List<TripPart> parts = trip.getParts();
            if (parts != null) {
                assertEquals("The first and last parts are walking parts, and we can use one transport without changing, so a trip with four parts should have one change:\n" + trip,
                    trip.getChanges(), parts.size() - 3);

                String firstDeparture = parts.get(1).getFromTime();
                assertEquals(trip.getDepartureTime(), firstDeparture);
                String lastArrival = parts.get(parts.size() - 2).getToTime();
                assertEquals(trip.getArrivalTime(), lastArrival);

                for (int i = 0; i < parts.size(); i++) {
                    TripPart part = parts.get(i);
                    boolean firstPart = (i == 0);
                    boolean lastPart = (i == parts.size() - 1);

                    assertTrue(part.getFromStop().length() > 0);
                    assertTrue(part.getToStop().length() > 0);

                    if (!(firstPart || lastPart)) {
                        assertEquals(5, part.getFromTime().length());

                        assertEquals("To time has the wrong length: "
                            + htmlResourceName + "/[" + part + "]",
                            5, part.getToTime().length());

                        // This part should have the same transport that was
                        // mentioned in the summary
                        assertEquals(trip.getTransports().get(i - 1),
                            part.getTransport());
                    }

                    assertTrue(part.getTransportDetails().length() > 0);
                }
            }
        }
    }

    /**
     * Verify that a trip list can be created from HTML.
     *
     * @throws Exception if testing goes exceptionally bad
     */
    public void testCreate() throws Exception {
        TripList tripList = new TripList(getContext(), null);
        tripList.populate(getResource("answer.html"));
        assertEquals(3, tripList.size());
        Iterator<Trip> tripIterator = tripList.iterator();

        Trip trip = tripIterator.next();
        assertEquals(trip, 0x000000,
            "Stockholms södra", "Bergshamra gård",
            "07:14",
            "07:40",
            26,
            1,
            "pendeltåg 36", "buss 540");
        assertNull(trip.getParts());
        assertEquals("http://reseplanerare.sl.se/bin/query.exe/sn?seqnr=1&ident=nz.019218217.1257227854&performGis=1&REQMapConnectionId=C0-0&REQMapWidth=358&REQMapHeight=358&application=DIRECTIONS&guiVCtrl_connection_directions_onlyElement_C0-0&SetGlobalOptionGO_callMapFromPosition=tpDetailsRouteComplete&REQMapConnectionLocationNameHandling=2&REQMapConnectionMarkerTransferLocation=STATION&REQMapConnectionSection=START&SetGlobalOptionGO_activeMapSlider=START&REQMapConnectionMarkerFirstLocation=ADDRESS&REQMapConnectionMarkerLastLocation=STATION",
            trip.getStartMapUrl());
        assertEquals("http://reseplanerare.sl.se/bin/query.exe/sn?seqnr=1&ident=nz.019218217.1257227854&performGis=1&REQMapConnectionId=C0-0&REQMapWidth=358&REQMapHeight=358&application=DIRECTIONS&guiVCtrl_connection_directions_onlyElement_C0-0&SetGlobalOptionGO_callMapFromPosition=tpDetailsRouteComplete&REQMapConnectionLocationNameHandling=2&REQMapConnectionMarkerTransferLocation=STATION&REQMapConnectionSection=END&SetGlobalOptionGO_activeMapSlider=END&REQMapConnectionMarkerFirstLocation=STATION&REQMapConnectionMarkerLastLocation=ADDRESS",
            trip.getEndMapUrl());

        trip = tripIterator.next();
        assertEquals(trip, 0x000001,
            "Mariatorget", "Bergshamra",
            "07:22",
            "07:40",
            18,
            0,
            "tunnelbanans röda linje 14");

        List<TripPart> tripParts = trip.getParts();
        assertEquals(3, tripParts.size());

        Iterator<TripPart> partsIterator = tripParts.iterator();
        TripPart tripPart = partsIterator.next();
        assertEquals(tripPart,
            "Stockholm, Åsögatan 90", "",
            "Mariatorget", "",
            "Gå",
            "Cirka 900 meter");
        tripPart = partsIterator.next();
        assertEquals(tripPart,
            "Mariatorget", "07:22",
            "Bergshamra", "07:40",
            "tunnelbanans röda linje 14",
            "mot: Mörby centrum");
        tripPart = partsIterator.next();
        assertEquals(tripPart,
            "Bergshamra", "",
            "Solna, Barks Väg 6", "",
            "Gå",
            "Cirka 150 meter");

        trip = tripIterator.next();
        assertEquals(trip, 0x000002,
            "Mariatorget", "Bergshamra",
            "17:27",
            "17:45",
            18,
            0,
            "tunnelbanans röda linje 14");
        assertNull(trip.getParts());

        // Verify that the correct exception is thrown if we get error pages
        try {
            tripList = new TripList(getContext(), null);
            tripList.populate(getResource("error.html"));
            fail("Parsing an error page should throw a TripException");
        } catch (TripException e) {
            assertEquals("Start, mål eller via är samma hållplats eller ligger för nära varandra.",
                e.getMessage());
        }

        try {
            tripList = new TripList(getContext(), null);
            tripList.populate(getResource("error2.html"));
            fail("Parsing an error page should throw a TripException");
        } catch (TripException e) {
            assertEquals("Det finns ingen hållplats i närheten av den angivna adressen.",
                e.getMessage());
        }

        try {
            tripList = new TripList(mContext, null);
            tripList.populate("file:///dev/null");
            fail("Parsing an empty file should throw a TripException");
        } catch (TripException e) {
            assertEquals("Hittade inga resor!", e.getMessage());
        }

        testSanity("answer.html");
        testSanity("answer2.html");
        testSanity("answer-expanded.html");
        testSanity("answer-expanded-allchecked.html");
    }

    /**
     * Get resource contents as an InputStream.
     *
     * @param resourceName The resource file to read from.
     *
     * @return An InputStream reading from the resource file.
     *
     * @throws Exception on trouble
     */
    private InputStream getResource(String resourceName) throws Exception {
        return getClass().getResourceAsStream(resourceName);
    }

    /**
     * Verify that the trip list can construct URLs for going earlier, later,
     * and for retrieving details for all trips.
     *
     * @throws Exception when tests go exceptionally bad
     */
    public void testUrls() throws Exception {
        TripList tripList = new TripList(getContext(), null);
        tripList.populate(getResource("answer.html"));

        String checkAllBoxes =
            "guiVCtrl_connection_detailsOut_select_C0-0=yes"
            + "&guiVCtrl_connection_detailsOut_select_C0-1=yes"
            + "&guiVCtrl_connection_detailsOut_select_C0-2=yes"
            + "&guiVCtrl_connection_detailsOut_add_selection=yes";

        // Verify that we can generate a "check all boxes" URL
        assertEquals("http://reseplanerare.sl.se/bin/query.exe/sn?seqnr=1&ident=nz.019218217.1257227854"
            + "&" + checkAllBoxes,
            tripList.getGetDetailsUrl());

        // Verify that we can generate a "go earlier" URL
        assertEquals("http://reseplanerare.sl.se/bin/query.exe/sn?seqnr=1&ident=nz.019218217.1257227854"
            + "&" + checkAllBoxes
            + "&REQ0HafasScrollDir=2",
            tripList.getGoEarlierUrl());

        // Verify that we can generate a "go later" URL
        assertEquals("http://reseplanerare.sl.se/bin/query.exe/sn?seqnr=1&ident=nz.019218217.1257227854"
            + "&" + checkAllBoxes
            + "&REQ0HafasScrollDir=1",
            tripList.getGoLaterUrl());
    }

    /**
     * Verify that we can interrupt trip data downloads.
     *
     * @throws Exception if testing goes exceptionally bad
     */
    public void testInterrupt() throws Exception {
        final TripList tripList = new TripList(getContext(), null);
        final Exception exceptionHolder[] = new Exception[1];
        Thread downloader = new Thread(new Runnable() {
            public void run() {
                try {
                    tripList.populate(new FileInputStream("/dev/zero"));
                } catch (Exception e) {
                    exceptionHolder[0] = e;
                }
            }
        }, "Map data downloader");
        downloader.start();

        // Give the downloader time to start waiting for /dev/zero
        Thread.sleep(500);

        tripList.interruptDownload();
        downloader.join(500);
        assertFalse("interruptDownload() should terminate the downloading thread",
            downloader.isAlive());

        // Verify that we got an exception saying that the download was
        // canceled.
        assertNotNull(exceptionHolder[0]);
        assertEquals(IOException.class, exceptionHolder[0].getClass());
        assertEquals("Nedladdningen avbröts", exceptionHolder[0].getMessage());
    }
}
